#define assert(expr) if (.not.(expr)) then; write(*,*) 'speed-test.f90', ':', __LINE__, ': Assertion: ', "expr", ' failed.'; stop; end if
program Speed_Test
  use mpi
  use omp_lib
  use M_LiSDAS
  implicit none
  type (T_Dat) :: dat
  type (T_Mdl) :: mdl
  type (T_SV)  :: sv
  type (T_SV2) :: sv2
  type (T_Mdl2):: mdl2
  real(8)      :: time_interval
  real(8)      :: Variance_System(max_num_Variable)
  real(8)      :: logLik
  integer :: i, j, n_iter, num_Particle_DPM , time
  integer :: ibTime, ieTime, ihz

  call Load_Snapshot("state.bin", dat, mdl, sv, time_interval, Variance_System)

  write(*,*) 'dim(x) = ', sv%num_Variable
  write(*,*) 'dim(m) = ', mdl%num_Reg_Modules
  write(*,*) 'dim(e) = ', mdl%num_Reg_Elements
  write(*,*) sv%StateVariable(1:19,1)

  !logLik = Wrap_Gauss_Likelihood(dat, sv, Variance_Measure)
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! cloning saved state 
  num_Particle_DPM = max_num_Particle_DPM
  assert(num_Particle_DPM .le. max_num_Particle_DPM)

  sv2%num_Variable       = sv%num_Variable 
  sv2%time_StateVariable = sv%time_StateVariable
  sv2%ID_Variable        = sv%ID_Variable
  forall (j = 1:size(sv%StateVariable, 2))
    forall (i = 1:size(sv%StateVariable, 1))
      sv2%StateVariable(1:num_Particle_DPM,i,j) = sv%StateVariable(i,j)
    endforall
  endforall

  mdl2%num_Reg_Elements = mdl%num_Reg_Elements 
  mdl2%num_Reg_Modules  = mdl%num_Reg_Modules 
  mdl2%List_Reg_Elements = mdl%List_Reg_Elements 
  mdl2%List_Reg_Modules = mdl%List_Reg_Modules 
  mdl2%map_e2m = mdl%map_e2m
  mdl2%iList = mdl%iList
  mdl2%iModuleList = mdl%iModuleList
  mdl2%iReactantList = mdl%iReactantList
  mdl2%num_List = mdl%num_List
  forall (j = 1:size(mdl%Para_Reg_Elements, 2))
    forall (i = 1:size(mdl%Para_Reg_Elements, 1))
      mdl2%Para_Reg_Elements(1:num_Particle_DPM,i,j) = mdl%Para_Reg_Elements(i,j)
    endforall
  endforall

  forall (j = 1:size(mdl%Para_Reg_Modules, 2))
    forall (i = 1:size(mdl%Para_Reg_Modules, 1))
      mdl2%Para_Reg_Modules(1:num_Particle_DPM,i,j) = mdl%Para_Reg_Modules(i,j)
    endforall
  endforall

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Speed test

  call system_clock(count_rate = ihz)

  !$omp parallel
  write(*,*) '#threads = ', omp_get_num_threads()
  !$omp end parallel

  ! - original version
  n_iter = 10
  call system_clock(count = ibTime)
  do j = 1, n_iter
    !$omp parallel do firstprivate(sv)
    do i = 1, num_Particle_DPM
    call Run_BioSimulator_version2a&
           (mdl, sv, time_interval, 0, 0, Variance_System)
    end do
    !$omp end parallel do
  end do
  call system_clock(count = ieTime)
  write(*,*) 'version 2a', real(ieTime - ibTime)/real(ihz)
  write(*,*) ieTime, ibTime

  ! - version of thread-iteration moved to the innermost loop
  call system_clock(count = ibTime)
  do j = 1, n_iter
    call Run_BioSimulator_version2b&
      (mdl2, sv2, time_interval, 0, 0, Variance_System, num_Particle_DPM)
  end do
  call system_clock(count = ieTime)
  write(*,*) 'version 2b', real(ieTime - ibTime)/real(ihz)
  write(*,*) ieTime, ibTime
end program
