!-------------------------------------------------------------------------------
! LiSDAS v2.7
!
! Copyright 2010
!   YOSHIDA, Ryo,
!   SAITO, Masaya,
!   NAGAO, Hiromichi,
!   NAKANO, Shin'ya,
!   HIGUCHI, Tomoyuki,
!   The Institute of Statistical Mathematics
!
! Dual licensed under the MIT or GPL Version 3 licenses.
! See 'LICENSE.MIT' for the definition of the MIT license, 
! and 'COPYING.GPL' for the definition of the GPL license.
!-------------------------------------------------------------------------------
! CAUTION: this function is duplicated in transition.
	double precision function Gauss_Likelihood(
*
*		-- Values for data and state variables	
     $		Data, StateVariable, time_Data, time_StateVariable, 
     $		num_Variable, max_time_Data, max_time_StateVariable, 
     $		max_num_Variable, 
*	
*		-- Indicators for observed time points and missing variables
     $		ID_Time_Data, ID_MissVariable, 
*
*		--	Observed noise variances
     $		Variance_Measure
*
     $	)     	
*
	implicit real*8 ( a-h, o-z )
*	
	dimension Data( max_num_Variable, max_time_Data )
	dimension StateVariable( max_num_Variable, * )
	dimension ID_Time_Data( max_time_Data )
	dimension ID_MissVariable( max_num_Variable )
	dimension Variance_Measure( max_num_Variable )
*
	integer*4 time_StateVariable
	integer*4 time_Data
	integer*4 time
*
	Gauss_Likelihood = 0.0d0
*	
	do i = 1, num_Variable
	sum = 0.0d0
	if ( ID_MissVariable(i) .EQ. 1 ) then
	do j = 1, time_Data
	time = ID_Time_Data(j)
	sum = sum + 
     $	( Data(i,j) - StateVariable(i,time) ) ** 2  
	enddo
	Gauss_Likelihood = Gauss_Likelihood + sum / 
     $	( 2.0d0 * Variance_Measure(i) )
	endif
	if( Gauss_Likelihood.GE.10d8 ) then
	write(*,*) '---Warning() -GaussLikelihood -overflow'
	stop
	endif
	enddo
*
	return
	end
*
*
	subroutine Get_Suffix( suffix )
*
	implicit none
	character suffix * 30 
*
	open( 5, file = 'suffix.set' ) 
	read( 5, * ) suffix
	close( 5 ) 
*
	return 
	end
*
*
	subroutine RandomSampling_Inits(
*
*		-- Replaced values for intial state variables are stored 
     $		StateVariable, num_Variable, 
     $		max_time_StateVariable, max_num_Variable,
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_Inits, ID_Prior_Inits,
*     
*		-- Range of hidden regulatory variables
     $     iDoRangeHidden, RangeHidden,
*
*       -- Missing indicator
     $     ID_MissVariable     
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
*
	parameter ( max_sampling = 1000 )
	dimension StateVariable( max_num_Variable, * )
	dimension Prior_Inits( max_num_Variable, 2 )
	dimension ID_Prior_Inits( max_num_Variable )
	dimension RangeHidden(2)
	dimension ID_MissVariable( max_num_Variable )
*
	do i = 1, num_Variable
*
	k = ID_Prior_Inits(i)
*
      if( ( iDoRangeHidden. EQ. 0 ). OR. 
     $    ( ID_MissVariable(k). EQ. 1 ) ) then
*
	do j = 1, max_sampling
	rand = Prior_Inits(i,1) +  dsqrt( Prior_Inits(i,2) ) * rgauss(ix)
	if( rand .GE. 0.0d0 ) exit
	if ( j .EQ. max_sampling ) then
	write(*,*) '---Warning() -RandomSampling_Inits'	
	stop
	endif
	enddo
	StateVariable(k,1) = rand
*
	endif
*
*
      if( ( iDoRangeHidden. EQ. 1 ). AND. 
     $    ( ID_MissVariable(k). EQ. 0 ) ) then
*
	rand = RangeHidden(2) + RangeHidden(1) * runif(ix)
	StateVariable(k,1) = rand
*
      endif

	enddo
*
	return
	end
*
*
	subroutine RandomSampling_Inits_NormalTruncated (
*
*		-- Replaced values for intial state variables are stored 
     $		StateVariable, num_Variable, 
     $		max_time_StateVariable, max_num_Variable,
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_Inits, ID_Prior_Inits,
*
*		-- Range of hidden regulatory variables
     $      iDoRangeHidden, RangeHidden,
*
*		-- Range of hidden regulatory variables
     $      bound_data,     
*
*       -- mRNA indicator
     $      imRNA     
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
*
	parameter ( max_sampling = 1000 )
	dimension StateVariable( max_num_Variable, * )
	dimension Prior_Inits( max_num_Variable, 2 )
	dimension ID_Prior_Inits( max_num_Variable )
	dimension imRNA( max_num_Variable )
	dimension RangeHidden(2)
	dimension bound_data( max_num_Variable, 2 )
*
*
	do i = 1, num_Variable
*
      ic = 0
*
	k = ID_Prior_Inits(i)
*
    1 continue
      ic = ic + 1 
	rand = Prior_Inits(i,1) +  dsqrt( Prior_Inits(i,2) ) * rgauss(ix)
*
	if ( ic .EQ. max_sampling ) then
	write(*,*) '---Warning() -RandomSampling_Inits'
*	do ii = 1, num_Variable
*	write(*,*) '---Variable#', ii	
*	write(*,*) '---Prior mean', Prior_Inits(ii,1)
*	write(*,*) '---Prior vari', Prior_Inits(ii,2)
*	enddo
	stop
	endif	
*
	if( rand. LE. 0.0d0 ) goto 1
*
	if( ( imRNA(k). NE. 0 ). AND. (
     $	( rand .GT. bound_data(k,1) ). OR. ( rand .LT. bound_data(k,2) ) 
     $  )
     $  ) goto 1
*
	if( ( imRNA(k). EQ. 0 ). AND. (
     $	( rand .GT. RangeHidden(1) ). OR. ( rand .LT. RangeHidden(2) ) 
     $  )
     $  ) goto 1
*     
	StateVariable(k,1) = rand
*
	enddo
*
	return
	end
*
*
*
	subroutine RandomSampling_Inits_NormalTruncated_version2 (
*
*		-- Replaced values for intial state variables are stored 
     $		StateVariable, num_Variable, 
     $		max_time_StateVariable, max_num_Variable,
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_Inits, ID_Prior_Inits,
*
*		-- Range of hidden regulatory variables
     $      iDoRangeHidden, RangeHidden,
*
*		-- Range of hidden regulatory variables
     $      bound_data,     
*
*       -- mRNA indicator
     $      imRNA     
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
*
	parameter ( max_sampling = 1000 )
	dimension StateVariable( max_num_Variable, * )
	dimension Prior_Inits( max_num_Variable, 2 )
	dimension ID_Prior_Inits( max_num_Variable )
	dimension imRNA( max_num_Variable )
	dimension RangeHidden(2)
	dimension bound_data( max_num_Variable, 2 )
*
*
	do i = 1, num_Variable
*
      ic = 0
*
	k = ID_Prior_Inits(i)
*
    1 continue
      ic = ic + 1 
	rand = Prior_Inits(i,1) +  dsqrt( Prior_Inits(i,2) ) * rgauss(ix)
*
*	if ( ic .EQ. max_sampling ) then
*	write(*,*) '---Warning() -RandomSampling_Inits'
*	do ii = 1, num_Variable
*	write(*,*) '---Variable#', ii	
*	write(*,*) '---Prior mean', Prior_Inits(ii,1)
*	write(*,*) '---Prior vari', Prior_Inits(ii,2)
*	enddo
*	stop
*	endif	
*
	if( rand. LE. 0.0d0 ) goto 1
*
	if( ( imRNA(k). NE. 0 ). AND. (
     $	( rand .GT. bound_data(k,1) ). OR. ( rand .LT. bound_data(k,2) ) 
     $  )
     $  ) goto 1
*
	if( ( imRNA(k). EQ. 0 ). AND. (
     $	( rand .GT. RangeHidden(1) ). OR. ( rand .LT. RangeHidden(2) ) 
     $  )
     $  ) goto 1
*     
	StateVariable(k,1) = rand
*
	enddo
*
	return
	end
*
*
	subroutine RandomSampling_Modules(
*
*		-- Replaced values for intial state variables are stored 
     $		Para_Reg_Modules, List_Reg_Modules, num_Reg_Modules, 
     $		max_num_Reg_Modules, max_tmp,
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_ParaModules, ID_Prior_ParaModules, 
     $		num_Para2, max_num_Para2
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
	parameter ( max_sampling = 1000 )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension Prior_ParaModules( max_num_Para2, 2 )
	dimension ID_Prior_ParaModules( max_num_Para2 )
*
	do i = 1, num_Para2
	k = ID_Prior_ParaModules(i)
	do j = 1, max_sampling
	rand = Prior_ParaModules(i,1) + 
     $	dsqrt( Prior_ParaModules(i,2) ) * rgauss( ix )
	if( rand .GE. 0.0d0 ) exit
	if ( j .EQ. max_sampling ) then
	write(*,*) '---Warning() -RandomSampling_Modules'	
	stop
	endif
	enddo
	do j = 1, num_Reg_Modules
	if( List_Reg_Modules(j,3). EQ. k ) 
     $	Para_Reg_Modules(j,1) = rand
	enddo
	enddo
*
	return
	end
*
*
	subroutine RandomSampling_Modules_S2(
*
*		-- Replaced values for intial state variables are stored 
     $		Para_Reg_Modules, List_Reg_Modules, num_Reg_Modules, 
     $		max_num_Reg_Modules, max_tmp,
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_ParaModules, ID_Prior_ParaModules, 
     $		num_Para2, max_num_Para2
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
	parameter ( max_sampling = 1000 )
	parameter ( itest = 0 )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension Prior_ParaModules( max_num_Para2, 2 )
	dimension ID_Prior_ParaModules( max_num_Para2 )
	dimension stateMaxMin( 12, 2 )
	dimension ratesTime( 12 )
*
	do i = 1, num_Para2
	k = ID_Prior_ParaModules(i)
	do j = 1, max_sampling
	rand = Prior_ParaModules(i,1) + 
     $	dsqrt( Prior_ParaModules(i,2) ) * rgauss( ix )
	if( rand .GE. 0.0d0 ) exit
	if ( j .EQ. max_sampling ) then
	write(*,*) '---Warning() -RandomSampling_Modules_S2'	
	stop
	endif
	enddo
	do j = 1, num_Reg_Modules
	if( List_Reg_Modules(j,3). EQ. k ) 
     $	Para_Reg_Modules(j,1) = rand
	enddo
	enddo
*
*---------------------------------------------------------------------------------
*	Maximum and minimum intensities of mRNAs
*---------------------------------------------------------------------------------
*
*	--- per ---
	stateMaxMin(1,1) = 1.80d0
	stateMaxMin(1,2) = 0.31d0
*	--- cry ---
	stateMaxMin(3,1) = 1.32d0
	stateMaxMin(3,2) = 0.80d0
*	--- rev/erb ---
	stateMaxMin(6,1) = 1.34d0
	stateMaxMin(6,2) = 0.78d0
*	--- clock ---
	stateMaxMin(8,1) = 4.50d0
	stateMaxMin(8,2) = 0.78d0
*	--- bmal ---
	stateMaxMin(10,1) = 1.28d0
	stateMaxMin(10,2) = 0.77d0
*
*-----------------------------------------------------------------------------------------------------------
*	Maximum and minimum intensities of proteins
*		* MUST be max(protein) / min(protein) = discount * max(mRNA) / min(mRNA) with 0 < discount < 1
*-----------------------------------------------------------------------------------------------------------
*
	discount = 1.3d0
*	--- Per ---
	irna = 1
	iprt = 2
	stateMaxMin(iprt,1) = 3d0
	stateMaxMin(iprt,2) = stateMaxMin(iprt,1) * stateMaxMin(irna,2) 
     $	/ ( discount * stateMaxMin(irna,1) )
*	stateMaxMin(iprt,2) = 0.01d0
*	write(*,*) stateMaxMin(iprt,1)/stateMaxMin(iprt,2)
*	write(*,*) stateMaxMin(irna,1)/stateMaxMin(irna,2)
*	write(*,*) stateMaxMin(iprt,1) * stateMaxMin(irna,2) 
*     $	- stateMaxMin(iprt,2) * stateMaxMin(irna,1)
*	write(*,*) stateMaxMin(iprt,1), stateMaxMin(iprt,2)
*	--- Cry ---
	irna = 3
	iprt = 4
	stateMaxMin(iprt,1) = 3d0
	stateMaxMin(iprt,2) = stateMaxMin(iprt,1) * stateMaxMin(irna,2) 
     $	/ ( discount * stateMaxMin(irna,1) )
*	stateMaxMin(iprt,2) = 0.01d0
*	write(*,*) stateMaxMin(iprt,1), stateMaxMin(iprt,2)
*	--- Rev/Erb ---
	irna = 6
	iprt = 7 
	stateMaxMin(iprt,1) = 2.5d0
	stateMaxMin(iprt,2) = stateMaxMin(iprt,1) * stateMaxMin(irna,2) 
     $	/ ( discount * stateMaxMin(irna,1) )
*	stateMaxMin(iprt,2) = 0.01d0
*	--- Clock ---
	irna = 8
	iprt = 9 
	stateMaxMin(iprt,1) = 1.3d0
	stateMaxMin(iprt,2) = stateMaxMin(iprt,1) * stateMaxMin(irna,2) 
     $	/ ( discount * stateMaxMin(irna,1) )
*	stateMaxMin(iprt,2) = 0.01d0
*	--- Bmal ---
	irna = 10
	iprt = 11 
	stateMaxMin(iprt,1) = 1.3d0
	stateMaxMin(iprt,2) = stateMaxMin(iprt,1) * stateMaxMin(irna,2) 
     $	/ ( discount * stateMaxMin(irna,1) )
*	stateMaxMin(iprt,2) = 0.01d0
*
	stateMaxMin(5,1) = 6.0d0	! Per_Cry max
	stateMaxMin(5,2) = 3.0d0	! Per_Cry min
	stateMaxMin(12,1) = 5.0d0	! Clock_Bmal max
	stateMaxMin(12,2) = 0.2d0	! Clock_Bmal min
*
*-----------------------------------------------------------------------------------------------------------
*	Degradation rates of Regulatory proteins
*-----------------------------------------------------------------------------------------------------------
*
	irun_DegRegPrt = 0
	if( irun_DegRegPrt. EQ. 1 ) then
	iprt = 5
	ib1 = 2
	ib2 = 4
	Para_Reg_Modules(19,1) = Para_Reg_Modules(18,1) * 
     $	stateMaxMin(ib1,1) * stateMaxMin(ib2,1) /
     $	stateMaxMin(iprt,1)
*
	iprt = 12
	ib1 = 11
	ib2 = 9
	Para_Reg_Modules(44,1) = Para_Reg_Modules(43,1) * 
     $	stateMaxMin(ib1,1) * stateMaxMin(ib2,1) /
     $	stateMaxMin(iprt,1)
	endif
*
*-----------------------------------------------------------------------------------------------------------
*	Degradation rates of mRNAs
*-----------------------------------------------------------------------------------------------------------
*
*	Response times in RNA decays (unit; # time steps)
	ratesTime(1) =  191d0	! per
	ratesTime(3) =  222d0	! cry
	ratesTime(6) =  254d0	! rev/erb
	ratesTime(10) =  70d0	! bmal
*
	scale = 0.80d0 + 0.4d0 * runif( ix )
*	scale = 1.0d0
*
	irun_DegRna = 1
	if( irun_DegRna. EQ. 1 ) then
	timeInterval = 0.1d0 
*	--- Degradation rate of per ---
	irna = 1
	Para_Reg_Modules(3,1) = ( 1.0d0 / timeInterval ) * ( 1.0d0 -  
     $ dexp( dlog( stateMaxMin(irna,2) ) - dlog( stateMaxMin(irna,1) ) ) 
     $	** ( 1.0d0 / ratesTime(irna) )  ) * scale
*	--- Degradation rate of cry ---
	irna = 3
	Para_Reg_Modules(10,1) = ( 1.0d0 / timeInterval )  * ( 1.0d0 -  
     $ dexp( dlog( stateMaxMin(irna,2) ) - dlog( stateMaxMin(irna,1) ) ) 
     $	** ( 1.0d0 / ratesTime(irna) ) ) * scale
*	--- Degradation rate of rev/erb ---
	irna = 6
	Para_Reg_Modules(23,1) = ( 1.0d0 / timeInterval )  * ( 1.0d0 -  
     $ dexp( dlog( stateMaxMin(irna,2) ) - dlog( stateMaxMin(irna,1) ) ) 
     $	** ( 1.0d0 / ratesTime(irna) ) ) * scale
*	--- Degradation rate of bmal ---
	irna = 10
	Para_Reg_Modules(36,1) = ( 1.0d0 / timeInterval )  * ( 1.0d0 -  
     $ dexp( dlog( stateMaxMin(irna,2) ) - dlog( stateMaxMin(irna,1) ) ) 
     $	** ( 1.0d0 / ratesTime(irna) ) ) * scale
*
*	write(*,*) 'per', 1/Para_Reg_Modules(3,1)
*	write(*,*) 'cry', 1/Para_Reg_Modules(10,1)
*	write(*,*) 'rev/erb', 1/Para_Reg_Modules(23,1)
*	write(*,*) 'bmal', 1/Para_Reg_Modules(36,1)
	endif
*
*--------------------------------------------
*	Baseline levels of mRNA syntheses
*--------------------------------------------
*
	irun_baselSS = 0
	if( irun_baselSS. EQ. 1 ) then
*	--- Baseline of per ---
	irna = 1
	Para_Reg_Modules(2,1) = Para_Reg_Modules(3,1) *stateMaxMin(irna,2)
*	--- Baseline of cry ---
	irna = 3
	Para_Reg_Modules(12,1) =Para_Reg_Modules(10,1)*stateMaxMin(irna,2)
*	--- Baseline of rev/erb ---
	irna = 6
	Para_Reg_Modules(21,1) =Para_Reg_Modules(23,1)*stateMaxMin(irna,2)
*	--- Baseline of bmal ---
	irna = 10
	Para_Reg_Modules(35,1) =Para_Reg_Modules(36,1)*stateMaxMin(irna,2)
	endif
*
	irun_baselZero = 1
	if( irun_baselZero. EQ. 1 ) then
	Para_Reg_Modules(2,1) = 0.0d0	! per
	Para_Reg_Modules(12,1) = 0.0d0	! cry
	Para_Reg_Modules(21,1) = 0.0d0	! rev/erb
	Para_Reg_Modules(35,1) = 0.0d0	! bmal
	endif
*
	scale = 0.80d0 + 0.4d0 * runif( ix )
*	scale = 1.0d0
*	Transcription rate of per
	Para_Reg_Modules(4,1) = 
     $	scale * ( Para_Reg_Modules(3,1) * stateMaxMin(1,1)
     $	- Para_Reg_Modules(2,1) )
*	Transcription rate of cry
	Para_Reg_Modules(11,1) = 
     $	scale * ( Para_Reg_Modules(10,1) * stateMaxMin(3,1)
     $	- Para_Reg_Modules(12,1) )
*	Transcription rate of rev/erb
	Para_Reg_Modules(22,1) = 
     $	scale * ( Para_Reg_Modules(23,1) * stateMaxMin(6,1)
     $	- Para_Reg_Modules(21,1) )
*	Transcription rate of bmal
	Para_Reg_Modules(37,1) = 
     $	scale * ( Para_Reg_Modules(36,1) *stateMaxMin(10,1)
     $	- Para_Reg_Modules(35,1) ) 
*	write(*,*) Para_Reg_Modules(23,1), Para_Reg_Modules(24,1)
*
*--------------------------------------------
*	Rates of translation
*--------------------------------------------
*
	irun_transL = 0
	if( irun_transL. EQ. 1 ) then
*	--- Translation of Per ---
	iprt = 2	! Per
	irna = 1	! per
	ibnd = 4	! Cry
	a = stateMaxMin(irna,2) * stateMaxMin(iprt,1)
	b = stateMaxMin(irna,1) * stateMaxMin(iprt,2)
	c = ( stateMaxMin(ibnd,1) - stateMaxMin(ibnd,2) ) *
     $	stateMaxMin(iprt,1) * stateMaxMin(iprt,2)
	Para_Reg_Modules(6,1) = Para_Reg_Modules(8,1) *  c / ( a - b )
	Para_Reg_Modules(7,1) = 
     $	( Para_Reg_Modules(6,1) * stateMaxMin(irna,1) 
     $	- Para_Reg_Modules(8,1) * stateMaxMin(ibnd,2) * 
     $	stateMaxMin(iprt,1)) / stateMaxMin(iprt,1)
	if( Para_Reg_Modules(6,1). LT. 0.0d0 ) then
	write(*,*) iprt, a - b
	stop
	endif
*	--- Translation of Cry ---
	iprt = 4	! Cry
	irna = 3	! cry
	ibnd = 2	! Per
	a = stateMaxMin(irna,2) * stateMaxMin(iprt,1)
	b = stateMaxMin(irna,1) * stateMaxMin(iprt,2)
	c = ( stateMaxMin(ibnd,1) - stateMaxMin(ibnd,2) ) *
     $	stateMaxMin(iprt,1) * stateMaxMin(iprt,2)
	Para_Reg_Modules(14,1) = Para_Reg_Modules(15,1) * c / ( a - b )
	Para_Reg_Modules(16,1) = 
     $	( Para_Reg_Modules(14,1) * stateMaxMin(irna,1) 
     $	- Para_Reg_Modules(15,1) * stateMaxMin(ibnd,2) * 
     $	stateMaxMin(iprt,1) ) / stateMaxMin(iprt,1)
*	write(*,*) Para_Reg_Modules(14,1) * stateMaxMin(irna,1) /
*     $	(Para_Reg_Modules(16,1)+Para_Reg_Modules(15,1)*
*     $	stateMaxMin(ibnd,2) )
*	write(*,*) Para_Reg_Modules(14,1) * stateMaxMin(irna,2) /
*     $	(Para_Reg_Modules(16,1)+Para_Reg_Modules(15,1)*
*     $	stateMaxMin(ibnd,1) )
	if( Para_Reg_Modules(14,1) .LT. 0.0d0 ) then
	write(*,*) iprt
	stop
	endif
*	--- Translation of Bmal ---
	iprt = 11	! Bmal
	irna = 10	! bmal
	ibnd = 9	! Clock
	a = stateMaxMin(irna,2) * stateMaxMin(iprt,1)
	b = stateMaxMin(irna,1) * stateMaxMin(iprt,2)
	c = ( stateMaxMin(ibnd,1) - stateMaxMin(ibnd,2) ) *
     $	stateMaxMin(iprt,1) * stateMaxMin(iprt,2)
	Para_Reg_Modules(39,1) = Para_Reg_Modules(41,1) * c / ( a - b )
	Para_Reg_Modules(40,1) = 
     $	( Para_Reg_Modules(39,1) * stateMaxMin(irna,1) 
     $	- Para_Reg_Modules(41,1) * stateMaxMin(ibnd,2) * 
     $	stateMaxMin(iprt,1) ) / stateMaxMin(iprt,1)
*	write(*,*) Para_Reg_Modules(39,1) * stateMaxMin(irna,1) /
*     $	(Para_Reg_Modules(40,1)+Para_Reg_Modules(41,1)*
*     $	stateMaxMin(ibnd,2) )
*	write(*,*) Para_Reg_Modules(39,1) * stateMaxMin(irna,2) /
*     $	(Para_Reg_Modules(40,1)+Para_Reg_Modules(41,1)*
*     $	stateMaxMin(ibnd,1) )
	if( Para_Reg_Modules(39,1). LT. 0.0d0 ) then
	write(*,*) iprt, a - b
	stop
	endif
*	--- Translation of Clock ---
	iprt = 9	! Clock
	irna = 8	! clock
	ibnd = 11	! Bmal
	a = stateMaxMin(irna,2) * stateMaxMin(iprt,1)
	b = stateMaxMin(irna,1) * stateMaxMin(iprt,2)
	c = ( stateMaxMin(ibnd,1) - stateMaxMin(ibnd,2) ) *
     $	stateMaxMin(iprt,1) * stateMaxMin(iprt,2)
	Para_Reg_Modules(31,1) = Para_Reg_Modules(33,1) * c / ( a - b )
	Para_Reg_Modules(32,1) = 
     $	( Para_Reg_Modules(31,1) * stateMaxMin(irna,1) 
     $	- Para_Reg_Modules(33,1) * stateMaxMin(ibnd,2) * 
     $	stateMaxMin(iprt,1) ) / stateMaxMin(iprt,1)
*	write(*,*) Para_Reg_Modules(31,1) * stateMaxMin(irna,1) /
*     $	(Para_Reg_Modules(32,1)+Para_Reg_Modules(33,1)*
*     $	stateMaxMin(ibnd,2) )
*	write(*,*) Para_Reg_Modules(31,1) * stateMaxMin(irna,2) /
*     $	(Para_Reg_Modules(32,1)+Para_Reg_Modules(33,1)*
*     $	stateMaxMin(ibnd,1) )
*	stop
	if( Para_Reg_Modules(31,1). LT. 0.0d0 ) then
	write(*,*) iprt, a - b
	stop
	endif
	endif
*
	do j = 1, num_Reg_Modules
	if( Para_Reg_Modules(j,1). LT. 0.0d0 ) then
     	Para_Reg_Modules(j,1) = 0.0d0
	write(*,*) j 
	stop
	endif
	enddo
*
	return
	end
*
*
	subroutine RandomSampling_Modules_AutoS2Rna(
*
*		-- Replaced values for parameters in module [Para_Reg_Modules] and relevant parameters 
     $		Para_Reg_Modules, List_Reg_Modules, 
     $		num_Reg_Modules, max_num_Reg_Modules, max_tmp,
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_ParaModules, ID_Prior_ParaModules, 
     $		num_Para2, max_num_Para2,
*
*		-- Additional parameters
     $		degrad_RNA, synthesis_RNA, 
     $      max_num_Variable, num_Variable,
     $		ID_MissVariable, strModules,
*
*		-- Noise control parameters in generating particles relevant to transcriprion parameters
     $	    scale_variance, method_sample, 
     $      input_zero_baseline 							
     $	)
*
	implicit real*8 ( a-h, o-z )
	parameter ( max_sampling = 1000 )
	parameter ( itest = 0 )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension Prior_ParaModules( max_num_Para2, 2 )
	dimension ID_Prior_ParaModules( max_num_Para2 )
*
	dimension degrad_RNA( max_num_Variable )
	dimension synthesis_RNA( max_num_Variable )
	dimension ID_MissVariable( max_num_Variable )
*
	character strModules( max_num_Reg_Modules, 3 ) * 30
*
	character tag_synth * (*)
	character tag_rnadeg * (*)
	character tag_base * (*)
	parameter ( tag_synth = 'Transcription' )
	parameter ( tag_rnadeg = 'RnaDegradation' )
	parameter ( tag_base = 'Baseline' )
*
*
      if( method_sample. eq. 1 ) then 

*     /Draw transcriptional parameters from DP base measure 
*     on rates of synthesis and degradtion
*     -------------------------------------------------------
	do j = 1, num_Reg_Modules
*
*     /Product ID
	iprd = List_Reg_Modules(j,2)
*
*	/ Rate of transcription
	if( ( ID_MissVariable(iprd). EQ. 1 ). AND. 
     $	( strModules(j,2). EQ. tag_synth ) ) then
	scale = 0.9d0 + 0.2d0 * runif( ix )
     	Para_Reg_Modules(j,1) = scale * synthesis_RNA(iprd)
	endif
*
*	/ Degradation rate
	if( ( ID_MissVariable(iprd). EQ. 1 ). AND. 
     $	( strModules(j,2). EQ. tag_rnadeg ) ) then
	scale = 0.9d0 + 0.2d0 * runif( ix )
	Para_Reg_Modules(j,1) = scale * degrad_RNA(iprd)
	endif
*
	enddo
*
*
      else
*
*     /Draw transcriptional parameters by Gaussian distribution
*     -------------------------------------------------------
	do i = 1, num_Para2
*
	k = ID_Prior_ParaModules(i)
*	
	do j = 1, max_sampling
	rand = Prior_ParaModules(i,1) + 
     $	    dsqrt( Prior_ParaModules(i,2) ) * rgauss(ix) 
     $      * scale_variance
*
	if( rand .GE. 0.0d0 ) exit
	if ( j .EQ. max_sampling ) then
	write(*,*) '---Warning() -RandomSampling_Modules_AutoS2Rna'	 
	stop
	endif
	enddo
*
	do j = 1, num_Reg_Modules
	if( List_Reg_Modules(j,3). EQ. k )
     $	    Para_Reg_Modules(j,1) = rand
	enddo
*
	enddo
*
	endif
*
*     / Baseline is set to zero
      if( input_zero_baseline.eq.1 ) then
	do i = 1, num_Reg_Modules
	if ( strModules(i,2). EQ. tag_base  ) 
     $	        Para_Reg_Modules(i,1) = 0.0d0
	enddo
	endif
*
	return
	end
*
*
      subroutine Shift_Noise_Variances( 
*      
     $      Variance_Measure, imRNA, ID_MissVariable, 
     $      cov, max_num_Variable, num_Variable
*     
     $  )	
*
	implicit real*8 ( a-h, o-z )
	integer*2 tab /2313/
	dimension Variance_Measure(*)
	dimension cov( max_num_Variable, * )
	dimension imRNA(*)
	dimension ID_MissVariable(*)
*
      loc = 0
      do i = 1, num_Variable
      if( ( imRNA(i).ne.0 ).and.( ID_MissVariable(i).eq.1 ) ) then
      Variance_Measure(i) = cov(i,i)
      loc = i
      endif
      enddo
*
      do i = 1, num_Variable
      if( ( imRNA(i).ne.0 ).and.( ID_MissVariable(i).ne.1 ) ) then
      Variance_Measure(i) = cov(loc,loc)
      endif
      enddo
*
      return
      end	
*
*
	subroutine RandomSampling_Modules_AutoMaxRna(
*
*		-- Replaced values for intial state variables are stored 
     $		Para_Reg_Modules, List_Reg_Modules, num_Reg_Modules, 
     $		max_num_Reg_Modules, max_tmp,
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_ParaModules, ID_Prior_ParaModules, 
     $		num_Para2, max_num_Para2,
*
*		-- 
     $		value_maxmin, loc_maxmin, max_num_Variable, num_Variable,
     $		ID_MissVariable, strModules,
*
*		-- Noise control parameters in generating particles relevant to transcriprion parameters
     $	    scaleUnifNoiseDegRNA, isample_G0,
*
     $      itranslation, icomplex             							
     $	)
*
	implicit real*8 ( a-h, o-z )
	parameter ( max_sampling = 1000 )
	parameter ( itest = 0 )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension Prior_ParaModules( max_num_Para2, 2 )
	dimension ID_Prior_ParaModules( max_num_Para2 )
*
	dimension value_maxmin( max_num_Variable, 2 )
	dimension loc_maxmin( max_num_Variable, 2 )
	dimension ID_MissVariable( max_num_Variable )
      dimension itranslation( max_num_Variable )
      dimension icomplex( max_num_Variable )      
*
	character strModules( max_num_Reg_Modules, 3 ) * 30	
*
	character tag_synth * (*)
	character tag_rnadeg * (*)
	character tag_base * (*)
	character tag_transl * (*)
	character tag_prodeg * (*)	
	character tag_bind * (*)		
*	
	parameter ( tag_synth = 'Transcription' )
	parameter ( tag_rnadeg = 'RnaDegradation' )
	parameter ( tag_base = 'Baseline' )
*
	parameter ( tag_transl = 'Translation' )
	parameter ( tag_prodeg = 'ProteinDegradation' )
      parameter ( tag_bind = 'Binding' )
*
*
      if ( isample_G0. eq. 0 ) then
*     / Random sampling for 'Parameters in Reaction Module'
	do i = 1, num_Para2
*
	k = ID_Prior_ParaModules(i)
*
*     / Generate a positive part Gaussian random number 
	do j = 1, max_sampling
	rand = Prior_ParaModules(i,1) + 
     $	dsqrt( Prior_ParaModules(i,2) ) * rgauss( ix )
	if( rand .GE. 0.0d0 ) exit
	if ( j .EQ. max_sampling ) then
	write(*,*) '---Warning() -RandomSampling_Modules_AutoMaxRna'	
	stop
	endif
	enddo
*
	do j = 1, num_Reg_Modules
*
	if( List_Reg_Modules(j,3). EQ. k ) 
     $	Para_Reg_Modules(j,1) = rand
*     
	enddo
*
	enddo
*
      endif	
*
*
      if( isample_G0. eq. 1 ) then
*      
	do j = 1, num_Variable
*
*     RNA degradation and synthesis
*     ----------------------------------------------------------- 
	if( ID_MissVariable(j). EQ. 1 ) then 
*
	do i = 1,  num_Reg_Modules
*
	if( ( List_Reg_Modules(i,2). EQ. j ). AND. 
     $	( strModules(i,2). EQ. tag_synth ) ) then
	locSynth = i
	endif
*
	if( ( List_Reg_Modules(i,2). EQ. j ). AND. 
     $	( strModules(i,2). EQ. tag_rnadeg ) ) then
	locDeg = i
	endif
*
	enddo
*
      rand = runif(ix)
      Para_Reg_Modules(locSynth,1) = rand
	Para_Reg_Modules(locDeg,1) = Para_Reg_Modules(locSynth,1) 
     $	/ value_maxmin( j, 1 )
*     	
      endif
*
      enddo
*
      endif      
*
*     ----------------------------------------------------------- 
*     Protein degradation and translation
*     -----------------------------------------------------------
      if( isample_G0. eq. 1 ) then
*      
      do j = 1, num_Variable
*
      mRNA = itranslation(j)       
      if( mRNA. NE. 0 ) then
*
	do i = 1,  num_Reg_Modules
*
	if( ( List_Reg_Modules(i,2). EQ. j ). AND. 
     $	( strModules(i,2). EQ. tag_transl ) ) then
	locTransl = i
	endif
*
	if( ( List_Reg_Modules(i,2). EQ. j ). AND. 
     $	( strModules(i,2). EQ. tag_prodeg ) ) then
	locProDeg = i
	endif
*
	enddo
*
      rand = runif(ix)
      Para_Reg_Modules(locTransl,1) = rand
	Para_Reg_Modules(locProDeg,1) = Para_Reg_Modules(locTransl,1) 
     $	* value_maxmin( mRNA, 1 )
*     	
	endif
*
	enddo
*
      endif	
*
*
      if( isample_G0. eq. 1 ) then
*      
      do j = 1, num_Variable
*
      if( icomplex(j). NE. 0 ) then
*
	do i = 1,  num_Reg_Modules
*
	if( ( List_Reg_Modules(i,2). EQ. j ). AND. 
     $	( strModules(i,2). EQ. tag_bind ) ) then
	locBind = i
	endif
*
	if( ( List_Reg_Modules(i,2). EQ. j ). AND. 
     $	( strModules(i,2). EQ. tag_prodeg ) ) then
	locProDeg = i
	endif
*
	enddo
*
      rand = runif(ix)
      Para_Reg_Modules(locBind,1) = rand
	Para_Reg_Modules(locProDeg,1) = Para_Reg_Modules(locBind,1)
*
      id_r = List_Reg_Modules( locBind, 3 )
*      write(*,*) id_r
*
      do i = 1, num_Reg_Modules
      if( List_Reg_Modules(i,3). eq. id_r ) then
      Para_Reg_Modules(i,1) = Para_Reg_Modules(locBind,1)
*      write(*,*) List_Reg_Modules(i,1), List_Reg_Modules(i,3), 
*     $ Para_Reg_Modules(i,2)
      endif 
      enddo
*     	
	endif
*
	enddo
*
      endif	
*
*
	do i = 1, num_Reg_Modules
	if ( strModules(i,2). EQ. tag_base  ) 
     $	Para_Reg_Modules(i,1) = 0.0d0
	enddo
*
	return
	end
*
*
	subroutine RandomSampling_Elements(
*
*		-- Replaced values for parameters in elements [Para_Reg_Elements] and related variables
     $		Para_Reg_Elements, List_Reg_Elements, 
     $		num_Reg_Elements, max_num_Reg_Elements,
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_ParaElements, ID_Prior_ParaElements, 
     $		num_Para1, max_num_Para1, scale_variance
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
	parameter ( max_sampling = 1000 )
	dimension Para_Reg_Elements( max_num_Reg_Elements, * )
	dimension List_Reg_Elements( max_num_Reg_Elements, * )
	dimension Prior_ParaElements( max_num_Para1, 4 )
	dimension ID_Prior_ParaElements( max_num_Para1 )
*
	do i = 1, num_Para1
	k = ID_Prior_ParaElements(i)
*
      do j = 1, max_sampling
      rg = rgauss(ix)
	rand1 = Prior_ParaElements(i,1) + 
     $	    scale_variance * dsqrt( Prior_ParaElements(i,3) ) * rg
	if( rand1 .GE. 0.0d0 ) goto 1
	enddo
    1 continue
*
      do j = 1, max_sampling
      rg = rgauss(ix)
	rand2 = Prior_ParaElements(i,2) +
     $	    scale_variance * dsqrt( Prior_ParaElements(i,4) ) * rg
	if( rand2 .GE. 0.0d0 ) goto 2
	enddo
    2 continue
*
	do j = 1, num_Reg_Elements
	if( List_Reg_Elements(j,1). EQ. k ) then
     	Para_Reg_Elements(j,1) = rand1
      Para_Reg_Elements(j,2) = rand2
	endif
	enddo
*
	enddo
*
	return
	end
*
*
	subroutine RandomSampling_Elements_S2(
*
*		-- Replaced values for intial state variables are stored 
     $		Para_Reg_Elements, List_Reg_Elements, 
     $		num_Reg_Elements, max_num_Reg_Elements, max_tmp,
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_ParaElements, ID_Prior_ParaElements, 
     $		num_Para1, max_num_Para1
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
	dimension Para_Reg_Elements( max_num_Reg_Elements, * )
	dimension List_Reg_Elements( max_num_Reg_Elements, * )
	dimension Prior_ParaElements( max_num_Para1, 4 )
	dimension ID_Prior_ParaElements( max_num_Para1 )
	dimension edgeProb( max_num_Para1, 2 )
	valueMax = 10.0d0
	valueMin = 1.0d-6
*
	do i = 1, num_Para1
	edgeProb(i,1) = 0.2d0
	edgeProb(i,2) = 0.8d0
	enddo
*
	do i = 1, num_Para1
	k = ID_Prior_ParaElements(i)
*
    1 continue
	rand1 = Prior_ParaElements(i,1) + 
     $	dsqrt( Prior_ParaElements(i,3) ) * rgauss( ix )
*
	seed = runif(ix)
	if( seed. LT. edgeProb(i,1) ) then
	rand1 = valueMin
	write(*,*) '----- zero', valueMin
	endif
	if( seed. GT. edgeProb(i,2) ) then
	rand1 = valueMax
	write(*,*) '----- inf', valueMax
	endif
*
	if( rand1 .LT. 0.0d0 ) goto 1
*
    2 continue
	rand2 = Prior_ParaElements(i,2) + 
     $	dsqrt( Prior_ParaElements(i,4) ) * rgauss( ix )
*
	if( rand2 .LT. 0.0d0 ) goto 2
*
	do j = 1, num_Reg_Elements
	if( List_Reg_Elements(j,1). EQ. k ) then
     	Para_Reg_Elements(j,1) = rand1
      Para_Reg_Elements(j,2) = rand2
	endif
	enddo
*
	enddo
*
	return
	end
*
*
	subroutine RandomSampling_Elements_S3(
*
*		-- Replaced values for intial state variables are stored 
     $		Para_Reg_Elements, List_Reg_Elements, num_Reg_Elements, 
     $		max_num_Reg_Elements, max_tmp,
*     
     $      itranscript, switch, List_Reg_Elements_switch,
     $      num_transcript, max_num_transcript,     
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_ParaElements, ID_Prior_ParaElements, 
     $		num_Para1, max_num_Para1
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Prior_ParaElements( max_num_Para1, 4 )
	dimension ID_Prior_ParaElements( max_num_Para1 )
      dimension itranscript( max_num_transcript, 3 )
      dimension switch( max_num_transcript )
      dimension List_Reg_Elements_switch( max_num_transcript ) 
*
      do i = 1, num_transcript
      if( runif(ix).ge.0.9d0 ) then
      k = List_Reg_Elements_switch(i)
      Para_Reg_Elements(k,1) = switch(i)
*      write(*,*) k, Para_Reg_Elements(k,1)
      endif 
      enddo 
*
	return
	end	
*
*
	subroutine RandomSampling_Elements_S4(
*
*		-- Replaced values for parameters in Reg_Elements
     $		Para_Reg_Elements, List_Reg_Elements, num_Reg_Elements, 
     $		max_num_Reg_Elements, max_tmp,
*     
     $      itranscript, switch, List_Reg_Elements_switch,
     $      num_transcript, max_num_transcript, bound_model,
     $      max_num_Variable,   
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_ParaElements, ID_Prior_ParaElements, 
     $		num_Para1, max_num_Para1, method
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Prior_ParaElements( max_num_Para1, 4 )
	dimension ID_Prior_ParaElements( max_num_Para1 )
      dimension itranscript( max_num_transcript, 3 )
      dimension switch( max_num_transcript )
      dimension List_Reg_Elements_switch( max_num_transcript ) 
	dimension bound_model( max_num_Variable, 2 )      
*
*      do i = 1, num_Reg_Elements
*      k = List_Reg_Elements(i,3)
*      if( List_Reg_Elements(i,2).eq.18 )
*     $  Para_Reg_Elements(i,1) = bound_model(k,2) 
*      if( List_Reg_Elements(i,2).eq.19 )
*     $  Para_Reg_Elements(i,1) = bound_model(k,1)
*      enddo
*
      do i = 1, num_transcript
      if( runif(ix).ge.0.5d0 ) then
*
*     /Transcription factor
      ii = itranscript(i,2)
*     /Row ID of [Para_Reg_Elements] matrix       
      k = List_Reg_Elements_switch(i)
*     /      
      scale = 0.5d0 + runif(ix) * 0.5d0
      if( itranscript(i,3). eq. -1 ) then
*     /Inhibitor      
      temp = bound_model(ii,2) + scale * 
     $  ( bound_model(ii,1) - bound_model(ii,2) )
      Para_Reg_Elements(k,1) = temp
      else
*     /Activator      
      temp = bound_model(ii,2) + (1-scale) * 
     $  ( bound_model(ii,1) - bound_model(ii,2) )
      Para_Reg_Elements(k,1) = temp 
      endif
*
      if( method.eq.1 ) then
      ii = List_Reg_Elements(k,1)
      do j = 1, num_Para1
      if( ID_Prior_ParaElements(j).eq.ii ) then 
      Prior_ParaElements(j,1) = temp
      goto 1
      endif
      enddo
    1 continue
      endif
*
      endif 
      enddo 
*
	return
	end
*
*
	subroutine RandomSampling_Elements_S5(
*
*		-- Replaced values for intial state variables are stored 
     $		Para_Reg_Elements, List_Reg_Elements, num_Reg_Elements, 
     $		max_num_Reg_Elements, max_tmp,
*     
     $      itranscript, switch, List_Reg_Elements_switch,
     $      num_transcript, max_num_transcript,     
*
*		-- Seed for random number generator
     $		ix, 
*
*		-- Prior distributions
     $		Prior_ParaElements, ID_Prior_ParaElements, 
     $		num_Para1, max_num_Para1
*     							
     $	)
*
	implicit real*8 ( a-h, o-z )
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Prior_ParaElements( max_num_Para1, 4 )
	dimension ID_Prior_ParaElements( max_num_Para1 )
      dimension itranscript( max_num_transcript, 3 )
      dimension switch( max_num_transcript )
      dimension List_Reg_Elements_switch( max_num_transcript ) 
*
      do i = 1, num_transcript
      if( runif(ix).ge.0.9d0 ) then
      k = List_Reg_Elements_switch(i)
*      Para_Reg_Elements(k,1) = switch(i)
      do j = 1, num_Para1
      if( ID_Prior_ParaElements(j).eq.k ) then
      Prior_ParaElements(j,1) = switch(i)
      goto 1
      endif
      enddo 
      endif
    1 continue  
      enddo 
*
	return
	end		
*
*
	subroutine Run_BioSimulator(
*
*		-- Reaction functions
     $		List_Reg_Elements, Para_Reg_Elements,
     $		List_Reg_Modules, Para_Reg_Modules,
     $		num_Reg_Elements, num_Reg_Modules, 
     $		max_num_Reg_Elements, max_num_Reg_Modules, max_tmp, 
*      
*		-- Parameters for simulation
     $		time_interval, time_StateVariable, max_time_StateVariable,
     $		includeNoise, ix, Variance_System, 
*
*		-- Others
     $		num_Variable, max_num_Variable, 
*     		
*		-- Outputs
     $		StateVariable
*
     $	)
*
	implicit real*8 ( a-h, o-z )
*
	dimension StateVariable( max_time_StateVariable, max_num_Variable )
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension Variance_System( max_num_Variable )
	dimension a( max_num_Reg_Elements )
	integer*4 time_StateVariable
	integer*4 time
*
	do time = 2, time_StateVariable
*
	do i = 1, num_Variable
	StateVariable(time,i) = 0.0d0
	enddo
*
	do i = 1, num_Reg_Elements
*
	ID_Reg = List_Reg_Elements(i,2)
	ID_Reactant = List_Reg_Elements(i,3)
*
	if( ID_Reg. EQ. 1 )  
     $	a(i) = StateVariable(time-1,ID_Reactant) / time_interval
*
	if( ID_Reg. EQ. 2 )  
     $	a(i) = 1.0d0
*
	if( ID_Reg. GE. 3. AND. ID_Reg. LE. 4 ) 
     $	a(i) = StateVariable(time-1,ID_Reactant)
*
	if( ID_Reg. EQ. 18 ) a(i) = ( StateVariable(time-1,ID_Reactant) ** 
     $	Para_Reg_Elements(i,2) ) / 
     $	( StateVariable(time-1,ID_Reactant) ** Para_Reg_Elements(i,2) 
     $	+ Para_Reg_Elements(i,1) **  Para_Reg_Elements(i,2) ) 
*
	if( ID_Reg. EQ. 19 ) a(i) = 1.0d0 / 
     $	( 1.0d0 + ( StateVariable(time-1,ID_Reactant) / 
     $	Para_Reg_Elements(i,1) ) **  Para_Reg_Elements(i,2) ) 
*
	enddo
*
	do i = 1, num_Reg_Modules
	ID_Product = List_Reg_Modules(i,2)
	Rate_Constant = time_interval * Para_Reg_Modules(i,1) 
     $	* Para_Reg_Modules(i,2)
	num_Children = List_Reg_Modules(i,4)
	Reg_Value = Rate_Constant
	do j = 1, num_Children
	k = List_Reg_Modules(i,4+j)
	Reg_Value = Reg_Value * a(k)
	enddo
*
	StateVariable(time,ID_Product) = 
     $	StateVariable(time,ID_Product) + Reg_Value
	enddo
*
	if ( includeNoise .EQ. 1 ) then
	do i = 1, num_Variable
	StateVariable(time,i) = StateVariable(time,i) + 
     $	dsqrt( Variance_System(i) ) * rgauss( ix )
	enddo
	endif
*
	do i = 1, num_Variable
	if ( StateVariable(time,i) .LT. 0.0d0 ) 
*     $	StateVariable(time,i) = 0.0d0
     $	StateVariable(time,i) = StateVariable(time-1,i)
	enddo
*
	enddo
*	
*
	return
	end
*
*
	subroutine Run_BioSimulator_version3(
*
*		-- Reaction functions
     $		List_Reg_Elements, Para_Reg_Elements,
     $		List_Reg_Modules, Para_Reg_Modules,
     $		num_Reg_Elements, num_Reg_Modules, 
     $		max_num_Reg_Elements, max_num_Reg_Modules, max_tmp, 
*      
*		-- Parameters for simulation
     $		time_interval, time_StateVariable, max_time_StateVariable,
     $		includeNoise, ix, Variance_System, 
*
*		-- Others
     $		num_Variable, max_num_Variable, 
*     		
*		-- Outputs
     $		StateVariable
*
     $	)
*
	implicit real*8 ( a-h, o-z )
*
	dimension StateVariable( max_num_Variable, * )
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension Variance_System( max_num_Variable )
	dimension a( max_num_Reg_Elements )
	integer*4 time_StateVariable
	integer*4 time
*
	do time = 2, time_StateVariable
*
	do i = 1, num_Variable
	StateVariable(i,time) = 0.0d0
	enddo
*
	do i = 1, num_Reg_Elements
*
	ID_Reg = List_Reg_Elements(i,2)
	ID_Reactant = List_Reg_Elements(i,3)
*
	if( ID_Reg. EQ. 1 )  
     $	a(i) = StateVariable(ID_Reactant,time-1) / time_interval
*
	if( ID_Reg. EQ. 2 )  
     $	a(i) = 1.0d0
*
	if( ID_Reg. GE. 3. AND. ID_Reg. LE. 4 ) 
     $	a(i) = StateVariable(ID_Reactant,time-1)
*
	if( ID_Reg. EQ. 18 ) a(i) = ( StateVariable(ID_Reactant,time-1) ** 
     $	Para_Reg_Elements(i,2) ) / 
     $	( StateVariable(ID_Reactant,time-1) ** Para_Reg_Elements(i,2) 
     $	+ Para_Reg_Elements(i,1) **  Para_Reg_Elements(i,2) ) 
*
	if( ID_Reg. EQ. 19 ) a(i) = 1.0d0 / 
     $	( 1.0d0 + ( StateVariable(ID_Reactant,time-1) / 
     $	Para_Reg_Elements(i,1) ) **  Para_Reg_Elements(i,2) ) 
*
	enddo
*
	do i = 1, num_Reg_Modules
	ID_Product = List_Reg_Modules(i,2)
	Rate_Constant = time_interval * Para_Reg_Modules(i,1) 
     $	* Para_Reg_Modules(i,2)
	num_Children = List_Reg_Modules(i,4)
	Reg_Value = Rate_Constant
	do j = 1, num_Children
	k = List_Reg_Modules(i,4+j)
	Reg_Value = Reg_Value * a(k)
	enddo
*
	StateVariable(ID_Product,time) = 
     $	StateVariable(ID_Product,time) + Reg_Value
	enddo
*
	if ( includeNoise .EQ. 1 ) then
	do i = 1, num_Variable
	StateVariable(i,time) = StateVariable(i,time) + 
     $	dsqrt( Variance_System(i) ) * rgauss( ix )
	enddo
	endif
*
	do i = 1, num_Variable
	if ( StateVariable(i,time) .LT. 0.0d0 ) 
*     $	StateVariable(time,i) = 0.0d0
     $	StateVariable(i,time) = StateVariable(i,time-1)
	enddo
*
	enddo
*	
*
	return
	end
*
*
!	subroutine Run_BioSimulator_version2(
!*
!*		-- Reaction functions
!     $		List_Reg_Elements, Para_Reg_Elements,
!     $		List_Reg_Modules, Para_Reg_Modules,
!     $		num_Reg_Elements, num_Reg_Modules, 
!     $		max_num_Reg_Elements, max_num_Reg_Modules, 
!     $      max_tmp, 
!*      
!*		-- Parameters for simulation
!     $		time_interval, time_StateVariable, max_time_StateVariable,
!     $		includeNoise, ix, Variance_System, 
!*
!*		-- Others
!     $		num_Variable, max_num_Variable, 
!*     		
!*		-- Outputs
!     $		StateVariable, map_e2m, nn
!*
!     $	)
!*
!	implicit real*8 ( a-h, o-z )
!*
!	dimension StateVariable( max_num_Variable, * )
!	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
!	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
!	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
!	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
!	dimension Variance_System( max_num_Variable )
!	dimension a(max_num_Reg_Modules)
!	integer*4 time_StateVariable
!	integer*4 time
!	dimension map_e2m(2,*)
!*
!	do time = 2, time_StateVariable
!*
!	do i = 1, num_Reg_Modules
!	a(i) = time_interval * Para_Reg_Modules(i,1) 
!     $	* Para_Reg_Modules(i,2)
!	enddo
!*
!	do i = 1, num_Variable
!	StateVariable(i,time) = 0.0d0
!	enddo
!*
!	do i = 1, nn
!      ie = map_e2m(1,i)
!      im = map_e2m(2,i)
!	ID_Reg = List_Reg_Elements(ie,2)
!	ID_Reactant = List_Reg_Elements(ie,3)
!*
!      ! Differentiation over time 
!	if( ID_Reg. EQ. 1 )
!     $	a(im) = a(im) * StateVariable(ID_Reactant,time-1)/time_interval
!*
!*	if( ID_Reg. EQ. 2 )
!*     $	a(im) = a(im) * Rate_Constant
!*
!	if( ID_Reg. GE. 3. AND. ID_Reg. LE. 4 ) 
!     $	a(im) = a(im) * StateVariable(ID_Reactant,time-1)
!*
!	if( ID_Reg. EQ. 18 ) 
!     $	a(im) = a(im)  /
!     $  ( 1.0d0 + 
!     $  ( Para_Reg_Elements(ie,1) / StateVariable(ID_Reactant,time-1) 
!     $  ) ** Para_Reg_Elements(ie,2) ) 
!*
!	if( ID_Reg. EQ. 19 ) 
!     $	a(im) = a(im) / 
!     $	( 1.0d0 + ( StateVariable(ID_Reactant,time-1) / 
!     $	Para_Reg_Elements(ie,1) ) **  Para_Reg_Elements(ie,2) ) 	
!      enddo
!*
!	do i = 1, num_Reg_Modules
!      ID_Product = List_Reg_Modules(i,2)
!	StateVariable(ID_Product,time) = 
!     $	StateVariable(ID_Product,time) + a(i)
!	enddo
!*
!	if ( includeNoise .EQ. 1 ) then
!	do i = 1, num_Variable
!	StateVariable(i,time) = StateVariable(i,time) + 
!     $	dsqrt( Variance_System(i) ) * rgauss( ix )
!	enddo
!	endif
!*
!	do i = 1, num_Variable
!	if ( StateVariable(i,time) .le. 0.0d0 ) 
!     $	StateVariable(i,time) = StateVariable(i,time-1)
!	enddo
!*
!	enddo
!*	
!	return
!	end
*
*
	subroutine Get_NoiseVariances(
*
*		-- Variances
     $		Variance_System, Variance_Measure, num_Variable,
     $		max_num_Variable,
*
*		-- Others
     $		suffix
*
     $	)
*
	implicit real*8 ( a-h, o-z )
	integer*2 tab /2313/
	parameter ( max_read = 2000 )
	parameter ( iprint = 0 )
*
	character  S_ObsVariance * ( * )
	character  E_ObsVariance * ( * )
	parameter  ( S_ObsVariance = '<begin:Variance_Noise_Observed>' ) 
	parameter  ( E_ObsVariance = '<end:Variance_Noise_Observed>' ) 
*
	character  S_SysVariance * ( * )
	character  E_SysVariance * ( * )
	parameter  ( S_SysVariance = '<begin:Variance_Noise_System>' ) 
	parameter  ( E_SysVariance = '<end:Variance_Noise_System>' ) 
*
	dimension Variance_Measure( max_num_Variable )
	dimension Variance_System( max_num_Variable )
	character suffix * 30
	character str(1000) * 50
	character Read_File * 30
	Read_File = trim( suffix )//'.mdl'
*
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_ObsVariance ) exit
	enddo
*
	do i = 1, num_Variable
	read( 5, * ) str(1), str(2)
	if( trim( str(1) ). EQ. E_ObsVariance ) exit
	read( str(1), * ) Variance_Measure(i)
	enddo
	close( 5 )
*
	if( iprint. EQ. 1) then
	write( *, * )
	write( *, '(a)' ) '<Variance-Noise-Observed>'
	write( *, '(100(a1,f7.3))' ) 
     $	( ( tab, Variance_Measure(i) ), i = 1, num_Variable )
	endif
*
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_SysVariance ) exit
	enddo
*
	do i = 1, num_Variable
	read( 5, * ) str(1), str(2)
	if( trim( str(1) ). EQ. E_SysVariance ) exit
	read( str(1), * ) Variance_System(i)
	enddo
	close( 5 )
*
	if( iprint. EQ. 1) then
	write( *, * )
	write( *, '(a)' ) '<Variance-Noise-System>'
	write( *, '(100(a1,f7.3))' ) 
     $	( ( tab, Variance_System(i) ), i = 1, num_Variable )
	endif
*
	return
	end
*		
*
	subroutine Get_Data(
*
*		-- Observed time course data
     $		Data, time_Data, num_Variable, max_time_Data, 
     $		max_num_Variable, 
*
*		-- Indicators for observed time points and missing variables
     $		ID_Time_Data, ID_MissVariable, 
*
*		-- Others
     $		suffix, Name_Variable 
     $	)
*
	implicit real*8( a-h, o-z )
*
	integer*2 tab /2313/
	integer*4 time_Data
	parameter ( max_read = 100000 )
	parameter ( iprint = 0 )
*		 
	dimension Data( max_num_Variable, max_time_Data )
	dimension ID_Time_Data( max_time_Data )
	dimension ID_MissVariable( max_num_Variable ) 
	character suffix * 30 
	character str(1000) * 50
*
	character  BEGIN_obsVAR * ( * ), END_obsVAR * ( * )
	parameter  ( BEGIN_obsVAR = '<begin:obs_var>' ) 
	parameter  ( END_obsVAR = '<end:obs_var>' ) 
*
	character  BEGIN_obsTIME * ( * ), END_obsTIME * ( * )
	parameter  ( BEGIN_obsTIME = '<begin:obs_time>' ) 
	parameter  ( END_obsTIME = '<end:obs_time>' )
*
	character  BEGIN_DATA * ( * ), END_DATA * ( * )
	parameter  ( BEGIN_DATA = '<begin:data>' ) 
	parameter  ( END_DATA = '<end:data>' ) 
*
	character  Name_Variable( max_num_Variable ) * 30 
	character  Read_File * 30
	character  Out_File_Data * 30
*
	Read_File = trim( suffix )//'.dat'
	Out_File_Data = trim( suffix )//'_data.log'
*
	do i = 1, num_Variable
	ID_MissVariable(i) = 0 
	enddo
*
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. BEGIN_obsVAR ) exit
	enddo
*
	do i = 1, max_read
	read( 5, * ) str(1), str(2)
	if( trim( str(1) ). EQ. END_obsVAR ) exit
	read( str(2), '(i)' ) k
	ID_MissVariable(k) = 1
	enddo
      num_ObsVariable = i - 1
*
	close( 5 )
*
*
	if( iprint. EQ. 1) then
	write( *, * )
	write( *, '(a)' ) '<Missing-Variable>'
	write( *, '(100(a1,i3))' ) 
     $	( ( tab, ID_MissVariable(i) ), i = 1, num_Variable )
	endif
*
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. BEGIN_obsTIME ) exit
	enddo
*
      do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ .END_obsTIME ) exit
	read( str(1),'(i)' ) ID_Time_Data(i)
	enddo
      time_Data = i - 1
	close( 5 )
*
*
	if( iprint. EQ. 1) then
	write( *, * )
	write( *, '(a)' ) '<Observed-Time-Points>'
	write( *, '(100(a1,i3))' ) 
     $	( ( tab, ID_Time_Data(i) ), i = 1, time_Data )
	endif
*
*
	do i = 1, num_Variable
	do j = 1, time_Data
	Data(i,j) = 0.0d0 
	enddo
	enddo
*
*
	itmp = time_Data + 2
*	
	open( 5, file = Read_FILE )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. BEGIN_DATA ) exit
	enddo
*
      do i = 1, num_ObsVariable
	read( 5, * ) ( str(k), k = 1, itmp )
	if( trim( str(1) ). EQ. END_DATA ) exit
	read( str(2), '(i)' ) k
	do j = 1, time_Data
	read( str(j+2), * ) Data(k,j)
	enddo
	enddo
*
	close( 5 )
*
*
	if( iprint. EQ. 1) then
	write( *, * )
	write( *, '(a)' ) '<Data>'
	do i = 1, num_Variable
	k = ID_MissVariable(i)
	if ( k .EQ. 1 ) 
     $	write( *, '( 100( a1, f7.3 ) )' ) 
     $	( ( tab, Data(i,j) ), j = 1, time_Data )
	enddo
	endif
*
*
	open( 5, file = Out_File_Data )
*
	write( 5, '( a, 100( a1,i3 ) )' ) 'ObservedTime',  
     $	( ( tab, ID_Time_Data(i) ), i = 1, time_Data )
*
	do i = 1, num_Variable
	k = ID_MissVariable(i)
	if ( k .EQ. 1 ) 
     $	write( 5, '( a, 100( a1, f7.3 ) )' ) Name_Variable(i),  
     $	( ( tab, Data(i,j) ), j = 1, time_Data )
	enddo
*
	close( 5 )
*
*
	return
	end
*
*
	subroutine Get_Reactions (
*	 
*		-- Reaction functions 
     $		List_Reg_Elements, Para_Reg_Elements,
     $		List_Reg_Modules, Para_Reg_Modules,
     $		num_Reg_Elements, num_Reg_Modules, 
     $		max_num_Reg_Elements, max_num_Reg_Modules, max_tmp, 
*
*		-- Variables
     $		ID_Variable, Name_Variable, num_Variable, 
     $		max_num_Variable,
*      
*		-- Parameters for simulation
     $		time_interval, time_StateVariable, 
*
*		-- Prior distributions        
     $		Prior_ParaElements, ID_Prior_ParaElements, 
     $		Prior_ParaModules, ID_Prior_ParaModules,
     $		Prior_Inits, ID_Prior_Inits, 
     $		num_Para1, num_Para2, max_num_Para1, max_num_Para2,
*
*		-- Others
     $		suffix, Name_Model,
     $		strElements, strModules, strParaElements, strParaModules
*
     $	)
*
	implicit real*8 (a-h,o-z)
*
	parameter ( max_read = 2000 )
	parameter ( iprint = 0 )
*
	character S_NameModel * (*)
	character E_NameModel * (*)
	parameter ( S_NameModel = '<begin:Model_Reference>' ) 
	parameter ( E_NameModel = '<end:Model_Reference>' ) 
*
	character S_Variable * (*)
	character E_Variable * (*)
	parameter ( S_Variable = '<begin:List_Species>' ) 
	parameter ( E_Variable = '<end:List_Species>' ) 
*
	character S_ReactionE * (*)
	character E_ReactionE * (*)
	parameter ( S_ReactionE = '<begin:List_ReactionElements>' ) 
	parameter ( E_ReactionE = '<end:List_ReactionElements>' )
*
	character S_ReactionM * (*)
	character E_ReactionM * (*)
	parameter ( S_ReactionM = '<begin:List_ReactionModules>' ) 
	parameter ( E_ReactionM = '<end:List_ReactionModules>' )
*
	character S_TimeInterval * (*)
	character E_TimeInterval * (*)
	parameter ( S_TimeInterval = '<begin:Time_Interval>' ) 
	parameter ( E_TimeInterval = '<end:Time_Interval>' )
*
	character S_TimeSim * (*)
	character E_TimeSim * (*)
	parameter ( S_TimeSim = '<begin:Time_Simulation>' ) 
	parameter ( E_TimeSim = '<end:Time_Simulation>' )
*
	character S_InitValue * (*)
	character E_InitValue * (*)
	parameter ( S_InitValue = '<begin:List_Initial_Values>' ) 
	parameter ( E_InitValue = '<end:List_Initial_Values>' ) 
*
	character S_ParaModule * (*)
	character E_ParaModule * (*)
	parameter ( S_ParaModule = '<begin:List_ParametersInModules>' ) 
	parameter ( E_ParaModule = '<end:List_ParametersInModules>' )
*
	character S_ParaElement * (*)
	character E_ParaElement * (*)
	parameter ( S_ParaElement = '<begin:List_ParametersInElements>') 
	parameter ( E_ParaElement = '<end:List_ParametersInElements>' )
*
	integer*2 tab /2313/
	integer*4 time_StateVariable
*
	character str( 100 ) * 50
	character labels( 100 ) * 50
*
	character suffix * 30
	character Name_Variable( max_num_Variable ) * 30
	dimension ID_Variable( max_num_Variable )
	character Name_Model * 30
*
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
*
	dimension Prior_ParaElements( max_num_Para1, 4 )
	dimension Prior_ParaModules( max_num_Para2, 2 )
	dimension Prior_Inits( max_num_Variable, 2 )
	dimension ID_Prior_ParaElements( max_num_Para1 )
	dimension ID_Prior_ParaModules( max_num_Para2 )
	dimension ID_Prior_Inits( max_num_Variable )
*
	character strElements( max_num_Reg_Elements, 4 ) * 30
	character strModules( max_num_Reg_Modules, 3 ) * 30
	character strParaElements( max_num_Para1, 4  )  * 30
	character strParaModules( max_num_Para2 ) * 50
*
	character  Read_File * 50
	Read_File = trim( suffix )//'.mdl'
*
*
* --- Get reference name of model 
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_NameModel ) exit
	enddo
	read( 5, * ) Name_Model
*
	close(5)
*
*
	if( iprint. EQ. 1 ) then 
	write( *, * )
	write( *, '(a)' ) '<Reference-Name-Of-Model>'
	write( * , '(a)' ) trim( Name_Model )
	endif
*
*
* --- Get list of variables, reference ID and names of variables
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ) .EQ. S_Variable ) exit
	enddo
*
	do i = 1, max_read
	read( 5, * ) str(1), str(2)
	if( trim( str(1) ) .EQ. E_Variable ) exit
	read( str(1), '(i)' ) ID_Variable(i)	
	Name_Variable(i) = str(2)
	enddo
*
	num_Variable = i-1
*
	close( 5 )
*
*
	if( iprint. EQ. 1) then
	write( *, * )
	write( *, '(a)' ) '<List-Of-Species>'
	write( *, '(100(a1,i3))' ) 
     $	( ( tab, ID_Variable(i) ), i = 1, num_Variable )
	write( *, '(100(a1,a))' ) 
     $	( ( tab, trim( Name_Variable(i) ) ), i = 1, num_Variable )
	endif
*
*
* --- Get list of reaction elements
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_ReactionE ) exit
	enddo
*
	read( 5, * ) ( labels(k), k = 1, 11 ) 
*
      do i = 1, max_read
    	read( 5, * ) ( str(k), k = 1, 11 )
	if( str(1). EQ. E_ReactionE ) exit
	read( str(1), '(i)' ) List_Reg_Elements(i,1)	! Reaction IDs
	read( str(5), '(i)' ) List_Reg_Elements(i,2)	! Type of reaction equation
	read( str(7), '(i)' ) List_Reg_Elements(i,3)	! Reactant IDs
	read( str(8:11), * ) Para_Reg_Elements(i,1:4)	! Kinetic parameters in each reaction equation
*
	strElements(i,1) = str(2)	! Product name
	strElements(i,2) = str(3)	! Product ID
	strElements(i,3) = str(4)	! Kinetic law name
	strElements(i,4) = str(6)
	enddo
*
	num_Reg_Elements = i - 1
*
	close(5)
*
	if ( iprint .EQ. 1 ) then
	write( *, * )
	write( *, '(a)' ) '<Elementary-Functions-In-Reaction-Model>'
	do i = 1, num_Reg_Elements
	write( *, '(i3,a1,i3,a1,i3,100(a1,f7.3))' ) 
     $	List_Reg_Elements(i,1), tab, List_Reg_Elements(i,2),
     $	tab, List_Reg_Elements(i,3), 
     $	( ( tab, Para_Reg_Elements(i,k) ), k = 1, 4 )
	enddo
	endif
*
*
* --- Get modules of individual reaction equations
*
	do i = 1, max_num_Reg_Modules
	do j = 1, max_tmp
	List_Reg_Modules(i,j) = 0
	enddo
	enddo
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_ReactionM ) exit
	enddo
*
	read( 5, * ) ( labels(k), k = 1, 11 ) 
*
      do i = 1, max_read
    	read( 5, * ) ( str(k), k = 1, 9 )
	if( str(1). EQ. E_ReactionM ) exit
	read( str(1), '(i)' ) List_Reg_Modules(i,1)		! Module ID
	read( str(3), '(i)' ) List_Reg_Modules(i,2)		! Product species ID
	read( str(6), '(i)' ) List_Reg_Modules(i,3)		! Reference IDs for rate constant parameters
	read( str(9), '(i)' ) List_Reg_Modules(i,4)		! Number of reactants species
	read( str(7:8), * ) Para_Reg_Modules(i,1:2)		! Rate constant parameters	
*
	strModules(i,1) = str(2)	! Product name
	strModules(i,2) = str(4)	! Kinetic law
	strModules(i,3) = str(5)	! Reaction partner
	enddo
*
	num_Reg_Modules = i - 1
*
	close(5)
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_ReactionM ) exit
	enddo
*
	read( 5, * ) 
*
      do i = 1, num_Reg_Modules
	n = List_Reg_Modules(i,4)
    	read( 5, * ) ( str(k), k = 1, 9 ), 
     $	( List_Reg_Modules(i,4 + k), k = 1, n )
	if( str(1). EQ. E_ReactionM ) exit
	enddo
*
	close(5)
*
	if ( iprint .EQ. 1) then
	write( *, * )
	write( *, '(a)' ) '<Modules-Of-Elementary-Reaction-Functions>' 
	do i = 1, num_Reg_Modules
	write( *, '(7(a1,i3),100(a1,f7.3))' ) 
     $	( ( tab, List_Reg_Modules(i,k) ), k = 1, 7 ), 
     $	( ( tab, Para_Reg_Modules(i,k) ), k = 1, 2 )
	enddo
	endif
*
*
* --- Read time interval
*
	open( 5, file = Read_File )
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_TimeInterval ) exit
	enddo
*
	read( 5, * ) time_interval
*
	close( 5 )
*
	if ( iprint .EQ. 1) then
	write( *, * ) 
	write( *, '(a)' ) '<Time-Interval>'
	write( *, '(f7.5)' ) time_interval 
	endif
*
*
* --- Read number of time points for simulation
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_TimeSim ) exit
	enddo
*
	read( 5, * )
	read( 5, * ) str(1), time_StateVariable
*
	close( 5 )
*
	if ( iprint .EQ. 1) then
	write( *, * ) 
	write( *, '(a)' ) '<Number-Of-Time-Points-Over-Simulation>'
	write( *, '(i)' ) time_StateVariable
	endif
*
*
* --- Read prior distribution over initial values
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_InitValue ) exit
	enddo
*
	read( 5, * )
	do i = 1, max_read
	read( 5, * ) ( str(k), k = 1, 5 )
	if( trim( str(1) ). EQ. E_InitValue ) exit
	read( str(4:5), * ) Prior_Inits(i,1:2)
	read( str(2), '(i)' ) ID_Prior_Inits(i)
	enddo
*
	close( 5 )
*
	if( iprint .EQ. 1 ) then
	write( *, * )
	write( *, '(a)' ) '<Prior-Over-Initial-Values>'
	do i = 1, num_Variable
	write( *, '(i3, 100(a1,f7.3))' ) 
     *	ID_Prior_Inits(i), ( ( tab, Prior_Inits(i,k) ), k = 1, 2 )
	enddo
	endif
*
*
* --- Read prior distribution over kientic parameters in Reaction Modules
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_ParaModule ) exit
	enddo
*
  	read( 5, * )
	do i = 1, max_read
	read( 5, * ) ( str(k), k = 1,5 )
	if( trim( str(1) ). EQ. E_ParaModule ) exit
	read( str(1), '( i )' ) ID_Prior_ParaModules(i)
	read( str(2:3), * ) Prior_ParaModules(i,1:2)
*
	strParaModules(i) = str(5)
	enddo
*
	num_Para2 = i - 1
*
	close( 5 )
*
	if( iprint .EQ. 1 ) then
	write( *, * )
	write( *, '(a)' ) '<Prior-Over-Rate-Constant-Parameters>'
	do i = 1, num_Para2
	write( *, '(i3, 100(a1,f7.3))' ) 
     $	ID_Prior_ParaModules(i), 
     $	( ( tab, Prior_ParaModules(i,k) ), k = 1, 2 )
	enddo
	endif
*
*
* --- Read prior distribution over kientic parameters
*
	open( 5, file = Read_File )
*
	do i = 1, max_read
	read( 5, * ) str(1)
	if( trim( str(1) ). EQ. S_ParaElement ) exit
	enddo
*
  	read( 5, * )
	do i = 1, max_read
	read( 5, * ) ( str(k), k = 1, 11 )
	if( trim( str(1) ). EQ. E_ParaElement ) exit
	read( str(1), '( i )' ) ID_Prior_ParaElements(i)
	read( str(8:11), * ) Prior_ParaElements(i,1:4)
*
	strParaElements(i,1) = str(2)	! Product name
	strParaElements(i,2) = str(3)	! Kinetic law
	strParaElements(i,3) = str(4)	! Kinatic law
	strParaElements(i,4) = str(5)	! Number of relevant parameters
	enddo
*
	num_Para1 = i - 1
*
	close( 5 )
*
	if( iprint .EQ. 1 ) then
	write( *, * )
	write( *, '(a)' ) '<Prior-Over-Parameters-In-Elements>'
	do i = 1, num_Para1
	write( *, '(i3, 100(a1,f7.3))' ) 
     $	ID_Prior_ParaElements(i), 
     $	( ( tab, Prior_ParaElements(i,k) ), k = 1, 4 )
	enddo
	endif
*
*
	return
	end
*
*
*
	subroutine Output_Reactions (
*	 
*		-- Reaction functions 
     $		List_Reg_Elements, Para_Reg_Elements,
     $		List_Reg_Modules, Para_Reg_Modules,
     $		num_Reg_Elements, num_Reg_Modules, 
     $		max_num_Reg_Elements, max_num_Reg_Modules, max_tmp, 
*
*		-- Variables
     $		ID_Variable, Name_Variable, num_Variable, 
     $		max_num_Variable,
*      
*		-- Parameters for simulation
     $		time_interval, time_StateVariable, 
*
*		-- Prior distributions        
     $		Prior_ParaElements, ID_Prior_ParaElements, 
     $		Prior_ParaModules, ID_Prior_ParaModules,
     $		Prior_Inits, ID_Prior_Inits, 
     $		num_Para1, num_Para2, max_num_Para1, max_num_Para2,
*
*		-- Others
     $		suffix, Name_Model, Output_File,
     $		StateVariable, max_time_StateVariable,
*
*		-- Variances
     $		Variance_System, Variance_Measure,
     $		strElements, strModules, strParaElements, strParaModules
*
     $	)
*
	implicit real*8 (a-h,o-z)
*
	parameter ( max_read = 2000 )
	parameter ( iprint = 1 )
*
	character S_NameModel * (*)
	character E_NameModel * (*)
	parameter ( S_NameModel = '<begin:Model_Reference>' ) 
	parameter ( E_NameModel = '<end:Model_Reference>' ) 
*
	character S_Variable * (*)
	character E_Variable * (*)
	parameter ( S_Variable = '<begin:List_Species>' ) 
	parameter ( E_Variable = '<end:List_Species>' ) 
*
	character S_ReactionE * (*)
	character E_ReactionE * (*)
	parameter ( S_ReactionE = '<begin:List_ReactionElements>' ) 
	parameter ( E_ReactionE = '<end:List_ReactionElements>' )
*
	character S_ReactionM * (*)
	character E_ReactionM * (*)
	parameter ( S_ReactionM = '<begin:List_ReactionModules>' ) 
	parameter ( E_ReactionM = '<end:List_ReactionModules>' )
*
	character S_TimeInterval * (*)
	character E_TimeInterval * (*)
	parameter ( S_TimeInterval = '<begin:Time_Interval>' ) 
	parameter ( E_TimeInterval = '<end:Time_Interval>' )
*
	character S_TimeSim * (*)
	character E_TimeSim * (*)
	parameter ( S_TimeSim = '<begin:Time_Simulation>' ) 
	parameter ( E_TimeSim = '<end:Time_Simulation>' )
*
	character S_InitValue * (*)
	character E_InitValue * (*)
	parameter ( S_InitValue = '<begin:List_Initial_Values>' ) 
	parameter ( E_InitValue = '<end:List_Initial_Values>' ) 
*
	character S_ParaModule * (*)
	character E_ParaModule * (*)
	parameter ( S_ParaModule = '<begin:List_ParametersInModules>' ) 
	parameter ( E_ParaModule = '<end:List_ParametersInModules>' )
*
	character S_ParaElement * (*)
	character E_ParaElement * (*)
	parameter ( S_ParaElement = '<begin:List_ParametersInElements>') 
	parameter ( E_ParaElement = '<end:List_ParametersInElements>' )
*
	character  S_ObsVariance * ( * )
	character  E_ObsVariance * ( * )
	parameter  ( S_ObsVariance = '<begin:Variance_Noise_Observed>' ) 
	parameter  ( E_ObsVariance = '<end:Variance_Noise_Observed>' ) 
*
	character  S_SysVariance * ( * )
	character  E_SysVariance * ( * )
	parameter  ( S_SysVariance = '<begin:Variance_Noise_System>' ) 
	parameter  ( E_SysVariance = '<end:Variance_Noise_System>' ) 
*
	integer*2 tab /2313/
	integer*4 time_StateVariable
*
	character str( 100 ) * 50
	character labels( 100 ) * 50
*
	character suffix * 30
	character Name_Variable( max_num_Variable ) * 30
	dimension ID_Variable( max_num_Variable )
	character Name_Model * 30
	character suppParam( 10 ) * 30
*
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
*
	dimension Prior_ParaElements( max_num_Para1, 4 )
	dimension Prior_ParaModules( max_num_Para2, 2 )
	dimension Prior_Inits( max_num_Variable, 2 )
	dimension ID_Prior_ParaElements( max_num_Para1 )
	dimension ID_Prior_ParaModules( max_num_Para2 )
	dimension ID_Prior_Inits( max_num_Variable )
*
	dimension StateVariable( max_num_Variable, * )
	dimension Variance_Measure( max_num_Variable )
	dimension Variance_System( max_num_Variable )
*
	character strElements( max_num_Reg_Elements, 4 ) * 30
	character strModules( max_num_Reg_Modules, 3 ) * 30
	character strParaElements( max_num_Para1, 4  )  * 30
	character strParaModules( max_num_Para2 ) * 50
*
	character  Output_File * ( * )
	Output_File = trim( Output_File )
*
	do i = 1, 4
	suppParam(i) = 'SupplParameter'
	enddo
*
	open( 5, file = Output_File )
*
	write( 5, * ) 'LiSDAS: format version = 1.0'
	write( 5, * )
	write( 5, * ) S_NameModel
 	write( 5, '(a1,a)' ) tab, Name_Model
	write( 5, * ) E_NameModel
	write( 5, * )
	write( 5, * ) S_Variable
	do i = 1, num_Variable
	write( 5, '(a1,i,a1,a)' ) tab, ID_Variable(i), tab,  
     $	trim(Name_Variable(i))
	enddo
	write( 5, * ) E_Variable
	write( 5, * )
	write( 5, * ) S_TimeInterval
	write( 5, '(a1,f)' ) tab, time_interval
	write( 5, * ) E_TimeInterval
	write( 5, * )
	write( 5, * ) S_TimeSim
	write( 5, '(a1,a,a1,i)' ) tab, 'start', tab, 1 
	write( 5, '(a1,a,a1,i)' ) tab, 'end', tab, time_StateVariable
	write( 5, * ) E_TimeSim
	write( 5, * )
	write( 5, * ) S_ReactionE
	write( 5,'( 1000(a1,a) )')
     $	tab, 'Process', tab, 'Product_Name', tab, 'Product_ID', tab, 
     $	'KineticLaw', tab, 'KineticLawID', tab, 'Reactant_Name', tab,
     $	'Reactant_ID', ( ( tab, suppParam(j) ), j =1,4 )	
 	do i = 1, num_Reg_Elements
	write( 5, '(a1,i,a1,a,a1,a,a1,a,a1,i,a1,a,a1,i,100(a1,f7.4) )' ) 
     $	tab, List_Reg_Elements(i,1), tab, trim(strElements(i,1)), tab, 
     $	trim(strElements(i,2)), tab, trim(strElements(i,3)), tab, 
     $	List_Reg_Elements(i,2), tab, 
     $	trim(strElements(i,4)), tab, List_Reg_Elements(i,3), 
     $	( ( tab, Para_Reg_Elements(i,j) ), j = 1,4 )
	enddo
	write( 5, * ) E_ReactionE
	write( 5, * )
	write( 5, * ) S_ReactionM
	write( 5,'( 1000(a1,a) )')
     $	tab, 'Module_ID', tab, 'Product_Name', tab, 'Product_ID', tab,
     $	'KineticLaw', tab, 'Reactant_Name', tab,'RateParameterID',tab, 
     $	'RateParameter', tab, 'SignRateParameter',tab, 'Num_Children',
     $	tab, 'Child_Process', tab, 'Child_Process'
 	do i = 1, num_Reg_Modules
	n = List_Reg_Modules(i,4)
	write( 5, '(a1,i,a1,a,a1,i,a1,a,a1,a,a1,i,2(a1,f7.4),100(a1,i) )' ) 
     $	tab, List_Reg_Modules(i,1), tab, trim(strModules(i,1)), tab, 
     $	List_Reg_Modules(i,2), tab, trim(strModules(i,2)), tab, 
     $	trim(strModules(i,3)), tab,
     $	List_Reg_Modules(i,3), 
     $	( ( tab, Para_Reg_Modules(i,j) ), j = 1,2 ), tab,
     $	List_Reg_Modules(i,4), 
     $	( ( tab, List_Reg_Modules(i,4+j) ), j = 1,n )
	enddo
	write( 5, * ) E_ReactionM
*
	write( 5, * ) 
	write( 5, * ) S_InitValue
	write( 5,'( 1000(a1,a) )' )
     $	tab, 'Species_Name', tab, 'Species_ID', tab, 'Value', tab,	
     $	'Prior_Mean', tab, 'Prior_Variance'
	do i = 1, num_Variable
	write( 5, '(a1,a,a1,i,100(a1,f7.4))' ) 
     $	tab, trim(Name_Variable(i)), tab, 
     $	ID_Variable(i), tab, StateVariable(i,1), 
     $	( ( tab, Prior_Inits(i,k) ), k = 1, 2 )
 	enddo
	write( 5, * ) E_InitValue
*
	write( 5, * ) 
	write( 5, * ) S_ParaElement
	write( 5,'( 1000(a1,a) )' )
     $	tab, 'Process', tab, 'Product', tab, 'KineticLaw', tab, 
     $	'KineticLaw', tab, 'NumberOfParameters', tab, 'Parameter',
     $	tab, 'Parameter', tab, 'PriorMean', tab, 'PriorMean', 
     $	tab, 'PriorVariance',  tab, 'PriorVariance'
	do i = 1, num_Para1
	n = ID_Prior_ParaElements(i)
	write( 5,'( 1000(a1,i,a1,a,a1,a,a1,a,a1,a, 100(a1,f7.4)) )' )
     $	tab, ID_Prior_ParaElements(i), 
     $	( tab, trim(strParaElements(i,k) ), k = 1,4 ), 
     $	( ( tab, Para_Reg_Elements(n,j) ), j=1,2 ),
     $	( ( tab, Prior_ParaElements(i,k) ), k = 1, 4 )
	enddo
	write( 5, * ) E_ParaElement
*
	write( 5, * ) 
	write( 5, * ) S_ParaModule
	write( 5,'( 1000(a1,a) )' )
     $	tab, 'kinetic_module_id', tab, 'prior_mean', tab, 
     $	'prior_variance', tab, 'scan_prob', tab, 'comments', 
     $	tab, 'estimate'
	do i = 1, num_Para2
	n = ID_Prior_ParaModules(i)
	do m = 1, num_Reg_Modules
	if( n. EQ. List_Reg_Modules(m,3) ) mm = m
	enddo
	write( 5,'( 1000(a1,i,2(a1,f7.4), a1, a, a1,a,a1,f7.4) )' )
     $	tab, ID_Prior_ParaModules(i),
     $	( ( tab, Prior_ParaModules(i,j) ), j=1,2 ),
     $	tab, 'scan_prob', tab, trim( strParaModules(i) ), 
     $	tab, Para_Reg_Modules(mm,1)
	enddo
	write( 5, * ) E_ParaModule
*
	write( 5, * ) 
	write( 5, * ) S_ObsVariance
	do i = 1, num_Variable
	write( 5,'( 1000(a1,f7.4,a1,i,100(a1,a) ) )' )
     $	tab, Variance_Measure(i), tab, i, tab, 'empty', tab, 'empty'
	enddo
	write( 5, * ) E_ObsVariance
*
	write( 5, * ) 
	write( 5, * ) S_SysVariance
	do i = 1, num_Variable
	write( 5,'( 1000(a1,f7.4,a1,i,100(a1,a) ) )' )
     $	tab, Variance_System(i), tab, i, tab, 'empty', tab, 'empty'
	enddo
	write( 5, * ) E_SysVariance
*
	write( 5, * )
*
	close(5)
*
	return
	end
*
*
	subroutine Output_BioSimulator( 
     $		StateVariable, num_Variable, time_StateVariable, 
     $		max_time_StateVariable, max_num_Variable, 
     $		Name_Variable,
     $		Output_File
     $	)
*
	implicit real*8 ( a-h, o-z )
	integer*2 tab /2313/
	integer*4 time_StateVariable
	dimension StateVariable( max_num_Variable, * ) 
	character Name_Variable( max_num_Variable ) * 30
	character(*) Output_File
*
	open( 5, file = Output_File, err=999)
	write( 5, '(a,1000(a1,a))') 'Time_VariableName', 
     $	( ( tab, trim( Name_Variable(i) ) ), i = 1, num_Variable )
	do i = 1, time_StateVariable
	write( 5, '(i,1000(a1,f8.4))') i, 
     $	( tab, StateVariable(j,i), j = 1, num_Variable )
	enddo
	close( 5 )
*
	return

 999  write(*,*) 'Output_BioSimulator: fail to open', trim(Output_File)
      stop
	end
*
*
	subroutine choles(x,mj,n,y,nj)
*
*	Inputs
*		x(i,j):	Symmetric matrix
*		n:		Dimension of x
*		mj:		Adjustable dimension of x
*		nj:		Adlustable dimension of y
*	Outputs:
*		y(i,j): Lower trianglular matrix x=y*y
*
	implicit real*8(a-h,o-z)
	dimension x(mj,mj),y(nj,nj)
	do 10 i=1,n
	do 10 j=1,n
   10 y(i,j)=0.0d0
	do 100 j=1,n
	sum1=x(j,j)
	do 20 k=1,j-1
   20 sum1=sum1-y(j,k)**2
	if(sum1.gt.0.0d0) y(j,j)=dsqrt(sum1)
	if(sum1.eq.0.0d0) y(j,j)=0.0d0
	do 40 i=j+1,n
	sum2=0.0d0
	do 30 k=1,j-1
   30 sum2=sum2+y(i,k)*y(j,k)
   40 y(i,j)=(x(i,j)-sum2)/y(j,j)
  100 continue
	return
	end
*
*
*
      double precision function rgauss( ix )
*
*	Gaussian random number generator (marsaglia's algorithm)
*
      implicit real*8( a-h, o-z )
*      data icount /0/
      if( ix. eq. 0 ) ix = 1990103011
*      if( icount. eq. 0 ) then
   10 u1 = runif( ix )
      u2 = runif( ix )
      v1 = 2.0 * u1 - 1.0
      v2= 2.0 * u2 - 1.0
      s = v1 ** 2 + v2 ** 2
      if( s. ge. 1.0 ) goto 10
      rgauss = v1 * dsqrt( -2.0 * dlog( s ) / s )
*      icount = 1
*      else
*      rgauss = v2 * dsqrt( -2.0 * dlog( s ) / s )
*      icount = 0
*      endif
      return
      end
*
*
      function runi(idum)
	integer idum,im1,im2,imm1,ia1,ia2,iq1,iq2,ir1,ir2,ntab,ndiv
	real*8 runi,am,eps,rnmx
	parameter(im1=2147483563,im2=2147483399,am=1./im1,imm1=im1-1,
     *  ia1=40014,ia2=40692,iq1=53668,iq2=52774,ir1=12211,
     *  ir2=3791,ntab=32,ndiv=1+imm1/ntab,eps=1.2e-7,rnmx=1.-eps)
*
      integer idum2,j,k,iv(ntab),iy
	save iv,iy,idum2
	data idum2/123456789/, iv/ntab*0/, iy/0/
	if(idum.le.0) then
	idum=max(-idum,1)
	idum2=idum
	do 11 j=ntab+8,1,-1
	k=idum/iq1
	idum=ia1*(idum-k*iq1)-k*ir1
	if(idum.lt.0) idum=idum+im1
	if(j.le.ntab) iv(j)=idum
   11 continue
      iy=iv(1)
      endif
	k=idum/iq1
	idum=ia1*(idum-k*iq1)-k*ir1
	if(idum.lt.0) idum=idum+im1
	k=idum2/iq2
	idum2=ia2*(idum2-k*iq2)-k*ir2
	if(idum2.lt.0) idum2=idum2+im2
	j=1+iy/ndiv
	iy=iv(j)-idum2
	iv(j)=idum
	if(iy.lt.1) iy=iy+imm1
	runi=min(am*iy,rnmx)
	return
	end
*
*
	double precision function runif( ix )
*
*	Uniform random number generator 
*
	real*8 a, c, em
	data em/1664501d0/, a/1229d0/, c/351750d0/
	ix = mod( ix * a + c, em )
	runif = ix / em
	return
	end
*
*
	subroutine Resampling_Binary( 
     $	w, npar, mpar, ix, ismp, nn )
*
	implicit real*8( a-h, o-z )
	dimension w(mpar), ismp(mpar), w_sum(0:mpar)
*
	w_sum(0)=0.D0
	do 1 i = 1, npar
	w_sum(i) = w(i) + w_sum(i-1)
    1	continue
*
	do 3 i = 1, nn
	y = runif(ix)
	index_min = 1
	index_max = npar
    2	continue
	index_check = ( index_max + index_min ) / 2
	if ( w_sum( index_check ).lt.y ) then
	index_min = index_check
	if( ( index_min.eq.index_max ).or.
     $	( ( index_max-index_min ).eq.1 ) ) then
	goto 10
	endif
	else
	index_max = index_check
	if(index_min.eq.index_max) then
	goto 10
	endif
	endif
	goto 2
   10	continue
	ismp(i) = index_max
    3	continue 
*
	return
	end
*
*
	subroutine Resampling( weight, npar, mpar, ix, ismp, nn )
*
	implicit real*8( a-h, o-z )
	dimension weight(*), ismp(*)
*
	do i = 1, nn
	y = runif(ix)
	sum = 0.0d0 
	do j = 1, npar 
	sum = sum + weight(j) 
	if( sum.ge.y ) then
	ismp(i) = j
	goto 1
	endif
      enddo 
    1 continue
      enddo 
*
	return
	end
*
*
	subroutine ArrInit( a, numRow, numCol, value )
*
	implicit real*8 ( a-h, o-z )
	dimension a( numRow, numCol )
*
!$omp parallel do
	do i = 1, numRow
	do j = 1, numCol
	a(i,j) = value
	enddo
	enddo
!$omp end parallel do	
*
	return
	end
*
*
	subroutine VecInit( a, len, value )
*
	implicit real*8 ( a-h, o-z )
	dimension a( len )
*
!$omp parallel do
	do i = 1, len
	a(i) = value
	enddo
!$omp end parallel do	
*
	return
	end
*
*
	subroutine ArrInit_Integer( int_mat, numCol, numRow, int_scal )
*
	implicit real*8 ( a-h, o-z )
	dimension int_mat( numCol, numRow )
*
!$omp parallel do
	do i = 1, numCol
	do j = 1, numRow
	int_mat(i,j) = int_scal
	enddo
	enddo
!$omp end parallel do	
*
	return
	end
*
*
	subroutine VecInit_Integer( int_vec, len, int_scal )
*
	implicit real*8 ( a-h, o-z )
	dimension int_vec( len )
*
!$omp parallel do
	do i = 1, len
	int_vec(i) = int_scal
	enddo
!$omp end parallel do	
*
	return
	end
*
*
      subroutine order_increase( integer_obj, iorder, n )
*
      implicit real*8 ( a-h, o-z )
      dimension integer_obj(*), iorder(*)
      dimension itemp(n)
*
!$omp parallel
*
!$omp do
      do i = 1, n
      itemp(i) = integer_obj(i)
      enddo
!$omp end do
*
!$omp end parallel
*
      min_overall = 10**7
      do i = 1, n
      min_local = 10**6      
      do j = 1, n
      if( itemp(j). LE. min_local ) then
      iorder(i) = itemp(j)
      min_local = itemp(j)
      location = j
      endif
      enddo
      itemp(location) = min_overall
      enddo
*
      return
      end
*
*
      subroutine rank_increase( integer_obj, irank, n )
*
      implicit real*8 ( a-h, o-z )
      dimension integer_obj(*), irank(*)
      dimension itemp(n), iorder(n)
*
!$omp parallel
*
!$omp do
      do i = 1, n
      itemp(i) = integer_obj(i)
      enddo
!$omp end do
*
!$omp end parallel
*
      ir = 0
      min_overall = 10**7
      do i = 1, n
      ir = ir + 1
      min_local = 10**6
      location_even = 0
      do j = 1, n
      if( itemp(j). LE. min_local ) then
      min_local = itemp(j)
      iorder(i)= min_local
      location = j
      endif
      enddo
      if( i. EQ. 1 ) then 
      itemp(location) = min_overall
      irank(location) = ir
      else
      if( iorder(i). EQ. iorder(i-1) ) then
      ir = ir - 1
      itemp(location) = min_overall
      irank(location) = ir            
      else      
      itemp(location) = min_overall
      irank(location) = ir
      endif
      endif
      enddo
*
      return
      end
*
*
      subroutine order_matrix_real_row( 
     $      a, ncol, nrow, max_col, max_row, iorder_row )
*
      implicit real*8(a-h,o-z)
      dimension a(max_row,max_col), b(max_row,max_col)
      dimension iorder_row(nrow) 
*
!$omp parallel do
      do i = 1, nrow
      do j = 1, nrow
      k = iorder_row(j)
      if( k. EQ. i ) then
      do ii = 1, ncol
      b(i,ii) = a(k,ii)
      enddo
      endif
      enddo
      enddo
!$omp end parallel do
*
!$omp parallel do
      do i = 1, nrow
      do j = 1, ncol
      a(i,j) = b(i,j)
      enddo
      enddo
!$omp end parallel do      
*
      return
      end
*
*
      subroutine order_matrix_real_col( 
     $      a, nrow, ncol, max_row, max_col, iorder_col )
*
      implicit real*8(a-h,o-z)
      dimension a(max_row,max_col), b(max_row,max_col)
      dimension iorder_col(ncol) 
*
*
!$omp parallel do
      do i = 1, ncol
      do j = 1, ncol
      k = iorder_col(j)
      if( k. EQ. i ) then
      do ii = 1, nrow
      b(ii,i) = a(ii,k)
      enddo
      endif
      enddo
      enddo
!$omp end parallel do
*
!$omp parallel do
      do i = 1, nrow
      do j = 1, ncol
      a(i,j) = b(i,j)
      enddo
      enddo
!$omp end parallel do      
*
      return
      end       
*
*
      subroutine order_vector_real( a, n, max_n, iorder )
*
      implicit real*8(a-h,o-z)
      dimension a(max_n), b(max_n)
      dimension iorder(n) 
*
!$omp parallel do
      do i = 1, n
      k = iorder(i)
      b(k) = a(i)
      enddo
!$omp end parallel do
*
!$omp parallel do
      do i = 1, n
      a(i) = b(i)
      enddo
!$omp end parallel do      
*
      return
      end
*
*
      subroutine order_vector_char( a, n, max_n, iorder )
*
      implicit real*8(a-h,o-z)
      character a(max_n)*(*), b(max_n)*30
      dimension iorder(n) 
*
!$omp parallel do
      do i = 1, n
      k = iorder(i)
      b(k) = a(i)
      enddo
!$omp end parallel do
*
!$omp parallel do
      do i = 1, n
      a(i) = b(i)
      enddo
!$omp end parallel do      
*
      return
      end
*
*
      subroutine order_vector_integer( a, n, max_n, iorder )
*
      implicit real*8(a-h,o-z)
      integer a(max_n), b(max_n)
      dimension iorder(n) 
*
!$omp parallel do
      do i = 1, n
      k = iorder(i)
      b(k) = a(i)
      enddo
!$omp end parallel do
*
!$omp parallel do
      do i = 1, n
      a(i) = b(i)
      enddo
!$omp end parallel do      
*
      return
      end
*
*
      subroutine replace_array_integer( 
     $  int_mat, ncol, nrow, mcol, mrow, icol_replace, m,
     $  num_original, num_replace, n
     $  )
*
      implicit real*8(a-h,o-z)
      dimension int_mat( mrow, mcol )
      dimension icol_replace(*)
      dimension num_original(*)
      dimension num_replace(*)
*      character str_attribute(mrow,*)*(*)
*      character str(*)
*      dimension icol_str_replace(*)     
*
      do i = 1, ncol
      do ii = 1, m
      if( i. EQ. icol_replace(ii) ) then
      irep = i
      icol_replace(ii) = 0
      goto 1
      endif
      enddo
    1 continue
      do k = 1, n
      do j = 1, nrow
      if( int_mat(j,irep). EQ. num_original(k) ) then
      int_mat(j,irep) = num_replace(k)
      endif
      enddo
      enddo 
      enddo
*
      return
      end                  
*
*      
      subroutine ave( object, n, sample_mean )
*
      implicit real*8 ( a-h, o-z )
      dimension object(*)
*
      sum = 0.0d0 
!$omp parallel do reduction(+:sum)
      do i = 1, n
      sum = sum + object(i)
      enddo
!$omp end parallel do
      sample_mean = sum / dble(n)
*      
      return
      end
*
*
      subroutine get_max( object, n, value_max, loc_max )
*
      implicit real*8 ( a-h, o-z )
      dimension object(*)
*
      value_max = object(1)
*
!$omp parallel
*
!$omp do reduction(max:value_max)
      do i = 1, n
      value_max = max(object(i),value_max)
      enddo     
!$omp end do
*
!$omp do
      do i = 1, n
      if( value_max.eq.object(i) ) loc_max = i
      enddo     
!$omp end do
*
!$omp end parallel      
*      
      return
      end
*
*
      subroutine get_min( object, n, value_min, loc_min )
*
      implicit real*8 ( a-h, o-z )
      dimension object(*)
*
      value_min = object(1)
*
!$omp parallel 
*
!$omp do reduction(min:value_min)
      do i = 1, n
      value_min = min(object(i),value_min)
      enddo     
!$omp end do
*
!$omp do
      do i = 1, n
      if( value_min.eq.object(i) ) loc_min = i
      enddo     
!$omp end do
*
!$omp end parallel
*      
      return
      end            
*
*
	subroutine SequentialAve( currMean, currSum, weight, sample, 
     $	lenRow, maxRow, maxCol )
*
	implicit real*8( a-h, o-z )
	dimension currMean( maxRow ), sample( maxRow, maxCol )
*
!$omp parallel do
	do i = 1, lenRow
	currMean(i) = ( currSum * currMean(i) 
     $	+ weight * sample(i,1) ) / ( weight + currSum )
	enddo
!$omp end parallel do	
*
	return
	end
*
*
	subroutine SequentialCovar( currCovar, currMean, currSum, weight, 
     $	sample, lenRow, maxRow, maxCol )
*
	implicit real*8( a-h, o-z )
	dimension currCovar( maxRow, maxRow ), currMean( maxRow ) 
	dimension sample( maxRow, maxCol )
*
!$omp parallel do
	do i = 1, lenRow
	do j = 1, lenRow
	currCovar(i,j) = ( currSum * currCovar(i,j) 
     $	+ currSum * ( sample(i,1) - currMean(i) ) * 
     $	( sample(j,1) - currMean(j) )  ) / 
     $	( weight + currSum )
	enddo
	enddo
!$omp end parallel do	
*
	return
	end
*
*
	subroutine OutputMatrix( a, ncol, nrow, maxcol, maxrow, fileName )
*
	implicit real*8 ( a-h, o-z )
	integer*2  tab /2313/
	dimension a( maxrow, maxcol )
	character fileName * (*)
*
	open( 5, file = fileName )
	do i = 1, nrow
	write( 5, '( 10000( f15.5, a1 ) )' ) 
     $	a(i,1), ( ( tab, a(i,j) ), j = 2, ncol )
	enddo
	close(5)
*
	return
	end
*
*
	subroutine OutputVector( a, n, maxn, fileName )
*
	implicit real*8 ( a-h, o-z )
	integer*2  tab /2313/
	dimension a( maxn )
	character fileName * (*)
*
	open( 5, file = fileName )
	do i = 1, n
	write( 5, '( f15.5 )' ) a(i)
	enddo
	close(5)
*
	return
	end      	
*
*
	subroutine OutputVector_Integer( a, n, maxn, fileName )
*
	implicit real*8 ( a-h, o-z )
	integer*2  tab /2313/
	integer a( maxn )
	character fileName * (*)
*
	open( 5, file = fileName )
	do i = 1, n
	write( 5, '( i5 )' ) a(i)
	enddo
	close(5)
*
	return
	end      	
*
*
	subroutine Save_Particle(
*
     $  Particle_ParaModules, Particle_ParaElements, Particle_Inits,
     $  nsave_particle, num_Reg_Modules, num_Reg_Elements, num_Variable,
     $  max_num_Reg_Modules, max_num_Reg_Elements, max_num_Variable,
     $  num_overlap_DP, str     	
*
     $  )
*
	implicit real*8 ( a-h, o-z )
	integer*2  tab /2313/
	dimension Particle_ParaElements( nsave_Particle, 
     $	                    max_num_Reg_Elements, 2 )
	dimension Particle_ParaModules( nsave_Particle, max_num_Reg_Modules )
	dimension Particle_Inits( nsave_Particle, max_num_Variable )
	dimension num_overlap_DP( nsave_Particle )	
	character fileName * 50
	character str * (*)
*
*
      fileName = 'demo_particle_element_1_'//trim( str )//'.log'      
	open( 5, file = trim(fileName) )
	do i = 1, nsave_Particle
	write( 5, '( i8, 10000( a1,f15.5 ) )' ) 
     $	num_overlap_DP(i), 
     $  ( ( tab, Particle_ParaElements( i, j, 1 ) ), 
     $  j = 1, num_Reg_Elements )
	enddo
	close(5)
*
*
      fileName = 'demo_particle_element_2_'//trim( str )//'.log'
	open( 5, file = trim(fileName) )
	do i = 1, nsave_Particle
	write( 5, '( i8, 10000( a1, f15.5 ) )' ) 
     $	num_overlap_DP(i), 
     $  ( ( tab, Particle_ParaElements( i, j, 2 ) ), 
     $  j = 1, num_Reg_Elements )
	enddo
	close(5)
*
      fileName = 'demo_particle_module_'//trim( str )//'.log'
	open( 5, file = trim(fileName) )
	do i = 1, nsave_Particle
	write( 5, '( i8, 10000( a1, f15.5 ) )' ) 
     $	num_overlap_DP(i), 
     $  ( ( tab, Particle_ParaModules( i, j ) ), 
     $  j = 1, num_Reg_Modules )
	enddo
	close(5)
*
      fileName = 'demo_particle_inits_'//trim( str )//'.log'
	open( 5, file = trim(fileName) )
	do i = 1, nsave_Particle
	write( 5, '( i8, 10000( a1, f15.5 ) )' )
     $	num_overlap_DP(i),
     $  ( ( tab, Particle_Inits( i, j ) ),
     $  j = 1, num_Variable )
	enddo
	close(5)
*
	return
	end	
*
*
	subroutine SummaryStatistics(
*	 
     $	    data, num_var, num_sample, 
     $      max_var, max_sample,
     $	    ave_var, value_maxmin, loc_maxmin 
*
     $	)
*
	implicit real*8( a-h, o-z )
*	
	parameter ( cntlMax = -10.0d20, cntlMin = 10.0d20 )
	dimension data( max_var, max_sample )
	dimension ave_var( max_var, 2 )
	dimension value_maxmin( max_var, 2 )
	dimension loc_maxmin( max_var, 2 )
*
*	/ Sample mean for each variable  
	do i = 1, num_var
	sum = 0.0d0
	do j = 1, num_sample
	sum = sum + data( i, j )
	enddo
	ave_var(i,1) = sum / dble(num_sample)
	enddo
*
*	/ Sample variance for each variable
	do i = 1, num_var
	sum = 0.0d0
	do j = 1, num_sample
	sum = sum + ( data( i, j ) - ave_var(i,1) ) ** 2
	enddo
	ave_var(i,2) = sum / dble(num_sample)
	enddo
*
*	/ Max and Min for each variable
	do i = 1, num_var
	value_maxmin(i,1) = cntlMax
	value_maxmin(i,2) = cntlMin
	do j = 1, num_sample
	if( data(i,j). GE. value_maxmin(i,1) ) then 
	value_maxmin(i,1) = data(i,j)
	loc_maxmin(i,1) = j
	endif
	if( data(i,j). LE. value_maxmin(i,2) ) then 
	value_maxmin(i,2) = data(i,j)
	loc_maxmin(i,2) = j
	endif
	enddo
	enddo
*
	return
	end
*
*
*-------------------------------------------------------------------------------------------------
*
*     Name:     Gauss_Kernel_Regression
*
*     Description:  Smoothing values is interpolated for given observations of each variables 
*                   in [y], using Gaussian kernel nonparametoric regression with fixed bandwidth 
*                   and ridge parameters
*
*-------------------------------------------------------------------------------------------------
*
      subroutine Gauss_Kernel_Regression(
*
*		--- [y] Observed data, [x] Smoothed estimates  over [ind_obs_time]      	
     $		y, x, tall, tobs, nvar, ins_tall, ins_tobs, ins_nvar,
*
*		--- [band] Bandwidths (variance) of Gauss kernels, [shrink] L2 regulalization parameter  	
     $		band, shrink,
*
*		--- [ind_obs_var] Observed variables for [y], [ind_obs_time] Observed times for [y] 
     $		ind_obs_var, ind_obs_time,
*
*		--- [coeff] Estimated coefficients (number of basis functions times number of variables)
     $		coeff, max_num_basis, resid, iDoPrint
     $	)
*
	implicit real*8 ( a-h, o-z )
*
	integer*2  tab /2313/
	integer*4  tall
	integer*4  tobs
	dimension  y( ins_nvar, ins_tobs )
	dimension  x( ins_nvar, ins_tall )
	dimension  ind_obs_time( ins_tobs )
	dimension  ind_obs_var( ins_nvar )
	dimension  coeff( max_num_basis, ins_nvar )
	dimension  basis( ins_tobs, max_num_basis )
	dimension  sqrt_basis( max_num_basis, max_num_basis )
	dimension  project( max_num_basis, ins_tobs )
	dimension  sample_mean( ins_nvar )
	dimension  band( ins_nvar )
	dimension  resid( ins_nvar, ins_tobs )
*
*	/ Number of basis functions is set to number of observed time points 
	num_basis = tobs
*
* --- For each variable in [y], fit Gaussian radious functions ---
*
* ----- OpenMP -----
!$omp parallel private(i,j,k,sum,xx,ave,bw,basis,sample_mean,
!$omp&    sqrt_basis,det,project,prd)
!$omp do
* ------------------
*
	do ii = 1, nvar   ! Loop for variables
*
	if( ind_obs_var(ii).eq.1 ) then		! if-loop (1) for variable in which data are observable 
*
      bw = band(ii)
*
*	/ Sample mean, and generating a basis function matrix  
	sum = 0.0d0	
	do i = 1, tobs
	sum = sum + y(ii,i)
	do j = 1, num_basis
	xx = dble( ind_obs_time(i) )
	ave = dble( ind_obs_time(j) )
	basis(i,j) = dGauss_Univ( xx, ave, bw )	! Basis functions are stored to [basis]
	enddo
	enddo
	sample_mean(ii) = sum / dble( tobs )	! Sample mean
*
*	/ Cross-product of basis function matrix with an additive shrinkage parameter
	do i = 1, num_basis
	do j = 1, num_basis
	sum = 0.0d0
	do k = 1, tobs
	sum = sum + basis(k,i) * basis(k,j)
	enddo
	sqrt_basis(i,j) = sum 
	if ( i.eq.j ) sqrt_basis(i,j) = sqrt_basis(i,j) + shrink
	enddo
	enddo
*
*	/ Inversion of the cross-product function matrix is stored at sqrt_basis
	call Det_Inv( sqrt_basis, det, num_basis, max_num_basis )
*
*     / Occurrence of sigularity in sqrt_basis 
	if( det.eq.0.0d0 ) then 
	write( *, '(a)' ) 
     $	'error-conf(1) --<Error in Gauss_Kernel: zero determinant>'
	stop
	endif
*
*	/ Projection matrix: sqrt_basis * basis
	do i = 1, num_basis
	do j = 1, tobs
	sum = 0.0d0
	do k = 1, num_basis
	sum = sum + sqrt_basis(i,k) * basis(j,k)
	enddo
	project(i,j) = sum
	enddo
	enddo
*
*	/ Regression coefficients
	do i = 1, num_basis
	sum = 0.0d0
	do k = 1, tobs
	sum = sum + project(i,k) * ( y(ii,k) - sample_mean(ii) )
	enddo
	coeff(i,ii) = sum
	enddo
*
*	/ Interporation of smoothed estimates over all data (time) points
	do i = 1, tall
	xx = dble(i)
	sum = 0.0d0
	do j = 1, num_basis
	ave = dble( ind_obs_time(j) )
	basis(1,j) = dGauss_Univ( xx, ave, bw )
	sum = sum + basis(1,j) * coeff(j,ii)
	enddo
	x(ii,i) = sum + sample_mean(ii)
	resid(ii,i) = y(ii,i) - x(ii,i)
	enddo
*
	endif
	enddo ! End loop for variable
*
* ----- OpenMP -----	
!$omp end do
!$omp end parallel	
* ------------------	
*
*     / Plot the smoothed estimates
      if( iDoPrint. EQ. 1 ) then
	open( 5, file = 'gauss_kernel.log' )
	write( 5, '( a, 1000( a1, i5 ) )' ) 
     $	'Gauss_Kernel_Regression', 
     $  ( ( tab, ind_obs_var(i) ), i= 1, nvar )
	do j = 1,tall
	write( 5, '( i, 1000( a1, f15.5 ) )' ) 
     $	j, ( ( tab, x(i,j) ), i= 1, nvar )
	enddo
	close( 5 )
	endif
*
	return
	end
*
*
*-------------------------------------------------------------------------------------------------
*
*     Name:     Gauss_Kernel_Regression_BandSelect
*
*     Description:  Leave-one-out cross validation for selecting bandwidth parameters 
*                   of Gaussian kernel regression models 
*
*-------------------------------------------------------------------------------------------------
*
	subroutine Gauss_Kernel_Regression_BandSelect(
*
*		--- [y] Observed data, [x] Smoothed estimates over [ind_obs_time]  
     $		y, x, tall, tobs, nvar, ins_tall, ins_tobs, ins_nvar,
*
*		--- [band] Bandwidth (variance) of Gauss kernels, [shrink] L2 regulalization parameter  	
     $		band, shrink, nbw,
*
*		--- [ind_obs_var] Observed variables for [y], [ind_obs_time] Observed times for [y] 
     $		ind_obs_var, ind_obs_time,
*
*		--- [coeff] Estimated coefficients (number of basis functions times number of variables)
     $		coeff, max_num_basis, iDoPrint
     $	)
*
	implicit real*8 ( a-h, o-z )
	integer*2  tab /2313/
	integer*4  tall
	integer*4  tobs
	dimension  y( ins_nvar, ins_tobs )
	dimension  x( ins_nvar, ins_tall )
	dimension  ind_obs_time( ins_tobs )
	dimension  ind_obs_var( ins_nvar )
	dimension  coeff( max_num_basis, ins_nvar )
	dimension  basis( ins_tobs, max_num_basis )
	dimension  sqrt_basis( max_num_basis, max_num_basis )
	dimension  aa( max_num_basis, max_num_basis )
	dimension  project( max_num_basis, ins_tobs )
	dimension  sample_mean( ins_nvar )
	dimension  band( ins_nvar )
	dimension  bands( nbw )
*	dimension  cv( ins_nvar, nbw )
*
      integer omp_get_thread_num
*
      if( iDoPrint. EQ. 1 ) then
      write(*,*) 
      write(*,'(a)') 
     $ '-Now searching for bandwidth of Gaussian kernels.....'
      endif
*
*     / Number of basis functions
*       When conducting leave-one-out CV, the number of basis functions becomes 
	iDimForInv = tobs - 1
*
*	/ Number of basis functions is set to number of observed time points 
	num_basis = tobs
*
* --- For each variable in [y], fit Gaussian radious functions ---
*
* ----- OpenMP -----
!$omp parallel private(bwcenter,bwUpper,bwLower,bands,
!$omp&    cvmin,cvsum,cv,i,j,k,jj,kk,
!$omp&    sum,xx,ave,bw,basis,sample_mean,
!$omp&    sqrt_basis,irow,icol,
!$omp&    aa,det,project,prd,loccvmin)
!$omp do
* ------------------
*
	do ii = 1, nvar       ! Loop for variables
*     
	if( ind_obs_var(ii).eq.1 ) then	! if-loop (1) for variable in which data are observable 	
*
      bwcenter = dble(tall)
      bwUpper = 2.0d0
      bwLower = 0.5d0
*
   10 continue
*    
	do i = 1, nbw
	bands(i) = ( bwLower + ( dble(i) - 1.0d0 ) * ( bwUpper - bwLower ) 
     $	/ ( dble(nbw) - 1.0d0 ) ) * bwcenter
	enddo
*
      cvmin = 10d10
	do jj = 1, nbw     ! Loop for bandwidth
*	
	bw = bands(jj)
	cvsum = 0.0d0
	do kk = 1, tobs       ! Loop for leave-one-out CV
*
*	/ Sample mean and generating a basis function matrix  
	sum = 0.0d0
	do i = 1, tobs
	if(i.ne.kk) then
	sum = sum + y(ii,i)
	do j = 1, tobs
	if(j.ne.kk) then
	xx = dble( ind_obs_time(i) )
	ave = dble( ind_obs_time(j) ) 
	basis(i,j) = dGauss_Univ( xx, ave, bw )	! Basis functions are stored to [basis]
	endif
	enddo
	endif
	enddo
	sample_mean(ii) = sum / dble( tobs )	! Sample mean
*
*	/ Product of basis function matrix with an additive shrinkage parameter
	do i = 1, tobs
	do j = 1, tobs
	if( i.ne.kk.and.j.ne.kk ) then
	sum = 0.0d0
	do k = 1, tobs
	if( k.ne.kk ) sum = sum + basis(k,i) * basis(k,j)
	enddo
	sqrt_basis(i,j) = sum 
	if ( i.eq.j ) sqrt_basis(i,j) = sqrt_basis(i,j) + shrink
	endif
	enddo
	enddo
*
      irow = 0
      do i = 1, tobs
      if( i.eq.kk ) goto 2
      irow = irow + 1
      icol = 0
      do j = 1, tobs
      if( j.eq.kk ) goto 1
      icol = icol + 1
      aa(irow,icol) = sqrt_basis(i,j)
*      write(*,*) irow, icol, aa(irow,icol)
    1 continue
      enddo
    2 continue
      enddo      
*
*	/ Inversion of basis function matrix is stored at sqrt_basis
	call Det_Inv( aa, det, iDimForInv, max_num_basis )
*
      irow = 0
      do i = 1, tobs
      if( i.eq.kk ) goto 4
      irow = irow + 1
      icol = 0
      do j = 1, tobs
      if( j.eq.kk ) goto 3
      icol = icol + 1
      sqrt_basis(i,j) = aa(irow,icol)
    3 continue
      enddo
    4 continue
      enddo      
*
	if( det.eq.0.0d0 ) then 
	write( *, '(a)' ) 
     $	'error-conf(1) --<Error in Gauss_Kernel_BandSelect: 
     $  zero determinant>'
	stop
	endif
*
*	/ Projection matrix: sqrt_basis * basis
	do i = 1, tobs
	do j = 1, tobs
	if( i.ne.kk.and.j.ne.kk ) then	
	sum = 0.0d0
	do k = 1, num_basis
	if( k.ne.kk ) sum = sum + sqrt_basis(i,k) * basis(j,k)
	enddo
	project(i,j) = sum
	endif
	enddo
	enddo
*
*	/ Regression coefficients
	do i = 1, tobs
	if( i.ne.kk ) then
	sum = 0.0d0
	do k = 1, tobs
      if( k.ne.kk ) 
     $ sum = sum + project(i,k) * ( y(ii,k) - sample_mean(ii) )
	enddo
	coeff(i,ii) = sum
	endif
	enddo
*
*	/ Interporation over all data (time) points
	do i = 1, tall
	xx = dble(i)
	sum = 0.0d0
	do j = 1, tobs
	if( j.ne.kk ) then
	ave = dble( ind_obs_time(j) )
	basis(1,j) = dGauss_Univ( xx, ave, bw )
	sum = sum + basis(1,j) * coeff(j,ii)
      endif
	enddo
	x(ii,i) = sum + sample_mean(ii)
	enddo
*
*     / CV error
	xx = dble(ind_obs_time(kk))
	sum = 0.0d0
	do j = 1, tobs
	if( j.ne.kk ) then
	ave = dble( ind_obs_time(j) )
	basis(1,j) = dGauss_Univ( xx, ave, bw )
	sum = sum + basis(1,j) * coeff(j,ii)
      endif
	enddo
	prd = sum + sample_mean(ii)
	cvsum = cvsum + ( y(ii,kk) - prd ) ** 2
*
	enddo
*
      cv = cvsum / dble( tobs )
*
      if( cv.lt.cvmin ) then
      cvmin = cv
      loccvmin = jj
      endif
*
	enddo
*
      if(loccvmin.eq.1.or.loccvmin.eq.nbw) then
      bwcenter = bands(loccvmin)
      goto 10
      endif
*
      if( iDoPrint. EQ. 1 ) then
!$    id = omp_get_thread_num()
      write(*,'(a,i3,f10.3,i3)') 'Varible', ii, bands(loccvmin), id
      endif
*      
      band(ii) = bands(loccvmin)
*	
	endif	
	enddo
*	
* ----- OpenMP -----	
!$omp end do
!$omp end parallel	
* ------------------
*      
*     / Plot the smoothed estimates
      if( iDoPrint. EQ. 1 ) then
	open( 5, file = 'gauss_kernel.log' )
	write( 5, '( a, 1000( a1, i5 ) )' ) 
     $	'Gaussian_Kernal_Regression_BandSelect', 
     $  ( ( tab, ind_obs_var(i) ), i = 1, nvar )
	do j = 1,tall
	write( 5, '( i, 1000( a1, f15.5 ) )' ) 
     $	j, ( ( tab, x(i,j) ), i = 1, nvar )
	enddo
	close( 5 )
	endif
*
	return
	end	
*
*
	subroutine Gauss_Kernel_Gradient(
*
*		--- Gradient vectors {gradient} of given basis functions is evaluated over 1:[timeAll]      	
     $		gradient, timeAll, timeMeasure, nvar, 
     $		max_nvar, max_timeAll, max_timeMeasure,
*
*		--- [band] Bandwidth (variance) of Gauss kernels, [centerKernel] Center of basis functions -> 1:[timeMeasure]  	
     $		band, centerKernel,
*
*		--- [measure_variable] Observed variables in data, [measure_time] Observed time points in data
     $		measure_variable, measure_time,
*
*		--- [coeff] Estimated coefficients (number of basis functions times number of variables)
     $		coeff, max_num_basis, num_basis,
*
*		--- Print indicator 'gauss_kernel_gradient.log'
     $		iprint
     $	)
*
	implicit real*8 ( a-h, o-z )
	integer*2  tab /2313/
	integer*4  timeAll
	integer*4  timeMeasure
	dimension gradient( max_nvar, max_timeAll )
	dimension centerKernel( max_timeMeasure )
	dimension measure_time( max_timeMeasure )
	dimension measure_variable( max_nvar )
	dimension coeff( max_num_basis, max_nvar )
	dimension band( max_nvar )
	character fileName * 15
*
*	*/ Centers of basis functions allocated to time points of observed data
	do i = 1, timeMeasure
	centerKernel(i) = dble( measure_time(i) )
	enddo
*
* --- For each variable (if it is obseved), compute gradients of fitted Gauss kernel regressors --- 
*
	do iv = 1, nvar
*
	if( measure_variable(iv). eq. 1 ) then
	bw = band(iv)
*
	do i = 1, timeAll
	sumGrad = 0.0d0
	do j = 1, timeMeasure
	xx = dble(i)
	ave = centerKernel(j)
	density = dGauss_Univ( xx, ave, bw )
	diff = ( xx - ave ) / ( 2.0d0 * bw )
      sumGrad = sumGrad - coeff( j, iv ) * density * diff
	enddo
	gradient( iv, i ) = sumGrad
	enddo
*
	endif
*
	enddo
*
	if( iprint. eq. 1 ) then
	call OutputMatrix( gradient, timeAll, nvar, 
     $	max_timeAll, max_nvar, 'gauss_kernel_gradient.log' )
	endif
*
	return
	end
*
*
	subroutine Gauss_Kernel_Duration(
     $		durTimeI, durTimeD, nDurTimeI, nDurTimeD,	 
     $		gradient, max_num_Variable, num_Variable, 
     $		max_time_StateVariable, time_StateVariable,
     $		nDecompTime, ID_MissVariable, iDoPrint
     $	)
*
	implicit real*8 ( a-h, o-z )
	integer*2 tab /2313/
	integer*4 time_StateVariable
	dimension ID_MissVariable( max_num_Variable )
	dimension gradient( max_num_Variable, max_time_StateVariable )
	dimension igradient( max_num_Variable, max_time_StateVariable )
	dimension durTimeI( max_num_Variable, nDecompTime, 2 )
	dimension durTimeD( max_num_Variable, nDecompTime, 2 )
	dimension nDurTimeI( max_num_Variable )
	dimension nDurTimeD( max_num_Variable )
*
*     / Initialization for duration time matrices
	call ArrInit( durTimeI, nDecompTime, num_Variable, 0.0d0 )
	call ArrInit( durTimeD, nDecompTime, num_Variable, 0.0d0 )
*
*     / Sign matrices based on increasing and decreasing patterns 
	do iv = 1, num_Variable
	do i = 1, time_StateVariable
	igradient(iv,i) = 0
	if( gradient(iv,i). GT. 0.0d0 ) igradient(iv,i) = 1
	if( gradient(iv,i). LT. 0.0d0 ) igradient(iv,i) = -1
	enddo
	enddo
*
	do iv = 1, num_Variable   ! Loop for each varible
*
	if( ID_MissVariable(iv). EQ. 1 ) then
*
	idurI = 0     ! Flag for increasing duration at the current time point
	idurD = 0     ! Flag for decreasing duration at the current time point 
	nI = 0    ! Count flag for increasing durations
	nD = 0    ! Count flag for decreasing durations
*
	do i = 1, time_StateVariable  ! Loop for time points
*
*     / Starting time for increasing duration
	if( ( igradient(iv,i) .EQ. 1 ). AND. ( idurI. EQ. 0 ) ) then
*	if( iDoPrint. EQ. 1 ) write(*,*) 'StI', dble(i)
	nI = nI + 1
	idurI = 1
	nDurTimeI(iv) = nI
	durTimeI(iv,nI,1) = dble(i)
	endif
*
*     End time for the increasing duration
	if( ( igradient(iv,i) .NE. 1 ). AND. ( idurI. EQ. 1 ) ) then
*	if( iDoPrint. EQ. 1 ) write(*,*) 'EnI', dble(i-1)
	idurI = 0
	durTimeI(iv,nI,2) = dble(i-1)
	endif
*
*     Starting time for decreasing duration
	if( ( igradient(iv,i) .EQ. -1 ). AND. ( idurD. EQ. 0 ) ) then
*	if( iDoPrint. EQ. 1 ) write(*,*) 'StD', dble(i)
	nD = nD + 1
	idurD = 1
	nDurTimeD(iv) = nD
	durTimeD(iv,nD,1) = dble(i)
	endif
*
*     End time for decreasing duration
	if( ( igradient(iv,i) .NE. -1 ). AND. ( idurD. EQ. 1 ) ) then
*	if( iDoPrint. EQ. 1 ) write(*,*) 'EnD', dble(i)
	idurD = 0
	durTimeD(iv,nD,2) = dble(i-1)
	endif
*
	if( i. EQ. time_StateVariable ) then
	if( idurD. EQ. 1 ) durTimeD(iv,nD,2) = i
	if( idurI. EQ. 1 ) durTimeI(iv,nI,2) = i
	endif  
	enddo
	endif
*
	enddo
*
*
	if( iDoPrint. EQ. 1 ) then
	open( 5, file = 'gauss_kernel_duration.log' )
	do iv = 1, num_Variable
*
	if( ID_MissVariable(iv). EQ. 1 ) then
*
	write(5,'(a,a1,i,a1,a)') 'Variable:', tab, iv, tab, 'increasing'
	do i = 1, nDurTimeI(iv)
	write(5,'(f10.5,a1,f10.5)') 
     $	durTimeI(iv,i,1), tab, durTimeI(iv,i,2)
	enddo
*
	write(5,'(a,a1,i,a1,a)') 'Variable:', tab, iv, tab, 'decreasing'
	do i = 1, nDurTimeD(iv)
	write(5,'(f10.5,a1,f10.5)') 
     $	durTimeD(iv,i,1), tab, durTimeD(iv,i,2)
	enddo
*
	endif
*
	enddo
	close(5)
	endif
*
	return
	end
*
*
	DOUBLE PRECISION FUNCTION dGauss_Univ(X,U,S)
*
*	Double Precision Function to Calculate Density Function of 
*	Multivariate Gaussian
*	 @ written by R.YOSHIDA 
* 
*	Input Variables 
*	  X: Variable
*	  K,KMAX: Dimension of X and its Adjustable Dimension
*	  U,S:	Parameters
*	
	IMPLICIT REAL*8(A-H,O-Z)
	DATA PI/2.506628275D0/
*	
	dGauss_Univ = DEXP( - ( (X-U)**2 )/ ( S*2.0D0 ) ) /( PI*DSQRT(S) )
*
	RETURN
	END
*
*
	SUBROUTINE Det_Inv(X,DET,M,MJ)
*
*	SUBROUTINE FOR CALCULATING MATRIX INVERSION AND DETERMINANT
*	 @ WRITTEN BY G.KITAGAWA
*      
*	Input Variables
*		X: M*M Matrix
*		M,MJ: Dimension of X and its Adjustable Dimension 
*
*	Output Variables 
*		X: Inverce Matrix of X
*		DET: Determinat 
*
	IMPLICIT REAL*8(A-H,O-Z)
	DIMENSION X(MJ,MJ),IND(300)
*
	DET=1.0D0
	DO 60 L=1,M
	XMAX=0.1D-10
	IMAX=0
	DO 10 I=L,M
	IF(DABS(X(I,L)).GT.DABS(XMAX)) THEN
	XMAX=X(I,L)
	IMAX=I
	ENDIF
   10 CONTINUE
      IND(L)=IMAX
	IF(IMAX.NE.L) THEN
	IF(IMAX.LE.0) THEN
	DET=0.0D0
	RETURN
	ENDIF
*	Row Interchange
	DO 20 J=1,M
	XTEMP=X(IMAX,J)
	X(IMAX,J)=X(L,J)
   20 X(L,J)=XTEMP
	DET=-DET
      ENDIF
	DET=DET*XMAX
	X(L,L)=1.0D0
	DO 30 J=1,M
   30 X(L,J)=X(L,J)/XMAX
	DO 50 I=1,M
	IF(I.NE.L) THEN
	XTEMP=X(I,L)
	X(I,L)=0.0D0
	DO 40 J=1,M
   40 X(I,J)=X(I,J)-XTEMP*X(L,J)
	ENDIF
   50 CONTINUE
   60 CONTINUE
*	Colum Interchange
      IF(M.GT.1)THEN
	DO 110 J=1,M-1
	JJ=IND(M-J) 
	IF(JJ.NE.M-J)THEN
	DO 100 I=1,M
	XTEMP=X(I,JJ)
	X(I,JJ)=X(I,M-J)
  100	X(I,M-J)=XTEMP
	ENDIF
  110 CONTINUE
      ENDIF
*
      RETURN
	END		 
*
*
	double precision function Combination( n, k )
*
	integer*4 n, k 
	real*8 r1, r2, r3
*
	r1 = 0.0d0
      r2 = 0.0d0
      r3 = 0.0d0
*
	do i = 1, n
	r1 = r1 + dlog( dble(i) )
	enddo
*
	do i = 1, k
	r2 = r2 + dlog( dble(i) )
	enddo
*
	do i = 1, ( n - k )
	r3 = r3 + dlog( dble(i) )
	enddo
*
	Combination = dexp(r1) / ( dexp( r2 + r3 ) )
*
	return
	end
*
*
	subroutine Duration_Select_Max(
*
*		--- Max duration time for each variable
     $		durMaxTimeI, durMaxTimeD,
*     	
*		--- Set of duration times in increasing and decreasing trends
     $		durTimeI, durTimeD, nDurTimeI, nDurTimeD, nDecompTime,
*
*		--- Others
     $		max_num_Variable, num_Variable, ID_MissVariable,
     $		max_time_StateVariable
*
     $	)
*
	implicit real*8 ( a-h, o-z )
	integer*2 tab /2313/
*
	dimension durTimeI( max_num_Variable, nDecompTime, 2 )
	dimension durTimeD( max_num_Variable, nDecompTime, 2 )
	dimension nDurTimeI( max_num_Variable )
	dimension nDurTimeD( max_num_Variable )
	dimension ID_MissVariable( max_num_Variable )
	dimension durMaxTimeI( max_num_Variable, 2 )
	dimension durMaxTimeD( max_num_Variable, 2 )
*
	call ArrInit( durMaxTimeI, max_num_Variable, 2, 0.0d0 )
	call ArrInit( durMaxTimeD, max_num_Variable, 2, 0.0d0 )
*
	do iv = 1, num_Variable
*
	if( ID_MissVariable(iv). EQ. 1 ) then
*
	durMax = 0.0d0
	do i = 1, nDurTimeI(iv)
	tmp = durTimeI( iv, i, 2 ) - durTimeI( iv, i, 1 )
	if( tmp. GT. durMax ) then
	durMax = tmp
	durMaxTimeI( iv, 1 ) = durTimeI( iv, i, 1 )
	durMaxTimeI( iv, 2 ) = durTimeI( iv, i, 2 )
	endif
	enddo
*
	durMax = 0.0d0
	do i = 1, nDurTimeD(iv)
	tmp = durTimeD( iv, i, 2 ) - durTimeD( iv, i, 1 )
	if( tmp. GT. durMax ) then
	durMax = tmp
	durMaxTimeD( iv, 1 ) = durTimeD( iv, i, 1 )
	durMaxTimeD( iv, 2 ) = durTimeD( iv, i, 2 )
	endif
	enddo
*
	endif	
*
	enddo
*
	return
	end
*
*
	subroutine Duration_SteadyState(
*
*		--- Rates of degradation and transcription
     $		degrad_RNA, synthesis_RNA, 
*
*		--- Time ranges (start and end) for increasing and decreasing phases
     $		durMaxTimeI, durMaxTimeD,
*
*		--- Missing time series indicator, time interval
     $		ID_MissVariable, time_interval,
*
*		--- State variables (reference)
     $		StateVariable, max_num_Variable, num_Variable, 
     $		max_time_StateVariable, time_StateVariable,
*
*       --- Turning parameters
     $      discountRateTranscpript
     $	)
*
	implicit real*8 ( a-h, o-z )
	integer*2 tab /2313/
	integer*4 time_StateVariable
*
	dimension StateVariable( max_num_Variable, * )
	dimension ID_MissVariable( max_num_Variable )
	dimension durMaxTimeI( max_num_Variable, 2 )
	dimension durMaxTimeD( max_num_Variable, 2 )
	dimension degrad_RNA( max_num_Variable )
	dimension synthesis_RNA( max_num_Variable )
	dimension discountRateTranscpript(2)
*
	threshold_minState = 1.0d-20
*
*     / Initialization
	call VecInit( degrad_RNA, max_num_Variable, 0.0d0 )
	call VecInit( synthesis_RNA, max_num_Variable, 0.0d0 )
*
*     / For each variable observable as a reference state [StateVariable],
*       evaluate reasonable values for [degrad_RNA] and [synthesis_RNA]
	do iv = 1, num_Variable
*
	if( ID_MissVariable(iv). EQ. 1 ) then
*
*     / Degradation rate
	istart = int( durMaxTimeD(iv,1) )
	iend = int( durMaxTimeD(iv,2) )
*
	ratio = StateVariable(iv,iend) * discountRateTranscpript(2) 
     $	    / StateVariable(iv,istart)
*
	degrad_RNA(iv) = 1.0d0 - ratio ** (1.0d0 / dble(iend - istart)) 
	degrad_RNA(iv) = degrad_RNA(iv) / time_interval
*
*     / Transcription rate
	istart = int( durMaxTimeI(iv,1) )
	iend = int( durMaxTimeI(iv,2) )
	coef = 1.0d0 - time_interval * degrad_RNA(iv) 
	sumPower = 1.0d0
	power = 1.0d0
	do i = 2, (iend-istart-1)
	power = dexp(dlog(power) + dlog(coef)) 
	sumPower = sumPower + power
	enddo
	power = dexp(dlog(power) + dlog(coef))
	a = StateVariable(iv,iend) *  discountRateTranscpript(1)
     $	 - power * StateVariable(iv,istart)
	synthesis_RNA(iv) = a / (time_interval * sumPower)
*
	endif
*
	enddo
*
	return
	end
*
*     
      subroutine whoami
!$      external omp_get_thread_num
      integer iam, omp_get_thread_num
*
!$      iam = omp_get_thread_num()
!$omp critical
      write(*,*) 'I am', iam
!$omp end critical
      return
      end
*
*
      subroutine test_omp( xx, yy,n )
*
      implicit real*8(a-h,o-z)
      dimension xx(n), yy(n)
      dimension pp(n)
      integer iam, omp_get_thread_num      
*
!$omp parallel private(pp,temp)
!$omp do      
      do i= 1,n
!$      iam = omp_get_thread_num()
      temp = xx(i)
      pp(i) = temp
      yy(i) = xx(i)
      write(*,'(a,i3,4f10.4)') 'thread', iam, temp,xx(i), pp(i), yy(i)
      enddo
!$omp end do
!$omp end parallel
*
      return
      end
*
*
      subroutine relation_translation(
*      
     $      List_Reg_Modules, List_Reg_Elements, strModules,
     $      num_Reg_Modules, num_Reg_Elements,
     $      max_num_Reg_Modules, max_num_Reg_Elements, max_tmp, 
     $      itranslation, num_Variable, max_num_Variable,
     $      Name_Variable,ID_Variable, imRNA
*
     $  )
*
      implicit real*8 (a-h,o-z)
      dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
      dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	character strModules( max_num_Reg_Modules, 3 ) * 30
	character Name_Variable(*)*(*)
	dimension ID_Variable(*)
      dimension itranslation(*)
      character tag_transl*(*)
      dimension imRNA( max_num_Variable )
*
      parameter ( tag_transl = 'Translation' )
*
*
*     / Initialization
      do i = 1, num_Variable              
      itranslation(i) = 0
      imRNA(i) = 0
      enddo
*
*     / Search protein and its mRNA     
      do i = 1, num_Reg_Modules
      write(*,*) strModules(i,2)
      if( strModules(i,2). eq. tag_transl ) then
      k = List_Reg_Modules(i,2)     ! Product ID (a protein)
      ii = List_Reg_Modules(i,5)     ! ID Reaction Element 
      do j = 1, num_Reg_Elements
      if( List_Reg_Elements(j,1).eq.ii ) then
      mRNA = List_Reg_Elements(j,3)
      imRNA(mRNA) = k
      exit
      endif
      enddo
      itranslation(k) = mRNA
      write(*,*) mRNA, k
      endif
      enddo
*
      return
      end
*
*
      subroutine relation_complex(
*
     $      Para_Reg_Modules,      
     $      List_Reg_Modules, List_Reg_Elements, strModules,
     $      num_Reg_Modules, num_Reg_Elements,
     $      max_num_Reg_Modules, max_num_Reg_Elements, max_tmp, 
     $      icomplex, num_Variable, max_num_Variable,
     $      Name_Variable, ID_Variable
*
     $  )
*
      implicit real*8 (a-h,o-z)
      dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
      dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
      dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )      
	character strModules( max_num_Reg_Modules, 3 ) * 30
	character Name_Variable(*)*(*)
	dimension ID_Variable(*)
      dimension icomplex(*)
*      dimension icomplex_pair( max_num_Variable, 10 )
*      dimension icomplex_number( max_num_Variable )      
      character tag_bind*(*)
*
      parameter ( tag_bind = 'Binding' )
*
*
*     / Initialization
      do i = 1, num_Variable              
      icomplex(i) = 0
      enddo
*
*     / Search protein and its mRNA     
      do i = 1, num_Reg_Modules
*
      if( 
     $  ( strModules(i,2). eq. tag_bind ). AND. 
     $  ( Para_Reg_Modules(i,2). gt. 0.0d0 ) 
     $  ) then
*     
      k = List_Reg_Modules(i,2)     ! Product ID (a complex)
      icomplex(k) = 1
*      
      write(*,*) k, Name_Variable(k)
*
      endif
*
      enddo
*
      return
      end
*
*
      subroutine draw_hill_function ()
*
      implicit real*8 ( a-h, o-z )
*
	integer*2  tab /2313/
      parameter ( n = 500, nparam = 10 )
      dimension xx(n)
      dimension power(nparam)
      dimension value_func(nparam, n)
      integer omp_get_thread_num      
*
      const = 0.5d0
      step = 1.0d0 / ( dble(n) - 1 )
*
* ------------------
!$omp   parallel 
!$omp   do
* ------------------
*
      do i = 1, n
      xx(i) = ( dble(i) - 1 ) * step
      enddo
*
* ------------------
!$omp   end do
* ------------------            
*
*!$    write(*,*)
*
* ------------------
!$omp   do
* ------------------
*
      do i = 1, nparam
      power(i) = dble(i)
      enddo
*
* ------------------
!$omp   end do
* ------------------      
*
*
* ------------------
!$omp   do
* ------------------
*
      do i = 1, nparam
      do j = 1, n
*
      temp = xx(j)**power(i)
      value_func(i,j) = temp / ( temp + const**power(i) )
*
*!$    id = omp_get_thread_num()
*!$    write(*,*) id, i, j
*
      enddo
      enddo
*
* ------------------
!$omp   end do
!$omp   end parallel
* ------------------      
*      
      open( 5, file = 'draw_hill_function.log' )
      write(5,'(a, 1000(a1,f10.5))') 'XXX', ( ( tab, xx(i) ), i= 1,n )
      do i = 1, nparam
      write(5,'(f10.5, 1000(a1,f10.5))') 
     $  power(i), ( ( tab, value_func(i,j) ), j= 1,n )
      enddo
      close( 5 )
*
      return
      end
*
*
      subroutine Shift_Prior_ParaElements (
*      
     $      Prior_ParaElements, max_num_Para1, num_Para1, ix, percent          
*
     $  )
*
      implicit real*8 (a-h, o-z)
*
	integer*2  tab /2313/
*
      parameter ( n = 100, ny = 100 )
      parameter ( num_smp = 1000, ngrid = 100 )
      parameter ( sdmin = 0.1d0, sdmax = 1.0d0 )
      parameter ( center = 0.5d0 )
*      
     	dimension Prior_ParaElements( max_num_Para1, 4 )
     	dimension xx(n)
     	dimension yy(ny), value_func(n,ny)
*
      dimension sd( ngrid )
      dimension pr( num_smp )
*
      integer omp_get_thread_num
*
*
      step = ( sdmax - sdmin ) / ( dble(ngrid) - 1.0d0 )
*
!$omp parallel private( smp, num_above_one )
*
!$omp do      
      do i = 1, ngrid
      sd(i) = sdmin + (dble(i)-1) * step
      enddo
!$omp end do
*
!$omp do
      do i = 1, ngrid
*      
!$    id = omp_get_thread_num()
      num_above_one = 0
      do j = 1, num_smp
      smp = center + sd(i) * rgauss(ix)
      if( smp. GE. 1.0d0 ) num_above_one = num_above_one + 1
      enddo
*
      pr(i) = dble( num_above_one ) / dble( num_smp )
*      write(*,'(i,i,2f10.5)') i, id, sd(i), pr(i)-percent
*
      enddo
!$omp end do
*
!$omp end parallel 
*
*
      min_loc = 0
      diff_min = 1.0d20
*
!$omp   parallel private( temp )
*
!$omp do      
      do i = 1, ngrid
      temp = dabs( pr(i) - percent )
!$omp critical      
      if( temp. LT. diff_min ) then
      diff_min = temp
      min_loc = i
      endif
!$omp end critical      
      enddo
!$omp end do
*
*      write(*,*) min_loc, diff_min
*      
!$omp end parallel
*
*
!$omp parallel do
      do i = 1, num_Para1
      Prior_ParaElements(i,1) = center
      Prior_ParaElements(i,3) = sd(min_loc) ** 2      
      enddo
!$omp end parallel do
*
*
      power = 40.0d0
     	step = 1.0d0 / ( dble(ny) - 1 )
*
!$omp parallel do
*
      do i = 1, ny
      yy(i) = (dble(i)-1) * step
      enddo
*
!$omp end parallel do
*
*
      k = 1
*
!$omp parallel do
*
      do i = 1, n
      xx(i) = Prior_ParaElements(k,1)
     $  + dsqrt( Prior_ParaElements(k,3) ) * rgauss(ix)
      enddo
*
!$omp end parallel do
*
*
!$omp parallel do
*
      do i = 1, n
      do j = 1, ny
      temp = yy(j) ** power
      value_func(i,j) = temp / ( temp + xx(i)**power )
      enddo
      enddo
*
!$omp end parallel do      
*            
      open( 5, file = 'draw_hill_function_rand.log' )
      write(5,'(a, 1000(a1,f10.5))') 'XXX', ( ( tab, yy(i) ), i= 1,ny )
      do i = 1, n
      write(5,'(f10.5, 1000(a1,f10.5))') 
     $  xx(i), ( ( tab, value_func(i,j) ), j= 1,ny )
      enddo
      close( 5 )
*
      return
      end
*
*
      subroutine Shift_Prior_Modules(
*      
     $      List_Reg_Modules, Para_Reg_Modules, max_num_Reg_Modules, 
     $      num_Reg_Modules, strModules, max_tmp,
     $      Prior_ParaModules, ID_Prior_ParaModules, 
     $      max_num_Para2, num_Para2, ix,
     $      degrad_RNA, synthesis_RNA, max_num_Variable, num_Variable,
     $      Name_Variable, ID_Variable, ID_MissVariable,
     $      itranslation, icomplex, value_maxminData, RangeHidden,
     $      input_zero_baseline
*
     $ )
*
      implicit real*8 (a-h,o-z)
	integer*2  tab /2313/      
      parameter ( iprint = 1 )
      dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )	
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	character strModules( max_num_Reg_Modules, 3 ) * 30	
	dimension Prior_ParaModules( max_num_Para2, 2 )	
	dimension ID_Prior_ParaModules( max_num_Para2 )
	dimension degrad_RNA( max_num_Variable )
	dimension synthesis_RNA( max_num_Variable )
	character Name_Variable( max_num_Variable ) * 30	! Names of state variables
	dimension ID_Variable( max_num_Variable )			! State variable IDs
	dimension ID_MissVariable( max_num_Variable )
*
	dimension value_maxminData( max_num_Variable, 2 )
      dimension itranslation( max_num_Variable )
      dimension icomplex( max_num_Variable )
      dimension RangeHidden(2)
*
      character tag_deg * (*)
      character tag_syn * (*)      
      parameter ( tag_deg = 'RnaDegradation' )
      parameter ( tag_syn = 'Transcription')
*
      character tag_degP * (*)
      character tag_synP * (*)      
      character tag_bind * (*)            
      parameter ( tag_degP = 'ProteinDegradation' )
      parameter ( tag_synP = 'Translation')
      parameter ( tag_bind = 'Binding')
*
      character tag_base * (*)
      parameter ( tag_base = 'Baseline' )      
*
      dimension pr(max_num_Variable), item(max_num_Variable), idraw(1)
*
*
*     / Print progress
*     ---------------------------------------------------------------------
      if( iprint. EQ. 1 ) then
      write(*,*)
      write(*,'(a)') '-@[Shift_Prior_Modules] Replacement of 
     $ prior distributions involving parameters in Reaction Modules'
      write(*,*)
      write(*,'(a)') '--Rates of degradation and transcription, 
     $ estimated from time course data'
      do i = 1, num_Variable
      write(*,'(a,a,a1,f10.5,a1,f10.5)') 
     $  '--', 
     $  Name_Variable(i), tab, degrad_RNA(i), tab, synthesis_RNA(i)
      enddo
      endif
*     ---------------------------------------------------------------------      
*
*
*     / Aggregate IDs of observed variables into a vector [item]
      num_obs = 0
      do i = 1, num_Variable
      if( ID_MissVariable(i). EQ. 1 ) then
      num_obs = num_obs + 1
      item(num_obs) = i
      endif
      enddo
*
*     / Vector of probabilities used below
      do i = 1, num_obs
      pr(i) = 1.0d0/dble(num_obs)
      enddo 
*
*
*     / Replacement of prior distributions related to mRNA attribute
*
      if( iprint.EQ.1 ) write(*,'(a)') '-@mRNA attribute' 
*
      do i = 1, num_Para2   !---(1)
      mRNA = 0       
*     / Parameter ID of Reaction Module            
      k = ID_Prior_ParaModules(i)
*     / Scan over list of reaction modules
      do j = 1, num_Reg_Modules     !---(2)
      if( List_Reg_Modules(j,3). EQ. k ) then  !---(3)
*     / Product ID
      kP = List_Reg_Modules(j,2)
      m = kP
*      
*     / Replacement of parameters involved in mRNA attribute, but data are missing     
      if( ID_MissVariable(kP). EQ. 0 ) then
    3 continue 
	call Resampling( pr, num_obs, max_num_Variable, ix, idraw, 1 )
	kk = idraw(1)
	if( item(kk). EQ. kP ) goto 3
	kP = item(kk)
      endif
*
*     / If the reaction module is associated to 'RnaDegradation' or 'Transcription',... 
      if( ( strModules(j,2). EQ. tag_deg ). OR. 
     $    ( strModules(j,2). EQ. tag_syn ) ) 
     $  mRNA = 1
*     
*     / RNA degradation
      if( strModules(j,2). EQ. tag_deg ) 
     $      Prior_ParaModules(i,1) = degrad_RNA(kP)
*     / Rate of transcription     
      if( strModules(j,2). EQ. tag_syn ) 
     $      Prior_ParaModules(i,1) = synthesis_RNA(kP)
*
*     / Print replaced value
      if( ( iprint. EQ. 1 ). AND. ( mRNA. EQ. 1 ) ) then
      write(*,'(a,a,a1,a,a1,f10.5)') 
     $      '-', Name_Variable(m), tab, strModules(j,2), tab,  
     $      Prior_ParaModules(i,1)
      endif
*      
      exit
      endif !---(3)
*
      enddo !---(2)
      enddo !---(1)
*
*            
*     / Replacement of parameters involved in translational process
*
      if( iprint.EQ.1 ) write(*,'(a)') '-@Monomer protein' 
*
      do i = 1, num_Variable    !---(1)
      mRNA = itranslation(i)
      if( mRNA. NE. 0 ) then    !---(2)
*      
      if( ID_MissVariable(mRNA). EQ. 0 ) then
    1 continue 
	call Resampling( pr, num_obs, max_num_Variable, ix, idraw, 1 )
	kk = idraw(1)
	if( item(kk). EQ. mRNA ) goto 1
	mRNA = item(kk)
      endif
*      
      i_bnd = 0
      i_syn = 0
      i_deg = 0
      i_bnd_prior = 0
      i_syn_prior = 0
      i_deg_prior = 0
*            
      do j = 1, num_Reg_Modules
      if( ( List_Reg_Modules(j,2).EQ.i ).AND.
     $    ( strModules(j,2). EQ. tag_degP ) ) 
     $      i_deg = List_Reg_Modules(j,3)
      if( ( List_Reg_Modules(j,2).EQ.i ).AND.
     $    ( strModules(j,2). EQ. tag_synP ) ) 
     $      i_syn = List_Reg_Modules(j,3)
      if( ( List_Reg_Modules(j,2).EQ.i ).AND.
     $    ( strModules(j,2). EQ. tag_bind ) ) 
     $      i_bnd = List_Reg_Modules(j,3)  
      enddo
*
      do j = 1, num_Para2
      if( ID_Prior_ParaModules(j). EQ. i_deg )
     $      i_deg_prior = j 
      if( ID_Prior_ParaModules(j). EQ. i_syn )
     $      i_syn_prior = j            
      if( ID_Prior_ParaModules(j). EQ. i_bnd )
     $      i_bnd_prior = j 
      enddo
*
*     / -deg*Pmax + syn*Mmax = 0 --> deg = syn*Mmax/Pmax    
      Prior_ParaModules(i_syn_prior,1) =
*     $  1.0d0 / (value_maxminData(mRNA,1)-value_maxminData(mRNA,2))
     $  1.8d0 / value_maxminData(mRNA,1)
*
      Prior_ParaModules(i_deg_prior,1) = 
     $  Prior_ParaModules(i_syn_prior,1) *
     $  value_maxminData(mRNA,1) / RangeHidden(1)
*
      if( iprint. EQ. 1 )
     $ write(*,'(a,a,a1,f10.5,a1,f10.5,a1,f10.5)')
     $  '-', Name_Variable(i), tab, 
     $  Prior_ParaModules(i_deg_prior,1), tab,
     $  Prior_ParaModules(i_syn_prior,1), tab,
     $  Prior_ParaModules(i_syn_prior,1) * value_maxminData(mRNA,1) 
     $  / Prior_ParaModules(i_deg_prior,1)
*
      endif !--(2)
      enddo !--(1)
*
*            
*     / Replacement of parameters involved in translational process
*
      if( iprint.EQ.1 ) write(*,'(a)') '-@Multimeric protein' 

      do i = 1, num_Variable
      if( icomplex(i). EQ. 1 ) then
      i_deg = 0
      i_bnd = 0
      i_deg_prior = 0
      i_bnd_prior = 0      
      do j = 1, num_Reg_Modules
      if( ( List_Reg_Modules(j,2).EQ.i ).AND.
     $    ( strModules(j,2). EQ. tag_degP ) ) 
     $      i_deg = List_Reg_Modules(j,3)
      if( ( List_Reg_Modules(j,2).EQ.i ).AND.
     $    ( strModules(j,2). EQ. tag_bind ) ) 
     $      i_bnd = List_Reg_Modules(j,3)
      enddo
*
      do j = 1, num_Para2
      if( ID_Prior_ParaModules(j). EQ. i_deg )
     $      i_deg_prior = j 
      if( ID_Prior_ParaModules(j). EQ. i_bnd )
     $      i_bnd_prior = j 
      enddo
*
*
      Prior_ParaModules(i_bnd_prior,1) = 
     $      0.8d0 / ( RangeHidden(1)-RangeHidden(2) )
      Prior_ParaModules(i_deg_prior,1) = 
     $      Prior_ParaModules(i_bnd_prior,1) 
*
      if( iprint. EQ. 1 )
     $ write(*,'(a,a,a1,f10.5,a1,f10.5)')
     $  '-', Name_Variable(i), tab, 
     $  Prior_ParaModules(i_deg_prior,1), tab,
     $  Prior_ParaModules(i_bnd_prior,1)
*     
      endif
      enddo      
*
*
*     / Replacement of baseline levels to zero
*
      if( input_zero_baseline. EQ. 1 ) then
*
      do i = 1, num_Reg_Modules
      if( strModules(i,2). EQ. tag_base ) then 
      id_prm = List_Reg_Modules(i,3)
      do j = 1, num_Para2
      if( ID_Prior_ParaModules(j). EQ. id_prm ) then
      Prior_ParaModules(j,1) = 0.0d0
      Prior_ParaModules(j,2) = 0.001d0
      endif
      enddo
      endif
      enddo
*      
      endif
*      
*	
	return
	end
*
*
      subroutine Shift_Prior_Elements_SpecficRegions(
*
     $      Prior_ParaElements, ID_Prior_ParaElements, 
     $      max_num_Para1, num_Para1,
     $      Para_Reg_Elements, List_Reg_Elements,  
     $      max_num_Reg_Elements, num_Reg_Elements,
     $      bound_model, num_Variable, max_num_Variable,
     $      discount_switch
*                  
     $  )
*
      implicit real*8 (a-h, o-z)
*
*     @ (1) Mean of parameter 1 (M-M constant), (2) mean of parameter 2 (power),
*       (3) Variance of parameter 1, and (4) Variance of parameter 2 	
	dimension Prior_ParaElements( max_num_Para1, * )
*
*     @	'Prior_ParaElements' is mapped to 'Reg_Elements' (IDs saved at the first column)
	dimension ID_Prior_ParaElements( max_num_Para1 )	
*
*	@ Michaelis-Menten-like kinetic values and steepness parameters of hill curve 
*       functions are saved in the 1st and 2nd columns	
	dimension Para_Reg_Elements( max_num_Reg_Elements, * )
*
*	@ (1) Element IDs, (2) kinetic law IDs, and (3) reactant state variables
	dimension List_Reg_Elements( max_num_Reg_Elements, * )
*
	dimension bound_model( max_num_Variable, * )
*
!$omp parallel private(k,ivari)
*
!$omp do
*
      do i = 1, num_Para1
      k = ID_Prior_ParaElements(i)
      do j = 1, num_Reg_Elements
      if( List_Reg_Elements(j,1). EQ. k ) then
*
      ivari = List_Reg_Elements(j,3)
*
      if( List_Reg_Elements(j,2). EQ. 18 )
     $      Prior_ParaElements(i,1) = discount_switch * 
     $      bound_model( ivari, 1 )
*
      if( List_Reg_Elements(j,2). EQ. 19 )
     $      Prior_ParaElements(i,1) = ( 1.0d0 - discount_switch ) * 
     $      bound_model( ivari, 1 )     
*
      endif
*      
      enddo
      enddo
*
!$omp end do
*
!$omp end parallel
*
      return
      end
*
*
      subroutine Shift_Prior_Inits(
*      
     $      Prior_Inits, ID_Prior_Inits,
     $      itranslation, icomplex, value_maxminData, 
     $      ave_varData, RangeHidden,
     $      Name_Variable, max_num_Variable, num_Variable,
     $      Data, max_time_Data, time_Data, 
     $      ID_MissVariable, ID_Variable,
     $      imRNA, ix, bound,
     $      network, network_attribute, max_num_edge, num_edge  
*
     $ )
*
      implicit real*8(a-h,o-z)
*
	integer*2 tab /2313/
	integer*4 time_StateVariable
	parameter ( iprint = 1 )	
	character Name_Variable( max_num_Variable ) * 30	
	dimension Prior_Inits( max_num_Variable, 2 )      
	dimension ID_Prior_Inits( max_num_Variable )
      dimension RangeHidden(2)
      dimension imRNA( max_num_Variable )
      dimension itranslation( max_num_Variable )
      dimension icomplex( max_num_Variable )
      dimension ID_MissVariable( max_num_Variable )
	dimension ID_Variable( max_num_Variable )
	dimension Data( max_num_Variable, max_time_Data )
	dimension ave_varData( max_num_Variable, 2 )
	dimension value_maxminData( max_num_Variable, 2 )
	dimension pr( max_num_Variable )
	dimension item( max_num_Variable )
	dimension ismp(1)
	dimension bound( max_num_Variable, 2 )
*
	dimension network( max_num_edge, 2 )
	character network_attribute( max_num_edge, 5 ) * 30
	dimension ipair( 50 )
*
      do i = 1, num_Variable
      bound(i,1) = value_maxminData(i,1)
      bound(i,2) = value_maxminData(i,2)      
      enddo
*
      if( iprint. EQ. 1 ) then
      write(*,*) 
      write(*,'(a)') '-@[Shift_Prior_Inits] 
     $  Replacement of prior distributions on initial states'
      endif
*
      num_item = 0
      discount = 0.01d0
      do i = 1, num_Variable
      k = ID_Prior_Inits(i)
      if( ID_MissVariable(k). EQ. 1 ) then
      Prior_Inits(i,1) = Data(k,1)
      Prior_Inits(i,2) = ave_varData(k,2) * discount
      num_item = num_item + 1
      item(num_item) = k
      if( iprint. EQ. 1 ) write(*,'(a,a1,a)')
     $  '-Replace...', tab, Name_Variable(k) 
      endif      
      enddo
*
      do i = 1, num_item
      pr(i) = 1.0d0 / dble(num_item)
      enddo
*
      do i = 1, num_Variable
      k = ID_Prior_Inits(i)      
      if( ( ID_MissVariable(k). EQ. 0 ). AND.
     $    ( imRNA(k). NE. 0 )  ) then
    1 continue  
      call Resampling( pr, num_item, 
     $	max_num_Variable, ix, ismp, 1 )
      mRNA = ismp(1)
      mRNA = item(mRNA)
*      mRNA = 17
      if( mRNA. EQ. k ) goto 1
      Prior_Inits(i,1) = Data(mRNA,1)
      Prior_Inits(i,2) = ave_varData(mRNA,2) * discount
      if( iprint. EQ. 1 ) write(*,'(6(a,a1))')
     $  '-This variable is mRNA, but missing...', 
     $  tab, trim(Name_Variable(k)),
     $  tab, '... so alternated by', tab, trim(Name_Variable(mRNA))
      bound(k,1) = value_maxminData(mRNA,1) 
      bound(k,2) = value_maxminData(mRNA,2)        
      endif
      enddo
*
      if( iprint. EQ. 1 ) then
      write(*,*)
      write(*,'(a)') 
     $  '-@[Shift_Prior_Inits] Bounds for regions of mRNA variables' 
      do i = 1, num_Variable
      write(*,'(a,a,a1,f10.5,a1,f10.5)') 
     $  '-', Name_Variable(i), tab, bound(i,1), tab, bound(i,2)
      enddo
      endif
*
      if( iprint. EQ. 1 ) then
      write(*,*) 
      write(*,'(a)') '-@[Shift_Prior_Inits] 
     $  Replacement of prior distributions on monomer protein'
      endif
*      
      do i = 1, num_Variable
      k = ID_Prior_Inits(i)
      mRNA = itranslation(k)
      if( mRNA. NE. 0 ) then
*
      if( ID_MissVariable(mRNA). EQ. 0 ) then
      call Resampling( pr, num_item, 
     $	    max_num_Variable, ix, ismp, 1 )
      mRNA = ismp(1)
      mRNA = item(mRNA)
      endif
*      
      relative = 1.0d0 - ( bound(mRNA,1) - Data(mRNA,1) )
     $  / ( bound(mRNA,1) - bound(mRNA,2) )
      Prior_Inits(i,1) = relative * ( RangeHidden(1) - RangeHidden(2) )
      Prior_Inits(i,2) = ( RangeHidden(1) - RangeHidden(2) ) * discount
*
      if( iprint. EQ. 1 ) write(*,'(3(a,a1),f10.5)')
     $  '-This variable is monomer protein...', 
     $  tab, trim(Name_Variable(k)),
     $  tab, 'whose prior mean is repalced to', tab, 
     $  Prior_Inits(i,1)    
*            
      endif      
      enddo
*
      if( iprint. EQ. 1 ) then
      write(*,*) 
      write(*,'(a)') '-@[Shift_Prior_Inits] 
     $  Replacement of prior distributions on multimeric protein'
      endif
*
      do i = 1, num_Variable
      k = ID_Prior_Inits(i)
      if( icomplex(k). EQ. 1 ) then
      num_pair = 0
      do j = 1, num_edge
*      write(*,'(5(a,a1))') ((network_attribute(j,ii),tab), ii=1,5)
      if( ( network(j,1). EQ. k ). AND.  
     $    ( network_attribute(j,3). EQ. 'Binding' ) ) then
      num_pair = num_pair + 1
      ipair(num_pair) = network(j,2)
      endif
      enddo
*
      product = 1.0d0
      pmax = -10d30
      do ii = 1, num_pair
      j = ipair(ii)
      product = product * Prior_Inits(j,1)
      if( Prior_Inits(j,1). gt. pmax )
     $  pmax = Prior_Inits(j,1)
      enddo
*      
*      Prior_Inits(k,1) = product
      Prior_Inits(k,1) = pmax
      Prior_Inits(k,2) = discount * ( RangeHidden(1) - RangeHidden(2) )
*
      if( iprint. EQ. 1 ) write(*,'(a,a1,a,a1,a,a1,f10.5,5(a1,i3))')
     $  '-This variable is multimeric protein...', 
     $  tab, trim(Name_Variable(k)), tab, 'relapced by', tab, 
     $  Prior_Inits(k,1), ((tab,ipair(ii)),ii=1,num_pair)      
      endif
      enddo
*
      return
      end           
*
*
      subroutine Set_Priors2Model(
*      
     $      List_Reg_Modules, List_Reg_Elements, 
     $      Para_Reg_Modules, Para_Reg_Elements, 
     $      max_num_Reg_Modules, max_num_Reg_Elements, max_tmp,
     $      num_Reg_Modules, num_Reg_Elements, strModules, strElements,
     $      Prior_ParaModules, Prior_ParaElements, 
     $      ID_Prior_ParaModules, ID_Prior_ParaElements, 
     $      max_num_Para2, max_num_Para1, num_Para2, num_Para1,
     $      Name_Variable, max_num_Variable, num_Variable,
     $      Prior_Inits, ID_Prior_Inits,
     $      StateVariable, time_StateVariable, max_time_StateVariable,
     $      iprint
*
     $ )
*
      implicit real*8(a-h,o-z)
	integer*2 tab /2313/   
	integer*4 time_StateVariable   
      dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
      dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )	      
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )	
	character strModules( max_num_Reg_Modules, 3 ) * 30
	character strElements( max_num_Reg_Elements, 4 ) * 30	
	dimension Prior_ParaModules( max_num_Para2, 2 )
	dimension Prior_ParaElements( max_num_Para1, 4 )		
	dimension ID_Prior_ParaModules( max_num_Para2 )
	dimension ID_Prior_ParaElements( max_num_Para1 )
	character Name_Variable( max_num_Variable ) * 30	
	dimension Prior_Inits( max_num_Variable, 2 )      
	dimension ID_Prior_Inits( max_num_Variable )
	dimension StateVariable( max_num_Variable, * )	
*
      if( iprint. EQ. 1 ) then
      write(*,*) 
      write(*,'(a)') '-@[Set_Priors2Model] 
     $  Overwrite reaction kinetics on [ReactionModules]
     $  by the specified prior mean'
      endif
*
      do i = 1, num_Para2
      k = ID_Prior_ParaModules(i)
      if( iprint. EQ.1 ) write(*,'(a,a1,i3)') 'Prior@Modules', tab, k
      do j = 1, num_Reg_Modules
      if( List_Reg_Modules(j,3). EQ. k ) then
      Para_Reg_Modules(j,1) = Prior_ParaModules(i,1)
      iv = List_Reg_Modules(j,2)
      if( iprint. EQ. 1 ) write(*,'(a,a1,a,a1,a,a1,a)') 
     $      'Replace', tab, strModules(j,2), tab, 
     $      '@Product', tab, Name_Variable(iv)
      endif 
      enddo
      enddo
*
      if( iprint. EQ. 1 ) then
      write(*,*) 
      write(*,'(a)') '-@[Set_Priors2Model] 
     $  Overwrite reaction kinetics on [ReactionElements]
     $  by the specified prior mean'
      endif
*      
*      write(*,*) (strElements(1,i),i=1,4)      
*
      do i = 1, num_Para1
      k = ID_Prior_ParaElements(i)
      if( iprint. EQ.1 ) write(*,'(a,a1,i3)') 'Prior@Elements', tab, k
      do j = 1, num_Reg_Elements
      if( List_Reg_Elements(j,1). EQ. k ) then
      Para_Reg_Elements(j,1) = Prior_ParaElements(i,1)
      Para_Reg_Elements(j,2) = Prior_ParaElements(i,2)      
      if( iprint. EQ. 1 ) write(*,'(10(a,a1))') 
     $      'Replace', tab, strElements(j,3), tab,
     $      '@Product', tab, strElements(j,1), tab,
     $      '@Reactant',  tab, strElements(j,4)
      endif
      enddo
      enddo      
*
*
      if( iprint. EQ. 1 ) then
      write(*,*) 
      write(*,'(a)') '-@[Set_Priors2Model] 
     $  Overwrite Initial States
     $  by the specified prior mean'
      endif
*      
      do i = 1, num_Variable
      k = ID_Prior_Inits(i)
      StateVariable(k,1) = Prior_Inits(i,1)
      if( iprint. EQ. 1 ) write(*,'(a,a1,a,a1,f10.5)') 
     $      'Replace', tab, Name_Variable(k), tab,
     $      Prior_Inits(i,1)
      enddo 
*
      return
      end
*
*
      subroutine get_network(
*      
*		-- Reaction functions       
     $      List_Reg_Elements, Para_Reg_Elements,
     $		List_Reg_Modules, Para_Reg_Modules,
     $		num_Reg_Elements, num_Reg_Modules, 
     $		max_num_Reg_Elements, max_num_Reg_Modules, max_tmp,
     $      strModules, strElements,
*       -- Variables
     $		ID_Variable, Name_Variable, num_Variable, 
     $		max_num_Variable, suffix, 
     $      itranslation, icomplex,
*       -- Network
     $      network, network_attribute, num_edge, max_num_edge,
     $      network_rna, num_edge_rna
*
     $  )
*
      implicit real * 8 (a-h,o-z)
*
	integer*2  tab /2313/
	parameter ( iPrint = 1 )
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	character strElements( max_num_Reg_Elements, 4 ) * 30
	character strModules( max_num_Reg_Modules, 3 ) * 30
	character Name_Variable( max_num_Variable ) * 30	! Names of state variables
	dimension ID_Variable( max_num_Variable )			! State variable IDs
	dimension network( max_num_edge, 2 )
	character network_attribute( max_num_edge, 5 ) * 30
	dimension network_rna( max_num_edge, 4 )
      dimension itranslation( max_num_Variable )
      dimension icomplex( max_num_Variable )      
      character tag_base * (*)
      character tag_diff * (*)
      character tag_bind * (*) 
      character tag_trc * (*)  
      character tag_degRna * (*)
      character tag_degPro * (*)            
      parameter ( tag_base = 'Baseline' )
      parameter ( tag_diff = 'DifferentiationOverTime' )
      parameter ( tag_bind = 'Binding' )
      parameter ( tag_trc = 'Transcription' )
      parameter ( tag_degRna = 'RnaDegradation' )
      parameter ( tag_degPro = 'ProteinDegradation' )      
      character suffix * 30
      character output_file * 30
      character sign * 10
      character AorI * 10
*
      output_file = trim(suffix)//'_graph.log'
*
      if( iPrint. EQ. 1 ) then
      write(*,*)
      write(*,'(a)') 
     $ '-@[get_network] Grapical structure of input model'       
      endif
*
*
      num_edge = 0
      do i = 1, num_Variable
      do j = 1, num_Reg_Modules
*      
      if(
     $      ( List_Reg_Modules(j,2). EQ. i ). AND.
     $      ( trim( strModules(j,2) ). NE. tag_base ). AND.
     $      ( trim( strModules(j,2) ). NE. tag_diff ). AND.
     $      ( trim( strModules(j,2) ). NE. tag_degPro ). AND.
     $      ( trim( strModules(j,2) ). NE. tag_degRna )  
     $      ) then
*     
      num_elements = List_Reg_Modules(j,4)
      if( Para_Reg_Modules(j,2). GT. 0 ) sign = 'Positive'
      if( Para_Reg_Modules(j,2). LT. 0 ) sign = 'Negative'
*      
      do ii = 1, num_elements
      do jj = 1, num_Reg_Elements
*      
      if( List_Reg_Elements(jj,1). EQ. List_Reg_Modules(j,4+ii) ) then
*     / Transcrition activator or repressor?
      AorI = 'FALSE'
      if( List_Reg_Elements(jj,2). EQ. 18 ) AorI = 'Activator'
      if( List_Reg_Elements(jj,2). EQ. 19 ) AorI = 'Inhibitor'      
      num_edge = num_edge + 1
      id_pa = List_Reg_Elements(jj,3)
      network(num_edge,1) = i
      network(num_edge,2) = id_pa
      network_attribute(num_edge,1) = Name_Variable(i)
      network_attribute(num_edge,2) = Name_Variable(id_pa)
      network_attribute(num_edge,3) = strModules(j,2)
      network_attribute(num_edge,4) = sign
      network_attribute(num_edge,5) = AorI      
*
      if( iPrint. EQ. 1 ) then
      write(*,'(a,a1,a,a1,a,a1,a,a1,a)') 
     $  trim( Name_Variable(i) ), tab, trim( Name_Variable(id_pa) ),
     $  tab, trim( strModules(j,2) ), tab, trim( sign ), 
     $  tab, trim( AorI ) 
      endif
*      
      endif
*      
      enddo
      enddo
*
      endif
*
      enddo
      enddo
*
      open( 5, file = output_file )
      do i = 1, num_edge
      write(5,'(i,a1,i, 5(a1,a))') 
     $  network(i,1), tab, network(i,2),
     $  ( ( tab, network_attribute(i,j) ), j = 1,5 ) 
      enddo
      close(5)
*
*
*     / Graph reduced for varaibles with mRNA attribute
*
      if( iPrint. EQ. 1 ) then 
      write(*,*)
      write(*,'(a)') 
     $ '-@[get_network] Reduced graph over the
     $ variables with mRNA attribute'
      endif
*      
      num_edge_rna = 0
      do i = 1, num_edge
*
      kC = network(i,1) ! Child in original graph
      kP = network(i,2) ! Parent in original graph     
*
*     If the child node has mRNA attribute, ...  
      if( ( itranslation(kC). EQ. 0 ). AND. 
     $    ( network_attribute(i,3). EQ. tag_trc ) ) then
*
*     If the parent node has 'single' protein attribute, ...            
      if( ( itranslation(kP). NE. 0 ) ) then
      num_edge_rna = num_edge_rna + 1
      id_pa = itranslation(kP)
      network_rna(num_edge_rna,1) = kC
      network_rna(num_edge_rna,2) = id_pa ! mRNA of the protein 
      network_rna(num_edge_rna,4) = i
      if( network_attribute(i,5). EQ. 'Activator' )  
     $  network_rna(num_edge_rna,3) = 1
      if( network_attribute(i,5). EQ. 'Inhibitor' )  
     $  network_rna(num_edge_rna,3) = -1         
*
      if( iPrint. EQ. 1 ) 
     $  write(*,'(a,a1,a, 5(a1,i5))') 
     $  trim( Name_Variable(kC) ), tab, trim( Name_Variable(id_pa) ),
     $  ( ( tab, network_rna(num_edge_rna,j) ), j = 3,4 )
*
      endif
*
*     If the parent node has 'monomer' protein attribute, ...
      if( ( icomplex(kP). EQ. 1 ) ) then
      do j = 1, num_edge
      if( ( network(j,1). EQ. kP ). AND. 
     $  ( network_attribute(j,3). EQ. tag_bind ) ) then
      num_edge_rna = num_edge_rna + 1
      id_pa = network(j,2)
      id_pa = itranslation( id_pa )
      network_rna(num_edge_rna,1) = kC
      network_rna(num_edge_rna,2) = id_pa
      network_rna(num_edge_rna,4) = i      
      if( network_attribute(i,5). EQ. 'Activator' )  
     $  network_rna(num_edge_rna,3) = 1
      if( network_attribute(i,5). EQ. 'Inhibitor' )  
     $  network_rna(num_edge_rna,3) = -1               
*
      if( iPrint. EQ. 1 ) 
     $  write(*,'(a,a1,a,5(a1,i5))') 
     $  trim( Name_Variable(kC) ), tab, trim( Name_Variable(id_pa) ),
     $  ( ( tab, network_rna(num_edge_rna,ii) ), ii = 3,4 )           
      endif
      enddo
      endif
*
      endif
      enddo
*
      return
      end	
*
*
      subroutine covariance_matrix( cov, ave, x, n, p, mn, mp )
*
      implicit real*8 (a-h,o-z)
      integer p
      dimension cov( mp, mp ), ave( mp )
      dimension x( mp, mn )
*
!$omp parallel private(sum)
!$omp do 
      do i = 1, p
      sum = 0.0d0
      do j = 1, n
      sum = sum + x(i,j)
      enddo
      ave(i) = sum / dble(n)
      enddo
!$omp end do
!$omp end parallel
*
!$omp parallel private(sum)
!$omp do
      do i = 1, p
      do j = 1, p
      sum = 0.0d0
      do ii = 1, n
      sum = sum + (x(i,ii) - ave(i)) * (x(j,ii) - ave(j))
      enddo
      cov(i,j) = sum / dble(n)
      enddo
      enddo
!$omp end do
!$omp end parallel
*
      return
      end
*
*
      subroutine get_bound_variable(
*
     $      Para_Reg_Modules, List_Reg_Modules, strModules,
     $      max_num_Reg_Modules, num_Reg_Modules, max_tmp,
     $      Name_Variable, StateVariable, 
     $      max_time_StateVariable, max_num_Variable, num_Variable,
     $      itranslation, icomplex, imRNA, 
     $      value_maxminData, RangeHidden,     
     $      bound_data, bound_model
*
     $  )
*
      implicit real*8(a-h,o-z)
	integer*2 tab /2313/   
*	integer*4 time_StateVariable   
      parameter ( iprint = 1 )
      dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	character strModules( max_num_Reg_Modules, 3 ) * 30
	character Name_Variable( max_num_Variable ) * 30
	dimension StateVariable( max_num_Variable, * )
	dimension bound_data( max_num_Variable, 2 )
	dimension bound_model( max_num_Variable, 2 )
*
      dimension imRNA( max_num_Variable )
      dimension itranslation( max_num_Variable )
      dimension icomplex( max_num_Variable )
      dimension RangeHidden(2)
      dimension value_maxminData( max_num_Variable, 2 )	
*
*
      if( iprint. EQ. 1 ) then
      write(*,*)
      write(*,'(a)') 
     $ '-@[get_bound_variable] Upper/Lower bounds of
     $ the endogenous variables defined by the specified model'       
      endif
*      
*     / Initialization
	call ArrInit( bound_model, max_num_Variable, 2, 0.0d0 )
*
*
*     / Upper/Lower bounds for mRNA variables
*
      if( iprint.EQ.1 ) write(*,'(a)') '-@mRNA attribute' 
*
      do i = 1, num_Variable
      if( imRNA(i). NE. 0 ) then
      deg = 0.0d0
      syn = 0.0d0     
      do j = 1, num_Reg_Modules
      if( List_Reg_Modules(j,2). EQ. i ) then
      if( strModules(j,2). EQ. 'RnaDegradation' ) 
     $      deg = Para_Reg_Modules(j,1)
      if( strModules(j,2). EQ. 'Transcription' ) 
     $      syn = Para_Reg_Modules(j,1)     
      endif
      enddo
      bound_model(i,1) = syn/deg
      bound_model(i,2) = 0.0d0
      if( iprint. EQ. 1 ) write(*,'(a,3(a1,f10.5))')
     $  Name_Variable(i), ( ( tab, bound_model(i,j) ), j=1,2 ),
     $  tab, value_maxminData(i,1)
      endif
      enddo
*
*
*     / Upper/Lower bounds for monomer protein variables
*
      if( iprint.EQ.1 ) write(*,'(a)') '-@Monomer protein attribute' 
*
      do i = 1, num_Variable
      if( itranslation(i). NE. 0 ) then
      irna = itranslation(i)
      deg = 0.0d0
      syn = 0.0d0
      do j = 1, num_Reg_Modules
      if( List_Reg_Modules(j,2). EQ. i ) then
      if( strModules(j,2). EQ. 'ProteinDegradation' ) 
     $  deg = Para_Reg_Modules(j,1)
      if( strModules(j,2). EQ. 'Translation' ) 
     $  syn = Para_Reg_Modules(j,1) * bound_data(irna,1)     
      endif
      enddo
*      write(*,*) Name_Variable(irna), syn, deg, bound_data(irna,1)
      bound_model(i,1) = syn/deg
      bound_model(i,2) = 0.0d0
      if( iprint. EQ. 1 ) write(*,'(a,3(a1,f10.5))')
     $  Name_Variable(i), ((tab,bound_model(i,j)),j=1,2),
     $  tab, RangeHidden(1)
      endif
      enddo
*
*
*     / Upper/Lower bounds for multimeric complex variables
*
      if( iprint.EQ.1 ) write(*,'(a)') '-@Multimeric complex attribute' 
*      
      do i = 1, num_Variable
      if( icomplex(i). EQ. 1 ) then
      deg = 0.0d0
      syn = 0.0d0
      do j = 1, num_Reg_Modules
      if( List_Reg_Modules(j,2). EQ. i ) then
      if( strModules(j,2). EQ. 'ProteinDegradation' ) 
     $  deg = Para_Reg_Modules(j,1)
      if( strModules(j,2). EQ. 'Binding' ) 
     $  syn = Para_Reg_Modules(j,1)     
      endif
      enddo
*      write(*,*) Name_Variable(irna), syn, deg, bound_data(irna,1)
      bound_model(i,1) = syn/deg
      bound_model(i,2) = 0.0d0
      if( iprint. EQ. 1 ) write(*,'(a,3(a1,f10.5))')
     $  Name_Variable(i), ((tab,bound_model(i,j)),j=1,2),
     $  tab, RangeHidden(1)
      endif
      enddo      
*
*
*     / Initial states
*
      if( iprint.EQ.1 ) then
      write(*,*)
      write(*,'(a)') '-@Initial states'
*      
      do i = 1, num_Variable
      if( itranslation(i). NE. 0 ) then
      write(*,*) 
      write(*,'(a,a1,f10.5)' ) 
     $ trim(Name_Variable(i)), tab, StateVariable(i,1)
      write(*,'(a1,a,a1,f10.5,a1,a,a1,f10.5)') 
     $ tab, 'Max', tab, RangeHidden(1), tab, 'Min', tab, RangeHidden(2)
      write(*,'(a1,a,a1,f10.5,a1,a,a1,f10.5)') 
     $ tab,'Relative Level', tab, 
     $  StateVariable(i,1) / ( RangeHidden(1) - RangeHidden(2) )     
*
      mrna = itranslation(i)
      write(*,'(a,a1,f10.5)' ) 
     $ trim(Name_Variable(mrna)), tab, StateVariable(mrna,1)
      write(*,'(a1,a,a1,f10.5,a1,a,a1,f10.5)') 
     $  tab, 'Max', tab, value_maxminData(mrna,1), tab, 
     $  'Min', tab, value_maxminData(mrna,2)
      relative = 1-( value_maxminData(mrna,1) - StateVariable(mrna,1) )/ 
     $  ( value_maxminData(mrna,1) - value_maxminData(mrna,2) )   
      write(*,'(a1,a,a1,f10.5,a1,a,a1,f10.5)') 
     $ tab, 'Relative Level', tab, relative
      endif
      enddo
*
      endif
*
	return
	end
*
*
      subroutine get_bound_variable_inLoop(
*
     $      Para_Reg_Modules, List_Reg_Modules, strModules,
     $      max_num_Reg_Modules, num_Reg_Modules, max_tmp,
     $      Name_Variable, max_num_Variable, num_Variable,
     $      itranslation, icomplex, imRNA, 
     $      value_maxminData, RangeHidden,     
     $      bound_data, bound_model, ix
*
     $  )
*
      implicit real*8(a-h,o-z)
	integer*2 tab /2313/   
	integer*4 time_StateVariable   
      parameter ( iprint = 0 )
      parameter ( bmin = 0.1d0, bmax = 0.5d0, cmin = 0.5d0 )
      dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	character strModules( max_num_Reg_Modules, 3 ) * 30
	character Name_Variable( max_num_Variable ) * 30
	dimension bound_data( max_num_Variable, 2 )
	dimension bound_model( max_num_Variable, 2 )
*
      dimension imRNA( max_num_Variable )
      dimension itranslation( max_num_Variable )
      dimension icomplex( max_num_Variable )
      dimension RangeHidden(2)
      dimension value_maxminData( max_num_Variable, 2 )	
*
*
      if( iprint. EQ. 1 ) then
      write(*,*)
      write(*,'(a)') 
     $ '-@[get_bound_variable_inLoop] Upper/Lower bounds of
     $ the endogenous variables defined by the specified model'       
      endif
*      
*     / Initialization
	call ArrInit( bound_model, max_num_Variable, 2, 0.0d0 )
*
*
*     / Upper/Lower bounds for mRNA variables
*
      if( iprint.EQ.1 ) write(*,'(a)') '-@mRNA attribute' 
*
      do i = 1, num_Variable
      if( imRNA(i). NE. 0 ) then
      deg = 0.0d0
      syn = 0.0d0     
      do j = 1, num_Reg_Modules
      if( List_Reg_Modules(j,2). EQ. i ) then
      if( strModules(j,2). EQ. 'RnaDegradation' ) 
     $      deg = Para_Reg_Modules(j,1)
      if( strModules(j,2). EQ. 'Transcription' ) 
     $      syn = Para_Reg_Modules(j,1)     
      endif
      enddo
      bound_model(i,1) = syn/deg
      bound_model(i,2) = 0.0d0
      if( iprint. EQ. 1 ) write(*,'(a,6(a1,f10.5))')
     $  Name_Variable(i), ( ( tab, bound_model(i,j) ), j=1,2 ),
     $  ( ( tab, bound_data(i,j) ), j=1,2 ), tab, syn, tab, deg
      endif
      enddo
*
*
*     / Upper/Lower bounds for monomer protein variables
*
      if( iprint.EQ.1 ) write(*,'(a)') '-@Monomer protein attribute' 
*
      do i = 1, num_Variable
      if( itranslation(i). NE. 0 ) then
      irna = itranslation(i)
      deg = 0.0d0
      syn = 0.0d0
      bnd = 0.0d0
      do j = 1, num_Reg_Modules
      if( List_Reg_Modules(j,2). EQ. i ) then
      if( strModules(j,2). EQ. 'ProteinDegradation' ) 
     $  deg = Para_Reg_Modules(j,1)
      if( strModules(j,2). EQ. 'Translation' ) 
     $  syn = Para_Reg_Modules(j,1) * bound_data(irna,1)
      if( strModules(j,2). EQ. 'Binding' ) 
     $  bnd = Para_Reg_Modules(j,1) 
      endif
      enddo
*      if( (deg+bnd). lt. 0.0d0 ) write(*,*) 
*     $  '---Warning() -get_bound_variable_inLoop -Binding constant' 
*      write(*,*) Name_Variable(irna), syn, deg, bound_data(irna,1)
      u = bmin + runif(ix) * (bmax - bmin)
      bound_model(i,1) = syn/(deg+bnd*u)
      bound_model(i,2) = 0.0d0
      if( iprint. EQ. 1 ) write(*,'(a,3(a1,f10.5))')
     $  Name_Variable(i), ((tab,bound_model(i,j)),j=1,2),
     $  tab, RangeHidden(1)
      endif
      enddo
*
*
*     / Upper/Lower bounds for multimeric complex variables
*
      if( iprint.EQ.1 ) write(*,'(a)') '-@Multimeric complex attribute' 
*      
      do i = 1, num_Variable
      if( icomplex(i). EQ. 1 ) then
      deg = 0.0d0
      syn = 0.0d0
      do j = 1, num_Reg_Modules
      if( List_Reg_Modules(j,2). EQ. i ) then
      if( strModules(j,2). EQ. 'ProteinDegradation' ) 
     $  deg = Para_Reg_Modules(j,1)
      if( strModules(j,2). EQ. 'Binding' ) 
     $  syn = Para_Reg_Modules(j,1)     
      endif
      enddo
      u = cmin + runif(ix) * ( RangeHidden(1) - cmin)      
*      write(*,*) Name_Variable(irna), syn, deg, bound_data(irna,1)
      bound_model(i,1) = syn*u/deg
      bound_model(i,2) = 0.0d0
      if( iprint. EQ. 1 ) write(*,'(a,3(a1,f10.5))')
     $  Name_Variable(i), ((tab,bound_model(i,j)),j=1,2),
     $  tab, RangeHidden(1)
      endif
      enddo
*	
	return
	end
*
*
      subroutine renumbering_reactions(
*	 
*		-- Reaction functions 
     $		List_Reg_Elements, Para_Reg_Elements,
     $		List_Reg_Modules, Para_Reg_Modules,
     $		num_Reg_Elements, num_Reg_Modules, 
     $		max_num_Reg_Elements, max_num_Reg_Modules, max_tmp, 
*
*		-- Variables
     $		ID_Variable, Name_Variable, num_Variable, 
     $		max_num_Variable,
*      
*		-- Prior distributions        
     $		Prior_ParaElements, ID_Prior_ParaElements, 
     $		Prior_ParaModules, ID_Prior_ParaModules,
     $		Prior_Inits, ID_Prior_Inits, 
     $		num_Para1, num_Para2, max_num_Para1, max_num_Para2,
*
     $		StateVariable, max_time_StateVariable, time_StateVariable,
     $		Data, max_time_Data, time_Data,     
*
*		-- Variances
     $		Variance_System, Variance_Measure,
     $		strElements, strModules, strParaElements, strParaModules
*      
     $ )
*
      implicit real*8(a-h,o-z)
      integer time_StateVariable, time_Data
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension Prior_ParaElements( max_num_Para1, 4 )
	dimension Prior_ParaModules( max_num_Para2, 2 )
	dimension Prior_Inits( max_num_Variable, 2 )
	dimension ID_Prior_ParaElements( max_num_Para1 )
	dimension ID_Prior_ParaModules( max_num_Para2 )
	dimension ID_Prior_Inits( max_num_Variable )
	dimension Variance_Measure( max_num_Variable )
	dimension Variance_System( max_num_Variable )
	character strElements( max_num_Reg_Elements, 4 ) * 30
	character strModules( max_num_Reg_Modules, 3 ) * 30
	character strParaElements( max_num_Para1, 4  )  * 30
	character strParaModules( max_num_Para2 ) * 50
	character Name_Variable( max_num_Variable ) * 30
	dimension ID_Variable( max_num_Variable )
	dimension Data( max_num_Variable, max_time_Data )
	dimension StateVariable( max_time_StateVariable, max_num_Variable )
*
      dimension number_variable( num_Variable )
      dimension number_modules( num_Reg_Modules )
      dimension number_elements( num_Reg_Elements )
*
      dimension number_variable_o( num_Variable )
      dimension number_modules_o( num_Reg_Modules )
      dimension number_elements_o( num_Reg_Elements )
*
      parameter ( max_ncol = 500, max_nrow = 500 )
      dimension temp_real( max_nrow, max_ncol )
      integer temp_int( max_nrow, max_ncol )
      integer temp_int_vec( max_nrow )
      character temp_char( max_nrow, max_ncol )*30
      dimension icol_replace(10)      
*
      call VecInit_Integer( number_variable, num_Variable, 0 )
      call VecInit_Integer( number_modules, num_Reg_Modules, 0 )
      call VecInit_Integer( number_elements, num_Reg_Elements, 0 )
*
      call VecInit_Integer( number_variable_o, num_Variable, 0 )
      call VecInit_Integer( number_modules_o, num_Reg_Modules, 0 )
      call VecInit_Integer( number_elements_o, num_Reg_Elements, 0 )
*
      do i = 1, num_Variable
      number_variable(i) = ID_Variable(i)
      enddo
*
*
      call order_increase( 
     $      number_variable, number_variable_o, num_Variable )
      call rank_increase( 
     $      number_variable, number_variable_o, num_Variable )     
*
*
      call order_matrix_real_row( Data, time_Data, num_Variable, 
     $        max_time_Data, max_num_Variable, number_variable_o )     
*
      call order_matrix_real_col(
     $     StateVariable, time_StateVariable, num_Variable, 
     $     max_time_StateVariable, max_num_Variable, number_variable_o )
*
      call order_vector_integer(
     $  ID_Variable, num_Variable, max_num_Variable, number_variable_o )
*
      call order_vector_char(
     $ Name_Variable, num_Variable, max_num_Variable, number_variable_o)     
*
      call ArrInit( temp_real, max_ncol, max_nrow, 0.0d0 )
      call ArrInit_Integer( temp_int, max_ncol, max_nrow, 0 )
*
      do k = 1, num_Variable
      do i = 1, num_Reg_Modules
      iord = number_variable_o(k)
*     / Product ID
      if( List_Reg_Modules(i,2). EQ. k ) 
     $      List_Reg_Modules(i,2) = iord
      enddo
      enddo
*
      do k = 1, num_Variable
      do i = 1, num_Reg_Elements
      iord = number_variable_o(k)
*     / Reactant variable IDs     
      if( List_Reg_Elements(i,3). EQ. k ) 
     $      List_Reg_Elements(i,3) = iord
      enddo
      enddo
*      
      do i = 1, num_Varible
      k = ID_Prior_Inits(i)
      do j = 1, 2
      temp_real(k,j) = Prior_Inits(i,j) 
      enddo
      enddo
*
      do i = 1, num_Variable
      do j = 1, 2
      Prior_Inits(i,j) = temp_real(i,j) 
      enddo
      enddo
*
*            
      return
      end
*
*
      subroutine Reference_Profiles(
*
     $      StateVariable, max_num_Variable, num_Variable, 
     $      max_time_StateVariable, time_StateVariable, 
     $      Name_Variable, ID_MissVariable,
     $      network, network_attribute, itranslation, icomplex,
     $      num_edge, max_num_edge, ix, RefState
* 
     $  ) 
*
      implicit real*8(a-h,o-z)
      integer*4 time_StateVariable
      dimension StateVariable( max_num_Variable,*)     
	dimension network( max_num_edge, 2 )
	character network_attribute( max_num_edge, 5 ) * 30
	dimension network_rna( max_num_edge, 4 )
      dimension itranslation( max_num_Variable )
      dimension icomplex( max_num_Variable )
      dimension iset_bind( 50 )
      character tag_bind * (*)      
      parameter ( tag_bind = 'Binding' )
      character Name_Variable( max_num_Variable ) * 30
	dimension ID_MissVariable( max_num_Variable )     
      character Output_File * 30
      dimension temp( time_StateVariable )
      dimension pr( num_Variable )
      dimension item( num_Variable )
	dimension RefState( max_num_Variable,* )      
*
!omp parallel
*
!omp do
      do i = 1, num_Variable
      pr(i) = 1.0d0 / dble(num_Variable)
      enddo
!omp enddo      
*
      do i = 1, num_Variable
      if( ID_MissVariable(i). EQ. 0 ) then
    1 continue
 	call Resampling( 
     $	    pr, num_Variable, num_Variable, ix, item, 1 
     $  )
      irna = item(1)
      if( ID_MissVariable(irna). EQ. 0 ) goto 1
!omp do       
      do j = 1, time_StateVariable
      StateVariable(i,j) = StateVariable(irna,j)
      enddo
!omp end do      
      endif
      enddo
*
*
!omp do private(irna,temp,vmax,vmin,imax,imin)
      do i = 1, num_Variable
      irna = itranslation(i)
      if( irna. NE. 0 ) then
      do j = 1, time_StateVariable
      StateVariable(i,j) = StateVariable(irna,j)
      temp(j) = StateVariable(i,j)
      enddo
      call get_max(temp, time_StateVariable, vmax, imax)
      call get_min(temp, time_StateVariable, vmin, imin)      
      do j = 1, time_StateVariable
      StateVariable(i,j) = ( StateVariable(i,j) - vmin ) 
     $  / ( vmax - vmin ) + 1.0d-4
      enddo
      endif
      enddo
!omp end do 
*
*
!omp do private(num_bind, iset_bind, prod, k)
      do i = 1, num_Variable
      if( icomplex(i). NE. 0 ) then
      num_bind = 0
      do j = 1, num_edge
      if( ( network(j,1). EQ. i ). AND. 
     $    ( network_attribute(j,3). EQ. tag_bind ) ) then
      num_bind = num_bind + 1
      iset_bind( num_bind ) = network(j,2)
      endif
      enddo
      do j = 1, time_StateVariable
      prod = 1.0d0
      do ii = 1, num_bind
      k = iset_bind(ii)
      prod = prod * StateVariable(k,j)
      enddo
      StateVariable(i,j) = prod
      enddo
      endif
      enddo
!omp end do 
*
!omp end parallel
*
	Output_File = 'reference_profile.log'
*	
	call Output_BioSimulator(
*    	
     $		StateVariable, num_Variable, time_StateVariable,  
     $		max_time_StateVariable, max_num_Variable,
     $		Name_Variable,
     $		Output_File
*      
     $	)
*
      do i = 1, num_Variable
      do j = 1, time_StateVariable 
      RefState(i,j) = StateVariable(i,j)
      enddo
      enddo
*
      return
      end
*
*
      subroutine Output_M2Constant(
*      
     $      Para_Reg_Modules, List_Reg_Modules, strModules,
     $      Para_Reg_Elements, List_Reg_Elements, strElements,
     $      max_num_Reg_Modules, num_Reg_Modules, max_tmp, 
     $      max_num_Reg_Elements, num_Reg_Elements,     
     $      imRNA, max_num_Variable, num_Variable,
     $      num_transcript, max_num_transcript, switch, itranscript,
     $      List_Reg_Elements_switch, Output_File
*     
     $  )
*
      implicit real*8(a-h,o-z)
	integer*2  tab /2313/      
*
*	@ Michaelis-Menten-like kinetic values and steepness parameters of hill curve 
*       functions are saved in the 1st and 2nd columns      
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
*	
*     @ Coefficients and their signs are saved in the 1st and 2nd columns
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )	
*
*	@ (1) Element IDs, (2) kinetic law IDs, and (3) reactant state variables
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
*
*	@ (1) Module IDs, (2) product variables, (3) parameter ID, (4) number of reactant variables 
*       (5-Last) Element IDs associated to the reaction module
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
*
	character strElements( max_num_Reg_Elements, 4 ) * 30
	character strModules( max_num_Reg_Modules, 3 ) * 30   ! (1) Product, (2) Reaction, (3) Reactants	
*
      dimension imRNA( max_num_Variable )
*
	character tag_synth * (*)
	parameter ( tag_synth = 'Transcription' )
*
      dimension itranscript( max_num_transcript, 3 )
      dimension switch( max_num_transcript )
      dimension List_Reg_Elements_switch( max_num_transcript )      
*
      character Output_File * (*)      
*
      num_transcript = 0
      do i = 1, num_Variable  !(1)
*      write(*,*) imRNA(i)
      if( imRNA(i). NE. 0 ) then   !(2)
      do j = 1, num_Reg_Modules    !(3)
      if( ( List_Reg_Modules(j,2). EQ. i ). AND.  
     $    ( strModules(j,2). EQ. tag_synth ) ) then !(4)
      do ii = 1, List_Reg_Modules(j,4) !(5)
      num_transcript = num_transcript + 1
      itranscript(num_transcript,1) = List_Reg_Modules(j,2)
      ide = List_Reg_Modules(j,4+ii)
      do jj = 1, num_Reg_Elements   !(6)
      if( List_Reg_Elements(jj,1). EQ. ide ) then  !(7)
      itranscript(num_transcript,2) = List_Reg_Elements(jj,3)
      switch(num_transcript) = Para_Reg_Elements(jj,1)
      List_Reg_Elements_switch(num_transcript) 
     $   = List_Reg_Elements(jj,1)      
      if( List_Reg_Elements(jj,2). EQ. 18 ) 
     $  itranscript(num_transcript,3) = 1
      if( List_Reg_Elements(jj,2). EQ. 19 ) 
     $  itranscript(num_transcript,3) = -1          
      endif !(7)
      enddo !(6)
      enddo !(5)
      endif !(4)
      enddo !(3)
      endif !(2)
      enddo !(1)
*
	open( 5, file = Output_File )
	do i = 1, num_transcript
	write( 5, '( i4, a1, i4, a1, i4, a1, i4, a1, f15.5)' ) 
     $	itranscript(i,1), tab, itranscript(i,2), tab, 
     $  itranscript(i,3), tab, List_Reg_Elements_switch(i), 
     $  tab, switch(i)
	enddo
	close(5)
*
      return
      end
*
*
      subroutine make_binary_array( a, binary, ncol, nrow, mrow, thr )
*
      implicit real*8 (a-h, o-z)
*      
      dimension a(mrow,*)
      integer binary(mrow,*)
*      
      do i = 1, nrow
      do j = 1, ncol
      if( a(i,j).ge.thr ) then 
      binary(i,j) = 1
      else
      binary(i,j) = 0
      endif
      enddo
      enddo
*
      return      
      end      
*
*
      subroutine make_binary_vector( a, binary, n, thr )
*
      implicit real*8 (a-h, o-z)
*      
      dimension a(*)
      integer binary(*)
      
      do i = 1, n
      if( a(i).ge.thr ) then 
      binary(i) = 1
      else
      binary(i) = 0
      endif
      enddo
*
      return      
      end
*
*
      subroutine gradient_state(
*
     $      StateVariable, bn_gradient, 
     $      time_StateVariable, num_Variable, 
     $      max_time_StateVariable, max_num_Variable
*     
     $  )
*
      implicit real*8 (a-h,o-z)
	integer*4 time_StateVariable  ! Number of simulation time steps           
	dimension 
     $	StateVariable( max_num_Variable, * )
	integer bn_gradient( max_num_Variable, max_time_StateVariable )
*
	do i = 1, num_Variable
	do j = 1, time_StateVariable
	if( j.eq.1 ) then 
	bn_gradient(i,j) = 0
	else
	a = StateVariable(i,j) - StateVariable(i,j-1)
	if( a.ge.0.0d0 ) bn_gradient(i,j) = 1
	if( a.lt.0.0d0 ) bn_gradient(i,j) = 0
	endif
	enddo
	enddo
*
      return
      end
*
*
      subroutine Get_Switch(
     $      s, bnGrad, nvari, nt, mvari, mt, id,
     $      itranscript, switch, num_transcript, 
     $      max_num_transcript, RefState     
     $     )
*
      implicit real*8(a-h, o-z)
*
	integer*2  tab /2313/
*
      parameter ( ngrid = 30 )
      integer bnGrad( mvari, mt )
      dimension s( mvari, mt )
	dimension RefStates( mvari, mt )        
      dimension grid( mvari, ngrid )
      dimension a(nt)
      integer bn(nt)
*
	dimension avevar( mvari, 2 )      ! Mean and variance of time course data
	dimension bounds( mvari, 2 )      ! Max and Min of time course data
	dimension lbounds( mvari, 2 )     ! Time points in which the max or min data value appears      
*      
      dimension itranscript( max_num_transcript, 3 )
      dimension switch( max_num_transcript )
*
      character outfile * 30      
*
      lag_trim = 20
      is = lag_trim         ! Starting time points
      ie = nt - lag_trim    ! End time points
*
*     Get max and min state variables
      call SummaryStatistics(
     $	    s, nvari, nt, mvari, mt, 
     $	    avevar, bounds, lbounds
     $	)
*
*     Search grids for the determination of transcription threshold      
      do i = 1, nvari
      r = (bounds(i,1) - bounds(i,2)) / (dble(ngrid) - 1.0d0)
      do j = 1, ngrid
      grid(i,j) = bounds(i,2) + dble(j-1) * r
      enddo
      enddo
*
*     For each pair of variables, search for a firing threshold... 
      do i = 1, num_transcript
*
      j = itranscript(i,1)  ! mRNA
      k = itranscript(i,2)  ! Regulator
*      
      do ii = is, ie
      a(ii) = s(k,ii)
      enddo
*
      min_grid = 10**6
      max_grid = -10**6 
*                 
      do ii = 1, ngrid
*
      thr = grid(k,ii)
      call make_binary_vector( a, bn, nt, thr )
*
      isum = 0
      do jj = is, ie
      isum = isum + iabs( bnGrad(j,jj) - bn(jj) )
      enddo
*
      if( isum.le.min_grid ) then
      min_grid = isum
      thr_min = thr
      endif
*
      if( isum.ge.max_grid ) then
      max_grid = isum
      thr_max = thr
      endif      
*      
      enddo
*
      if( itranscript(i,3).eq.-1 ) 
     $  switch(i) = thr_max
*
      if( itranscript(i,3).eq.1 ) 
     $  switch(i) = thr_min
*            
      enddo
*
      return
      end      		 
*
*
      subroutine sampling_priors(
*      
     $      Prior_ParaElements, Prior_ParaModules, Prior_Inits, 
     $      ID_Prior_ParaElements, ID_Prior_ParaModules, 
     $      ID_Prior_Inits,
     $      PriorG0_ParaElements, PriorG0_ParaModules, 
     $      PriorG0_Inits,
     $      max_num_Para1, max_num_Para2, num_Para1, num_Para2,
     $      max_num_Variable, num_Variable,
     $      max_num_Reg_Elements, max_num_Reg_Modules,
     $      num_Reg_Elements, num_Reg_Modules,
     $      ix,
     $      Particle_ParaElements, Particle_ParaModules,
     $      Particle_Inits, nsave_Particle,
     $      List_Reg_Elements, List_Reg_Modules, strModules,
     $      ir, method
*
     $ )
*
      implicit real*8(a-h,o-z)
	character tag_base * (*)      
	parameter ( tag_base = 'Baseline' )
*	      
	character strModules( max_num_Reg_Modules, 3 ) * (*)      
*	@ (1) Element IDs, (2) kinetic law IDs, and (3) reactant state variables
	dimension List_Reg_Elements( max_num_Reg_Elements, * )
*
*	@ (1) Module IDs, (2) product variables, (3) parameter ID, (4) number of reactant variables 
*       (5-Last) Element IDs associated to the reaction module
	dimension List_Reg_Modules( max_num_Reg_Modules, * )
*
*	/ Maintained particles for the parameters in "Reaction Elements", 
*       "Reaction Modules", and "Initial States"
	dimension Particle_ParaElements( nsave_Particle, 
     $	                max_num_Reg_Elements, 2 )
	dimension Particle_ParaModules( nsave_Particle, max_num_Reg_Modules )
	dimension Particle_Inits( nsave_Particle, max_num_Variable )
*	
*     @ (1) Mean of parameter 1 (M-M constant), (2) mean of parameter 2 (power),
*       (3) Variance of parameter 1, and (4) Variance of parameter 2 
	dimension Prior_ParaElements( max_num_Para1, 4 )
*
*	@ (1) Mean, (2) variance
	dimension Prior_ParaModules( max_num_Para2, 2 )
*
*     @ (1) Mean, (2) Variance	
	dimension Prior_Inits( max_num_Variable, 2 )
*
*     @	'Prior_ParaElements' is mapped to 'Reg_Elements' (IDs saved at the first column)
	dimension ID_Prior_ParaElements( max_num_Para1 )
*
*     @ 'Prior_ParaModules' is mapped to 'Reg_Modules' (IDs saved at the third column)
	dimension ID_Prior_ParaModules( max_num_Para2 )
*
*     @ 'Prior_Inits' is mapped to 'StateVariable'	
	dimension ID_Prior_Inits( max_num_Variable )
*	
*     / Parameters of the base measure of DP for "Reaction Elements", 
*       "Reaction Modules", and "Initial States" 
	dimension PriorG0_ParaElements( max_num_Para1, 4 )
	dimension PriorG0_ParaModules( max_num_Para2, 2 )
	dimension PriorG0_Inits( max_num_Variable, 2 )
	dimension num_overlap_DP( nsave_Particle )
*
      if( method.eq.1 ) then
      w1 = 0.8d0
      s = 0.01d0
      endif
*      
      if( method.eq.2 ) then
      w1 = 1.0d0
      s = 0.0d0
      endif
*      
      w2 = 1.0d0 - w1
*
!$omp parallel firstprivate(k,ix)
*      
!$omp do
	do i = 1, num_Para1
	k = ID_Prior_ParaElements(i)
	do j = 1, num_Reg_Elements
	if( List_Reg_Elements(j,1). EQ. k ) then
   10 continue	
     	Prior_ParaElements(i,1) = w1*Particle_ParaElements(ir,j,1) 
     $	    + w2*PriorG0_ParaElements(i,1) 
     $      + s * dsqrt( PriorG0_ParaElements(i,3) ) * rgauss(ix)
      if( (runif(ix).ge.0.9d0). and. (method.eq.1) ) then
      Prior_ParaElements(i,1) = 0.1d0 + runif(ix) * 1.0d0  
      endif
      if( Prior_ParaElements(i,1).lt. 0.0d0 ) goto 10
     	Prior_ParaElements(i,2) = w1*Particle_ParaElements(ir,j,2)
     $	    + w2*PriorG0_ParaElements(i,2) 
     $      + s * dsqrt( PriorG0_ParaElements(i,4) ) * rgauss(ix)
      if( Prior_ParaElements(i,2).lt. 0.0d0 ) goto 10     
*	write(*,'(a,i3,e14.5,e14.5,e14.5)')
*     $	'ele', i, Prior_ParaElements(i,1),
*     $	PriorG0_ParaElements(i,1), Particle_ParaElements(ir,j,1)
	goto 1
	endif
	enddo
    1 continue
	enddo
!$omp end do	
*
*
!$omp do
	do i = 1, num_Para2
	k = ID_Prior_ParaModules(i)
	do j = 1, num_Reg_Modules
	if( List_Reg_Modules(j,3). EQ. k ) then
   20 continue	
     	Prior_ParaModules(i,1) = w1*Particle_ParaModules(ir,j)
     $	    + w2*PriorG0_ParaModules(i,1) 
     $      + s * dsqrt( PriorG0_ParaModules(i,2) )* rgauss(ix)
      if( Prior_ParaModules(i,1).lt. 0.0d0 ) goto 20     
	if ( strModules(j,2). EQ. tag_base  ) Prior_ParaModules(i,1) = 0.0d0	
     	Prior_ParaModules(i,2) = PriorG0_ParaModules(i,2)
	goto 2
	endif
	enddo
    2 continue
*	write(*,'(a,i3,e14.5,e14.5,e14.5)') 
*     $	'mod', i, Prior_ParaModules(i,1),
*     $	PriorG0_ParaModules(i,1), Particle_ParaModules(ir,j)	
	enddo
!$omp end do	
*
*
!$omp do
	do i = 1, num_Variable
!$    id = omp_get_thread_num()	
	k = ID_Prior_Inits(i)
   30 continue	
	Prior_Inits(i,1) = w1 * Particle_Inits(ir,k)
     $	    + w2*PriorG0_Inits(i,1) 
     $      + s * dsqrt( PriorG0_Inits(i,2) ) * rgauss(ix)
      if( Prior_Inits(i,1).lt. 0.0d0 ) goto 30     
	Prior_Inits(i,2) = PriorG0_Inits(i,2)
*	write(*,'(a,i3,e14.5,e14.5,e14.5)') 'ini', i, Prior_Inits(i,1), 
*     $	PriorG0_Inits(i,1), Particle_Inits(ir,k) 	
	enddo
!$omp end do	
*
!$omp end parallel
*
      return
      end
*
*
      subroutine scaling( w, n, pr )
*      
      implicit real*8(a-h,o-z)
      dimension w(*), pr(*)
*
      sum = 0.0d0
!$omp parallel 
*
!$omp do reduction(+:sum)
      do i = 1, n
      sum = sum + w(i)
      enddo
!$omp end do
*
!$omp do
      do i = 1, n
      pr(i) = w(i)/sum
      enddo
!$omp end do
*      
!$omp end parallel
*
      return
      end
*
*
      subroutine scaling_byrow( w, n, m, nn )
*      
      implicit real*8(a-h,o-z)
      dimension w(nn,*)
*
!$omp parallel 
!$omp do reduction(+:sum)
      do j = 1, m
      sum = 0.0d0
      do i = 1, n
      sum = sum + w(i,j)
      enddo
      if( sum.eq.0.0d0 ) sum = 1.0d-20 
      do i = 1, n
      w(i,j) = w(i,j) / sum
      enddo
      enddo
!$omp end do      
!$omp end parallel
*
      return
      end      
*
*
      subroutine reaction_lists(
*      
     $      List_Reg_Modules, num_Reg_Modules, max_num_Reg_Modules,
     $      strModules, ID_Prior_ParaModules, max_num_Para2, num_Para2,
     $      List_Reg_Elements, num_Reg_Elements, max_num_Reg_Elements,
     $      strElements, ID_Prior_ParaElements, max_num_Para1,num_Para1,
     $      ID_Prior_Inits, num_Variable, max_num_Variable,
     $      list_m, list_e, list_i,
     $      list_mm, list_ee, list_ii     
*     
     $  )
*
      implicit real*8(a-h, o-z)
*
*	@ (1) Module IDs, (2) product variables, (3) parameter ID, (4) number of reactant variables 
*       (5-Last) Element IDs associated to the reaction module
	dimension List_Reg_Modules( max_num_Reg_Modules, * )
*	@ (1) Element IDs, (2) kinetic law IDs, and (3) reactant state variables
	dimension List_Reg_Elements( max_num_Reg_Elements, * )	
*     @ 'Prior_ParaModules' is mapped to 'Reg_Modules' (IDs saved at the third column)	
	dimension ID_Prior_ParaModules( * )
*     @	'Prior_ParaElements' is mapped to 'Reg_Elements' (IDs saved at the first column)
	dimension ID_Prior_ParaElements( * )
*     @ 'Prior_Inits' is mapped to 'StateVariable'	
	dimension ID_Prior_Inits( * )	
	character strElements( max_num_Reg_Elements, * ) * (*)
	character strModules( max_num_Reg_Modules, * ) * (*)   ! (1) Product, (2) Reaction, (3) Reactants
*
	dimension list_m( * )
	dimension list_e( * ) 
	dimension list_i( * )
	dimension list_mm( max_num_Para2, * )
	dimension list_ee( max_num_Para1, * ) 
	dimension list_ii( max_num_Variable, * )	
*
	character tag_base * (*)
	parameter ( tag_base = 'Baseline' )	
*
*
*!$omp parallel 
*
*!$omp do private(k,j,n)
      do i = 1, num_Para1
      k = ID_Prior_ParaElements(i)
      n = 0
      do j = 1, num_Reg_Elements
      if( List_Reg_Elements(j,1).eq.k ) then
      list_e(i) = j
      n = n + 1
      list_ee(i,1) = n
      list_ee(i,1+n) = j
      endif
      enddo
      enddo
*!$omp end do      
*
*!$omp do private(k,j,n)
      do i = 1, num_Para2
      k = ID_Prior_ParaModules(i)
      n = 0
      do j = 1, num_Reg_Modules
      if( ( List_Reg_Modules(j,3).eq.k ). and.
     $    ( strModules(j,2).ne.tag_base ) ) then
      list_m(i) = j
      n = n + 1
      list_mm(i,1) = n
      list_mm(i,1+n) = j
      endif
      if( ( List_Reg_Modules(j,3).eq.k ). and.
     $    ( strModules(j,2).eq.tag_base ) ) then 
      list_m(i) = 0
      endif
      enddo
      enddo
*!$omp end do
*
*!$omp do
      do i = 1, num_Variable
      list_i(i) = ID_Prior_Inits(i)
      enddo
*!$omp end do      
*
*!$omp end parallel
*      	
      return
      end	
*      
*      
      subroutine Gradient_Particle_Move(
*      
     $      w, m, mm,
     $      Particle_ParaElements, max_num_Reg_Elements,
     $      num_Reg_Elements, 
     $  	Particle_ParaModules, max_num_Reg_Modules,
     $      num_Reg_Modules,
     $      Particle_Inits, max_num_Variable, num_Variable,
     $      list_m, list_e, list_i, num_Para1, num_Para2,
     $      grad_m, grad_e, grad_i,
     $      list_mm, list_ee, list_ii, max_num_Para1, max_num_Para2,
*
     $      Para_Reg_Elements, Para_Reg_Modules,
     $      List_Reg_Elements, List_Reg_Modules, max_tmp,
     $      time_interval, time_StateVariable, max_time_StateVariable,
     $		includeNoise, ix, Variance_System, Variance_Measure,  
     $		time_Data, max_time_Data,
     $		StateVariable, Data, ID_Time_Data, ID_MissVariable
*          
     $     )
*
      implicit real*8(a-h, o-z)
	integer*2  tab /2313/
	parameter ( num = 300 )
	dimension w(mm), pr(mm)
	dimension Particle_ParaElements(mm,max_num_Reg_Elements,2)
	dimension Particle_ParaModules(mm,max_num_Reg_Modules)
	dimension Particle_Inits(mm,max_num_Variable)
	dimension T_Particle_ParaElements(mm,max_num_Reg_Elements,2)
	dimension T_Particle_ParaModules(mm,max_num_Reg_Modules)
	dimension T_Particle_Inits(mm,max_num_Variable)	
	dimension list_m(*)
	dimension list_e(*)
	dimension list_i(*)
	dimension list_mm( max_num_Para2, * )
	dimension list_ee( max_num_Para1, * ) 
	dimension list_ii( max_num_Variable, * )	
	dimension grad_m(mm,*)
	dimension grad_e(mm,*)
	dimension grad_i(mm,*)
	dimension v(num, mm)
	dimension cov( num, num ), ave( num )
*
	dimension StateVariable( max_num_Variable, max_time_StateVariable)
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	integer*4 time_StateVariable	
	integer*4 time_Data
	dimension Variance_Measure( * )
	dimension Variance_System( * )
	dimension Data( max_num_Variable, * )	
	dimension ID_Time_Data( * )
	dimension ID_MissVariable( * )	
	character Output_File * 30
	character str * 15
	character str2 * 15
	dimension valueLik( mm )
*
      width = 0.1d0
      call scaling( w, m, pr )
*
!$omp parallel do private(k)
      do i = 1, num_Para1
      k = list_e(i)
      do j = 1, m
      v(i,j) = Particle_ParaElements(j,k,1)
      enddo
      enddo
!$omp end parallel do
*
      call covariance_matrix( cov, ave, v, m, num_Para1, mm, num )      
*
!$omp parallel do private(k,u,s,x,sum)
      do i = 1, num_Para1
      k = list_e(i)
      s = cov(i,i)
*      write(*,*) i, 'variance', s      
      do j = 1, m
      x = Particle_ParaElements(j,k,1)
      sum = 0.0d0
      do l = 1, m
      u = Particle_ParaElements(l,k,1)
      sum = sum + w(l)*dGauss_Univ( x, u, s )*(x-u)/s
      enddo
      grad_e(j,i) = sum
*      write(*,'(i4,a1,i4,a1,f10.5,a1,e15.5,a1,e15.5)') 
*     $       i,tab,j, tab, x, tab, sum, tab, w(j)
      enddo
      enddo
!$omp end parallel do
*
*
!$omp parallel do private(k)
      do i = 1, num_Para2
      k = list_m(i)
      if( k.ne.0 ) then
      do j = 1, m
      v(i,j) = Particle_ParaModules(j,k)
      enddo
      else
      do j = 1, m
      v(i,j) = runif(ix)
      enddo
      endif
      enddo
!$omp end parallel do
*
      call covariance_matrix(cov, ave, v, m, num_Para2, mm, num)      
*
!$omp parallel do private(k,u,s,x,sum)
      do i = 1, num_Para2
      k = list_m(i)
      if( k.ne.0 ) then
      s = cov(i,i)
*      write(*,*) i, 'variance', s      
      do j = 1, m
      x = Particle_ParaModules(j,k)
      sum = 0.0d0
      do l = 1, m
      u = Particle_ParaModules(l,k)
      sum = sum + w(l)*dGauss_Univ( x, u, s )*(x-u)/s
      enddo
      grad_m(j,i) = sum
*      write(*,'(i4,a1,i4,a1,f10.5,a1,e15.5,a1,e15.5)') 
*     $       i,tab,j, tab, x, tab, sum, tab, w(j)
      enddo
      endif
      enddo
!$omp end parallel do
*
*
!$omp parallel do private(k)
      do i = 1, num_Variable
      k = list_i(i)
      do j = 1, m
      v(i,j) = Particle_Inits(j,k)
      enddo
      enddo
!$omp end parallel do
*
*
      call covariance_matrix(cov, ave, v, m, num_Variable, mm, num)      
*
!$omp parallel do private(k,u,s,x,sum)
      do i = 1, num_Variable
      k = list_i(i)
      s = cov(i,i)
*      write(*,*) i, 'variance', s      
      do j = 1, m
      x = Particle_Inits(j,k)
*      write(*,*) k,x 
      sum = 0.0d0
      do l = 1, m
      u = Particle_Inits(l,k)
      sum = sum + w(l)*dGauss_Univ( x, u, s )*(x-u)/s
      enddo
      grad_i(j,i) = sum
*      write(*,'(i4,a1,i4,a1,f10.5,a1,e15.5,a1,e15.5)') 
*     $       i,tab,j, tab, x, tab, sum, tab, w(j)
      enddo
      enddo
!$omp end parallel do
*
      call scaling_byrow( grad_i, m, num_Variable, mm )
      call scaling_byrow( grad_e, m, num_Para1, mm )
      call scaling_byrow( grad_m, m, num_Para2, mm )      
*
      r = 0.001d0
*      
*!$omp parallel private(k)
*!$omp do
*      do i = 1, num_Para1
*      k = list_e(i)
*      do j = 1, m
*      T_Particle_ParaElements(j,k,1) = Particle_ParaElements(j,k,1) + 
*     $      r * grad_e(j,i)
*      enddo
*      enddo
*!$omp end do      
*
*!$omp do
*      do i = 1, num_Para2
*      k = list_m(i)
*      if( k.ne.0 ) then
*      do j = 1, m
*      T_Particle_ParaModules(j,k) = Particle_ParaModules(j,k) + 
*     $      r * grad_m(j,i)
*      write(*,'(i4,a1,i4,a1,e15.5,a1,e15.5,a1,e15.5)') 
*     $       i,tab,j, tab, grad_m(j,i), tab, r
*      enddo
*      endif
*      enddo
*!$omp end do      
*
*!$omp do
*      do i = 1, num_Variable
*      k = list_i(i)
*      do j = 1, m
*      Particle_Inits(j,k) = Particle_Inits(j,k) + 
*     $      r * grad_i(j,i)
*      enddo
*      enddo
*!$omp end do
*!$omp end parallel            
*
*
* (OpenMP START)
* --------------------------------------------------------------------------
!$omp   parallel firstprivate(
!$omp&     StateVariable, Para_Reg_Modules, Para_Reg_Elements, 
!$omp&     str, str2, Output_File, ki, n, id)
* -------------------------------------------------------------------------	
*
!$omp do
	do k = 1, m
*
!$    id = omp_get_thread_num()
*
	do i = 1, num_Reg_Elements
	do j = 1,2
	Para_Reg_Elements(i,j) = Particle_ParaElements(k,i,j)
	enddo
	enddo
*
	do i = 1, num_Reg_Modules
      Para_Reg_Modules(i,1) = Particle_ParaModules(k,i)
	enddo
*
	do i = 1, num_Variable
      StateVariable(i,1) = Particle_Inits(k,i)
	enddo
*
      do i = 1, num_Para1
      n = list_ee(i,1)
      do j = 1, n
      ki = list_ee(i,1+j)
      Para_Reg_Elements(kj,1) = Para_Reg_Elements(ki,1) + 
     $      r * grad_e(k,i)
      enddo
      enddo
*
      do i = 1, num_Para2
*      write(*,*) id, i, list_m(i)
      if( list_m(i). ne. 0 ) then
      n = list_mm(i,1)
      do j = 1, n
      ki = list_mm(i,1+j)
      Para_Reg_Modules(kj,1) = Para_Reg_Modules(ki,1) + 
     $      r * grad_m(k,i)
      enddo
      endif
      enddo	
*      stop
*
      do i = 1, num_Variable
      StateVariable(i,1) = StateVariable(i,1) + 
     $      r * grad_i(k,i)
      enddo
*
*      write(*,*) max_tmp,	max_num_Reg_Modules, max_num_Reg_Elements
*      write(*,*) time_interval, time_StateVariable, 
*     $ max_time_StateVariable
*      write(*,*) num_Variable, max_num_Variable, ix
*      stop
*
*	/ Run initial simulation
	call Run_BioSimulator(
*		-- Reaction functions
     $		List_Reg_Elements, Para_Reg_Elements,
     $		List_Reg_Modules, Para_Reg_Modules,
     $		num_Reg_Elements, num_Reg_Modules, 
     $		max_num_Reg_Elements, max_num_Reg_Modules, max_tmp, 
*		-- Parameters for simulation
     $		time_interval, time_StateVariable, max_time_StateVariable,
     $		includeNoise, ix, Variance_System,  
*		-- Others
     $		num_Variable, max_num_Variable,
*		-- Outputs
     $		StateVariable
     $	)
*
*      do ii = 1, 10 ! time_StateVariable
*      write(*,'(100f10.5)') (StateVariable(ii,jj),jj=1,num_Variable)
*      enddo
*      stop
*
	valueLik(k) = Gauss_Likelihood(
*
*		-- Values for data and state variables	
     $		Data, StateVariable, time_Data, time_StateVariable, 
     $		num_Variable, max_time_Data, max_time_StateVariable, 
     $		max_num_Variable,
*	
*		-- Indicators for observed time points and missing variables
     $		ID_Time_Data, ID_MissVariable,
*
*		--	Observed noise variances
     $		Variance_Measure
*
     $	)     
*
      valueLik(k) = dexp( - valueLik(k) / 2.0d0 )
*      write(*,*) k, w(k), valueLik(k)
*
      if( valueLik(k).ge.w(k) ) then
*
*     /Replace particles
	do i = 1, num_Reg_Elements
	do j = 1,2
	Particle_ParaElements(k,i,j) = Para_Reg_Elements(i,j)
	enddo
	enddo
*
	do i = 1, num_Reg_Modules
      Particle_ParaModules(k,i) = Para_Reg_Modules(i,1) 
	enddo
*
	do i = 1, num_Variable
      Particle_Inits(k,i) = StateVariable(i,1)
	enddo
*      
      endif      
*      
      enddo
!$omp end do
*      
* (OpenMP END)
* --------------------------------------------------------------------------
!$omp end parallel
* --------------------------------------------------------------------------
*
	return
	end
*
*
      subroutine Run_SimulationOverParticleSet(
*      
     $      Particle_ParaElements, Particle_ParaModules,
     $      Particle_Inits, num_Particle_DPM, nsave_Particle,
     $      Para_Reg_Elements, Para_Reg_Modules,
     $      List_Reg_Elements, List_Reg_Modules,
     $      num_Reg_Elements, num_Reg_Modules,
     $      max_num_Reg_Elements, max_num_Reg_Modules, max_tmp,
     $      time_interval, time_StateVariable, max_time_StateVariable,
     $		includeNoise, ix, Variance_System, Variance_Measure,  
     $		num_Variable, max_num_Variable, time_Data, max_time_Data,
     $		StateVariable, Data, ID_Time_Data, ID_MissVariable,
     $      weight_DPM
*     
     $ )
*
      implicit real*8(a-h, o-z)
	dimension StateVariable( max_num_Variable, max_time_StateVariable )
	dimension Para_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension Para_Reg_Modules( max_num_Reg_Modules, max_tmp )
	dimension List_Reg_Elements( max_num_Reg_Elements, max_tmp )
	dimension List_Reg_Modules( max_num_Reg_Modules, max_tmp )
	integer*4 time_StateVariable	
	integer*4 time_Data
	dimension Particle_ParaElements( nsave_Particle, 
     $	                    max_num_Reg_Elements, * )
	dimension Particle_ParaModules( nsave_Particle, * )
	dimension Particle_Inits( nsave_Particle, * )
	dimension Variance_Measure( * )
	dimension Variance_System( * )
	dimension Data( max_num_Variable, * )	
	dimension ID_Time_Data( * )
	dimension ID_MissVariable( * )	
	character Output_File * 30
	character str * 15
	character str2 * 15
	dimension valueLik( nsave_Particle )
	dimension weight_DPM(*)
*
* (OpenMP START)
* --------------------------------------------------------------------------
!$omp   parallel firstprivate(
!$omp&     StateVariable, Para_Reg_Modules, Para_Reg_Elements, 
!$omp&     str, str2, Output_File, i,j,k )
* -------------------------------------------------------------------------	
*
!$omp do
	do k = 1, num_Particle_DPM
*
!$    id = omp_get_thread_num()
*
	do i = 1, num_Reg_Elements
	do j = 1,2
	Para_Reg_Elements(i,j) = Particle_ParaElements(k,i,j)
	enddo
	enddo
*
	do i = 1, num_Reg_Modules
      Para_Reg_Modules(i,1) = Particle_ParaModules(k,i)
	enddo
*
	do i = 1, num_Variable
      StateVariable(i,1) = 	Particle_Inits(k,i)
	enddo
*
*	/ Run initial simulation
	call Run_BioSimulator(
*		-- Reaction functions
     $		List_Reg_Elements, Para_Reg_Elements,
     $		List_Reg_Modules, Para_Reg_Modules,
     $		num_Reg_Elements, num_Reg_Modules, 
     $		max_num_Reg_Elements, max_num_Reg_Modules, max_tmp, 
*		-- Parameters for simulation
     $		time_interval, time_StateVariable, max_time_StateVariable,
     $		includeNoise, ix, Variance_System,  
*		-- Others
     $		num_Variable, max_num_Variable,
*		-- Outputs
     $		StateVariable
     $	)
*
	valueLik(k) = Gauss_Likelihood(
*
*		-- Values for data and state variables	
     $		Data, StateVariable, time_Data, time_StateVariable, 
     $		num_Variable, max_time_Data, max_time_StateVariable, 
     $		max_num_Variable,
*	
*		-- Indicators for observed time points and missing variables
     $		ID_Time_Data, ID_MissVariable,
*
*		--	Observed noise variances
     $		Variance_Measure
*
     $	)     
*
      valueLik(k) = dexp( - valueLik(k) / 2.0d0 )
      write(*,*) k, weight_DPM(k), valueLik(k)
*      
      enddo
!$omp end do
*      
* (OpenMP END)
* --------------------------------------------------------------------------
!$omp end parallel
* --------------------------------------------------------------------------
*
      return
      end
